/*
   This file is part of the MoCA video library.
   Name: ObjectSegmentationTest.cpp
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  
   \author    Stephan Kopf
   
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de
	
	 (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	 
*/


// C++ includes
#include <string>
#include <iostream>
using namespace std;

// MoCA includes
#include <ObjectMaskFeature.h>
#include <ObjectSegmentationFeature.h>
#include <DrawLine.h>
#include <IImageStream.h>
#include <RegionFeature.h>

#include <CameraTransformation.h>
#include <CameraTransformationPrecise.h>
#include <LogoDetection.h>
#include <VideoEncoder.h>
#include <InverseFilter.h>


const string objectSegmentationVersion = "1.0";
void usage(bool extendedHelp=false)
{
	cerr << endl;
	cerr << "ObjectSegmentation " << objectSegmentationVersion 
	 	 << "    12.Feb.2002   (c) Stephan Kopf, University of Mannheim" << endl;
	cerr << "-------------------------------------------------------------------------------" << endl;
	cerr << endl;
	cerr << "Usage: ObjectSegmentation input-file [options]" << endl;
	cerr << endl;
	
	if (extendedHelp) {
		cerr << "Options:" << endl;
		cerr << " -recurse           the input-file is the name of a directory. Objects are" << endl;
		cerr << "                    searched in all *.mpg files in this directory (and sub-" << endl;
		cerr << "                    directories). The parameter output-file is the " << endl;
		cerr << "                    destination directory. The prefix \"objects_\" is added" << endl;
		cerr << "                    to the filename of the XML file." << endl;
		cerr << " -keepMetadata      metadata files are not deleted when program terminates" << endl;
		cerr << endl;
	} // if
	
	if (!extendedHelp) {
		cerr << "Description:" << endl;
		cerr << "The object segmentation application analyses one MPEG-video, finds the shots" << endl;
		cerr << "of the video and analyses each shot. The algorithms search for moving " << endl;
		cerr << "objects in each shot. If there is a moving object in a shot, the" << endl;
		cerr << "application extracts the contour of this object." << endl;
		cerr << endl;
	} // if
	
	cerr << "Example:" <<  endl;
	cerr << "ObjectSegmentation source.mpg " << endl;
	cerr << "Analyses the moving objects of \"source.mpg\"." << endl;
	cerr << endl;
	exit(1);
	return;
} // usage

struct ObjectSegmentationParameters
{
	ObjectSegmentationParameters () 
	{
		recurse        = false;
		deleteMetadata = true;
		srcFilename    = "";
	}
	bool recurse;
	bool deleteMetadata;
	string srcFilename;
};

void parseParameters (int argc, char* argv[], ObjectSegmentationParameters &param)
{
	string s;
	if (argc < 2) usage();
	if (argc == 2) {
		s=argv[1];
		for (string::size_type k=0; k<s.size(); k++)
			s[k]=toupper(s[k]);
		if (s == "HELP") usage();
		if (s == "-HELP") usage(true);
		if (s == "--HELP") usage(true);
		if (s == "/?") usage(true);
	} // if
	
	// set filenames
	param.srcFilename = argv[1];
	
	// read options
	string s1;
	for (int i=2; i<argc; i++) {
		s=argv[i];
		FileHandling::upper(s);
		
		// recurse
		s1=s.substr(0,8);	
		if (s1 == "-RECURSE") {
			param.recurse = true;
			continue;
		} // if

		// keep metadata
		s1=s.substr(0,13);	
		if (s1 == "-KEEPMETADATA") {
			param.deleteMetadata = false;
			continue;
		} // if
		
		usage(false);
	} // for
	
	return;
} // parseParameters



void searchObjectsInFile (ObjectSegmentationParameters &param)
{

	try {
		// validate filenames
		FileHandling::convertToValidUnixFilename (param.srcFilename);
		
		cerr << endl;
		cerr << "source MPEG file: " << param.srcFilename << endl;

		time_t start = time(NULL);


		// start process: object masks
		{
			ObjectMaskData objectMasks;
			int minimumShotLength = 12;
			bool preciseCalculation=false;
			ObjectMask objRecognition(param.srcFilename, minimumShotLength, preciseCalculation);
			objRecognition.process(objectMasks);
			//objRecognition.storeFrames(objectMasks);
			//cerr << objectMasks << endl;
			//cerr << "exit in ObjectSegmentationTest" << endl;
		}

		VideoDecoder reader (param.srcFilename);
		int32 frames = reader.getSize()[DIM_Z];
		ieee64 vidLength  = (frames / reader.getFrameRate())/60;
		ieee64 calcLength = (time(NULL) - start)/60.0;
		cerr << endl 
			 << "-----------------------------------------" << endl 
			 <<"calculation    : " <<  (int32)calcLength << " min" << endl
			 <<"video          : " <<  (int32)vidLength << " min" << endl
			 <<"factor realtime: " <<  (int32)(calcLength / vidLength) << endl
			 << "-----------------------------------------" << endl;

		//char beep = 7; cout << beep; cout << beep;


	} catch (...)
	{
		cerr << endl << "Exception in in ObjectSegmentationTest!" << endl;
	}
	
	return;
} // searchObjectsInFile 



void analyseShot (const Video<uint8> &vid)
{

	// get object masks
	bool preciseCalculation=true;
	VideoSize vidSize = vid.getSize();

	ObjectMaskData objectMasks;
	ObjectMask objRec (vidSize, preciseCalculation, POSITION_CENTER);
	objRec.process(vid, objectMasks);
	cerr << objectMasks << endl;


	return;
}

int main(int argc, char* argv[])
{

	// analyse parameters
	ObjectSegmentationParameters param;
	parseParameters (argc, argv, param);

	
	if (!param.recurse) {
		
		searchObjectsInFile (param);		
		
	} else {

		
		// timer
		time_t startTime = time(NULL);
		ieee64 vidLength  = 0.0; // in minutes
		ieee64 calcLength = 0.0; // in minutes


		// set source path name
		string fullPathSrc = "";
		{
			CreateMocaPath mocaPath1 (param.srcFilename, false);
			if (mocaPath1.pathFile.length() == mocaPath1.path.length()) {
				fullPathSrc = mocaPath1.path;
			} else {
				string s = param.srcFilename + '/';
				CreateMocaPath mocaPath2 (s, false);
				fullPathSrc = mocaPath2.path;
			} // else
		}
		
		
		// set destination path name
		cerr << endl;
		cerr << "source path: " << fullPathSrc << endl;
		
		// read all MPEG files in directory
		vector<FileInfo> mpgFiles;
		FileList files(fullPathSrc, true);    
		files.process("*.mpg", mpgFiles);


		// open mpeg files
		vector<FileInfo>::iterator it;
		for (it = mpgFiles.begin(); it != mpgFiles.end(); it++) {

			// set filenames
			param.srcFilename = (*it).name;
			searchObjectsInFile (param);

			// set timer
			VideoDecoder reader (param.srcFilename);
			int32 frames = reader.getSize()[DIM_Z];
			vidLength  += (frames / reader.getFrameRate())/60;			

		} // for
		cout << endl << mpgFiles.size() << " files found." << endl;

		// timer
		calcLength = (time(NULL) - startTime)/60.0;
		cerr << endl 
			 << "-----------------------------------------" << endl 
			 <<"calculation time      : " <<  (int32)calcLength << " minutes" << endl
			 <<"total length of videos: " <<  (int32)vidLength << " minutes" << endl
			 <<"factor realtime       : " <<  (int32)(calcLength / vidLength) << endl
			 << "-----------------------------------------" << endl;
	} // else
	
	return 0;
} // main


