/*
   This file is part of the MoCA video library.
   Name: CompareImages.cpp
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

   \author    Stephan Kopf

              Lehrstuhl fuer Praktische Informatik IV
              Universitaet Mannheim
              68131 Mannheim
              Germany
              email: kopf@informatik.uni-mannheim.de

   (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt

*/

#include <iostream>
#include <Image.h>
#include <IImageStream.h>
#include <OImageStream.h>
#include <MocaExceptions.h>


#include <CannyFilter.h>
#include <RotationFilter.h>


#include <GrayscaleFilter.h>

int main(int argc, char* argv[])
{
    if (argc != 4) {
        cerr << "Usage:   CompareImages src1 src2 border" << endl;
        cerr << "example: CompareImages test1.png test2.jpg 0" << endl;
        return 1;
    } // if

    int border = atoi (argv[3]);
    Image<uint8> img1;
    cout << "read image: " << argv[1] << endl;
    try {
        // open input filestream and read image
        IImageStream<uint8> iis(argv[1]);
        iis.read(img1);
    } catch (MocaException e) {
        cout << "IImageStream " << e.what() << endl;
        exit (-1);
    } // try
    
    Image<uint8> img2;
    cout << "read image: " << argv[2] << endl;
    try {
        // open input filestream and read image
        IImageStream<uint8> iis(argv[2]);
        iis.read(img2);
    } catch (MocaException e) {
        cout << "IImageStream " << e.what() << endl;
        exit (-1);
    } // try


    cout << "compare images:" << endl;
    ImageSize size1 = img1.getSize();
    ImageSize size2 = img2.getSize();
    if (size1[DIM_X] != size2[DIM_X]) {
        cout << "different size X"  << size1[DIM_X] << "  " << size2[DIM_X] << endl;
        exit(1);
    } // if
    if (size1[DIM_Y] != size2[DIM_Y]) {
        cout << "different size Y" << size1[DIM_Y] << "  " << size2[DIM_Y] << endl;
        exit(1);
    } // if
    if (size1[DIM_BAND] != size2[DIM_BAND]) {
        cout << "different size BAND" << size1[DIM_BAND] << "  " << size2[DIM_BAND] << endl;
        exit(1);
    } // if

    int x,y,b;
    ieee64 diff=0;
    ieee64 maxDiff=0;
    ieee64 maxDest=0;
    ieee64 tmp;
    
    Image<uint8> dest (img1);
    size2[DIM_BAND] = 1;
    dest.realloc(size2, ImageSize(0,0), BORDER_REFLECTING);
    string diffName = "difference.png";
    for (y=border; y<size1[DIM_Y]-border; y++) {
        for (x=border; x<size1[DIM_X]-border; x++) {
            ieee64 sumBand=0;
            for (b=0; b<size1[DIM_BAND]; b++) {
                tmp = abs(img1(x,y,b) - img2(x,y,b));
                diff += tmp;
                sumBand+=tmp;
                if (tmp > maxDiff) {
                    maxDiff = tmp;
                } // if
            } // for
            if (sumBand >= 255) {
                sumBand = 255;
            } // if
            dest(x,y,0) = (uint8) sumBand;
            if (dest(x,y,0) > maxDest) {
                maxDest = dest(x,y,0);
            } // if
        } // for
    } // for

    // set border to zero
    for (x=0; x<size1[DIM_X]; x++) {
        for (y=0; y<border; y++) {
            dest(x,y) = 0;
            dest(x,size1[DIM_Y]-y-1) = 0;
        } // for
    } // for
    for (y=0; y<size1[DIM_Y]; y++) {
        for (x=0; x<border; x++) {
            dest(x,y) = 0;
            dest(size1[DIM_X]-x-1,y) = 0;
        } // for
    } // for

    for (y=border; y<size1[DIM_Y]-border; y++) {
        for (x=border; x<size1[DIM_X]-border; x++) {
            for (b=0; b<size1[DIM_BAND]; b++) {
                ieee64 v = dest(x,y,0) * 255.0 / maxDest;
                if (v>255) {
                    v = 255;
                } // if
                dest(x,y,0) = (uint8) v;
            } // for
        } // for
    } // for
    
    
    cout << endl;    
    cout << "total difference     : " << (int) diff << endl; 
    cout << "difference per pixel : " << diff / (size1[DIM_X]*size1[DIM_Y]*size1[DIM_Z]*size1[DIM_BAND]) << endl;
    cout << "maximum difference   : " << (int) maxDiff << endl; 
    cout << "difference image     : " << diffName << endl;
    try {
        // open output filestream and write image
        OImageStream<uint8> ois(diffName);
        ois.write(dest);
    } catch (MocaException e) {
        cout << "IImageStream " << e.what() << endl;
        exit (-1);
    } // try
    
    
    return 0;
} // main


