/*
    AudioCut.cpp     detects audio cuts
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// MoCA includes
#include <AudioCut.h>


std::ostream& operator<< (ostream &os, const AudioCutData &data) 
{
	os << "version          : " << data.version << endl;
	os << "filename         : " << data.filename << endl;
	os << "number of blocks : " << data.numberOfBlocks<< endl;
	os << "number of cuts   : " << data.numberOfCuts << endl;
	os << "minimum number of cuts      : " << data.minimumNumberOfCuts << endl;
	os << "blocklength in milli seconds: " << data.blockLengthMilliSeconds << endl;
	
	os << endl;
	
	for (int32 i=0; i<data.numberOfBlocks; i++) {
		if (data.cuts(i) != 0) {
			if (i<10) os << " ";
			if (i<100) os << " ";
			if (i<1000) os << " ";
			if (i<10000) os << " ";
			os << i << "\t  " << (int32)data.cuts(i) << endl;
		} // if
	} // for

	return os;
} // operator <<

OTagStream& operator<< (OTagStream& os, const AudioCutData &data) 
{
    TagFormat::writeStartTag (os, "AUDIO_CUTS");
	os << endl;
	
    TagFormat::writeNodeValue(os, "version", data.version, 1);
    TagFormat::writeNodeValue(os, "filename", data.filename, 1);
    TagFormat::writeNodeValue(os, "numberOfBlocks", data.numberOfBlocks, 1);
    TagFormat::writeNodeValue(os, "numberOfCuts", data.numberOfCuts, 1);
    TagFormat::writeNodeValue(os, "minimumNumberOfCuts", data.minimumNumberOfCuts, 1);
    TagFormat::writeNodeValue(os, "blockLengthMilliSeconds", data.blockLengthMilliSeconds, 1);
	os << endl;

	TagFormat::writeStartTag (os, "CUTS", 1);
	os << endl;

	for (int32 i=0; i<data.numberOfBlocks; i++) {
		os << setfill(' ') << setw(6) << i << "  " << (int32)data.cuts(i) << endl;
	} // for
	
	TagFormat::writeEndTag (os, "CUTS", 1);
	os << endl;

	
    TagFormat::writeEndTag (os, "AUDIO_CUTS");
	os << endl;
	
    return os;
} // operator<<


ITagStream& operator>> (ITagStream& is, AudioCutData& data)
{
	
    char c;
    bool failFlag = true;
	
	if (is >> c) {
		failFlag = false;
		is.putback(c);

		TagFormat::readStartTag (is, "AUDIO_CUTS");
		
		TagFormat::readNodeValue(is, "version", data.version);
		TagFormat::readNodeValue(is, "filename", data.filename);
		TagFormat::readNodeValue(is, "numberOfBlocks", data.numberOfBlocks);
		TagFormat::readNodeValue(is, "numberOfCuts", data.numberOfCuts);
		TagFormat::readNodeValue(is, "minimumNumberOfCuts", data.minimumNumberOfCuts);
		TagFormat::readNodeValue(is, "blockLengthMilliSeconds", data.blockLengthMilliSeconds);

		if (is.rdstate() & ios::failbit) {
			throw MocaException ("AudioCut: Illegal index file format!");
		} // if

		// initialize size of cuts
		data.cuts.resize(SignalSize(data.numberOfBlocks));
		for (int32 i=0; i<data.numberOfBlocks; i++) {
			data.cuts(i) = 0;
		} // for
		
		// read cuts
		if (is >> c) {
			is.putback(c);
			
			// read hard cuts
			TagFormat::readStartTag(is, "CUTS");
			
			int32 f; // frame number
			int32 v; // value
			for (int32 i=0; i<data.numberOfBlocks; i++) {
				is >> f;
				is >> v;
				data.cuts(i) = v;
			} // for
			
			TagFormat::readEndTag(is, "CUTS");
			if (is.rdstate() & ios::failbit) {
				throw MocaException ("AudioCut: Illegal index file format!");
			} // if
		} // if
		
		TagFormat::readEndTag (is, "AUDIO_CUTS");
	} // if
	
	if (is.rdstate() & ios::failbit) {
		failFlag = true;
	} // if
    if (failFlag) {
		throw MocaException ("AudioCut: Illegal index file format!");
	} else {
        is.clear();
    } // else

    return is;
} // operator>> 





AudioCut::AudioCut(const string &pathFilename, int32 minimumNumberOfCuts, int32 blockLengthMilliSeconds) 
{
	this->pathFilename = pathFilename;
	this->blockLengthMilliSeconds = blockLengthMilliSeconds;
	this->minimumNumberOfCuts = minimumNumberOfCuts;
} 


void 
AudioCut::process(AudioCutData& data) 
{
	// read data
	ReadMocaData<AudioCutData> reader (pathFilename, AUDIO_CUT_FILENAME);
	bool dataValid = reader.process (AUDIO_CUT_VERSION, data); 
	if (dataValid) {
		if ((blockLengthMilliSeconds != data.blockLengthMilliSeconds) ||
			(minimumNumberOfCuts != data.minimumNumberOfCuts)) {
			dataValid=false;
		} // if
	} // if

	if (!dataValid) {
		
		// create new index
		cerr << "\rAudioCut: Creating index file ...          " << endl;
		readValues(data);
		
		// save index file
		WriteMocaData<AudioCutData> writer (pathFilename, AUDIO_CUT_FILENAME);
		writer.process(data);
		cerr << "\rAudioCut: Creating index file ... done          " << endl;
	} // if

	return;
} // process


void 
AudioCut::readValues (AudioCutData& data) 
{

	// get path of filename, do not create new directory
	CreateMocaPath mocaPath (pathFilename, false);
	data.filename = mocaPath.file;
	data.minimumNumberOfCuts = minimumNumberOfCuts;
	data.blockLengthMilliSeconds = blockLengthMilliSeconds;



	int32 samplesSilentLong =0;   // silence is longer than 5 seconds
	int32 samplesSilentShort=0;   // silence is shorter than 1 second
	{
		IWaveAudio reader (pathFilename);
		WaveHeader header = reader.getHeader();
		ieee64 energySamplesPerSecond = 1000.0 / blockLengthMilliSeconds;
		samplesSilentLong  = (int32)(energySamplesPerSecond*5.0); // silence is longer than 5 seconds
		samplesSilentShort = (int32)(energySamplesPerSecond*1.0); // silence is shorter than 1 second
	}


	// read the energy of the audio signal (one energy value for both channels) and
	// get average values
	Signal<int32> energy;
	SignalSize energySize;
	int32 avgEnergy=0;
	int32 lowEnergyTh0 = 0;
	int32 lowEnergyTh1 = 0;
	int32 lowEnergyTh2 = 0;
	{
		AudioEnergyData energyData;
		AudioEnergy audioEnergy(pathFilename, blockLengthMilliSeconds);
		audioEnergy.process(energyData);

		// combine both channels to one energy value
		energySize = energyData.size;
		energySize[DIM_BAND]=1;
		energy.resize(energySize);
		
		ieee64 tmpAvgEnergy=0.0;
		if (energyData.size[DIM_BAND] == 2) {
			for (int32 i=0; i<energySize[DIM_X]; i++) {
				int32 v=(int32)((energyData.energy(i,0)+energyData.energy(i,1))/2.0);
				tmpAvgEnergy+=v;
				energy(i)=v;
			} // for
		} else {
			// analyse only the first channel
			for (int32 i=0; i<energySize[DIM_X]; i++) {
				int32 v=energyData.energy(i,0);
				tmpAvgEnergy+=v;
				energy(i)=v;
			} // for
		} // else
		tmpAvgEnergy/=energySize[DIM_X];
		avgEnergy=(int32)tmpAvgEnergy;

		// smooth energy data
		GaussFilter<int32> gauss(5);
		gauss.process(energy);
	
		// get average energy of lower energy values
		int32 n=0;
		tmpAvgEnergy=0.0;
		for (int32 i=0; i<energySize[DIM_X]; i++) {
			if (energy(i)<avgEnergy) {
				tmpAvgEnergy+=energy(i);
				n++;
			} // if
		} // for
		if (n==0) n=1;
		lowEnergyTh0 = (int32)tmpAvgEnergy/n;
		lowEnergyTh1 = lowEnergyTh0 / 2;
		lowEnergyTh2 = lowEnergyTh0 / 4;
	}

	/*
	// print values
	cout << "start-energy" << endl;
	for (int32 i=0; i<energySize[DIM_X]; i++) {
		cout << (int32)energy(i) << endl;
	} // for
	cout << "end-energy" << endl;
	exit(1);
	*/


	// repeat, until data.numberOfCuts < minimumNumberOfCuts 
	int16 thSilent = 55;
	do {
		// set variables
		data.version = AUDIO_CUT_VERSION;
		data.numberOfCuts   = 0;
		
		// set values to data
		data.blockLengthMilliSeconds = blockLengthMilliSeconds;
		data.filename = pathFilename;
		data.numberOfBlocks = energySize[DIM_X];
		data.numberOfCuts = 0;
		data.cuts.resize(energySize);
		for (int32 i=0; i<energySize[DIM_X]; i++) {
			data.cuts(i) = 0;
		} // for
		

		// initialize array
		SignalSize tmpCutsSize = energy.getSize();
		tmpCutsSize[DIM_BAND]=2;
		Signal<int16> tmpCuts (tmpCutsSize);
		//const int32 thAudio = 20;
		for (int32 i=0; i<tmpCutsSize[DIM_X]; i++) {
			tmpCuts(i,0)=0;
			
			// possible candidate for audio cut (energy is lower than average energy value)
			if (energy(i)<avgEnergy) {
				
				int32 totalSilence=0;
				if (energy(i) < avgEnergy) {
					totalSilence++;
					if (energy(i) < lowEnergyTh0) totalSilence++;
					if (energy(i) < lowEnergyTh1) totalSilence+=2;
					if (energy(i) < lowEnergyTh2) totalSilence+=4;
					if (i>0) totalSilence+=tmpCuts(i-1,0);
				} // if
				
				if (totalSilence > 32000) totalSilence = 32000;
				if (totalSilence < 0) totalSilence = 0;
				tmpCuts(i,0)=totalSilence;
			} // if 
		} // for
		for (int32 i=tmpCutsSize[DIM_X]-1; i>=0; i--) {
			tmpCuts(i,1)=0;
			
			// possible candidate for audio cut (energy is lower than average energy value)
			if (energy(i)<avgEnergy) {
				
				int32 totalSilence=0;
				if (energy(i) < avgEnergy) {
					totalSilence++;
					if (energy(i) < lowEnergyTh0) totalSilence++;
					if (energy(i) < lowEnergyTh1) totalSilence+=2;
					if (energy(i) < lowEnergyTh2) totalSilence+=4;
					if (i<tmpCutsSize[DIM_X]-1) totalSilence+=tmpCuts(i+1,1);
				} // if
				if (totalSilence > 32000) totalSilence = 32000;
				if (totalSilence < 0) totalSilence = 0;
				tmpCuts(i,1)=totalSilence;
			} // if 
			
			// get maximum of both peaks, store values in tmpCuts(i,0)
			if (tmpCuts(i,0) < tmpCuts(i,1)) {
				tmpCuts(i,0)=tmpCuts(i,1);
			} // if
		} // for
		
		// print values
		/*
		cout << "start" << endl;
		for (int32 i=0; i<tmpCutsSize[DIM_X]; i++) {
		cout << (int32)tmpCuts(i,0) << endl;
		} // for
		cout << "end" << endl;
		exit(1);
		*/
	
		
		// find silent areas, mark these areas (significance of silence)
		// read values from tmpCuts(i,0), store them in data.cuts(j)
		for (int32 i=0; i<tmpCutsSize[DIM_X]; i++) {
			tmpCuts(i,1)=0;
		} // for
		int32 nSilentAreas=0;
		for (int32 i=0; i<tmpCutsSize[DIM_X]; i++) {
		        //int32 tt0=tmpCuts(i,0);
			if (tmpCuts(i,0) > thSilent) {
				
				nSilentAreas++;
				
				// find minimum
				int32 center = i;
				while ((center<tmpCutsSize[DIM_X]-1) && (tmpCuts(center+1,0) <= tmpCuts(center,0))) center++;
				while ((center>0) && (tmpCuts(center-1,0) <= tmpCuts(center,0))) center--;
				
				// find left peak
				int32 left = center;
				while ((left>0) && (tmpCuts(left-1,0) >= tmpCuts(left,0))) left--;
				
				// find right peak
				int32 right = center;
				while ((right<tmpCutsSize[DIM_X]-1) && 
					((tmpCuts(right+1,0) >= tmpCuts(right,0)) || (tmpCuts(right+1,0) > thSilent))) right++;
				
				// modify values of tmpCuts(j,1)
				ieee64 lengthOfSilenceFactor = 1.0;
				if (right+1-left < samplesSilentLong)  lengthOfSilenceFactor = 0.8;
				if (right+1-left < samplesSilentShort) lengthOfSilenceFactor = 0.5;
				if (!(((left>0) && (tmpCuts(left-1,0) != 0)) || ((right<tmpCutsSize[DIM_X]-1) && (tmpCuts(right+1,0) != 0)))) {
					int32 tmpAvg1 = ((int32)tmpCuts(left,0) + tmpCuts(right,0)) / 2;
					// tmpAvg between 0..5000
					tmpAvg1 = (int32) (lengthOfSilenceFactor*tmpAvg1);
					if (tmpAvg1<0) tmpAvg1=0;
					if (tmpAvg1>5000) tmpAvg1=5000;
					
					// scale tmpAvg1 to 5..255, with special values 0..4
					// 0=no silence, 
					// 1=silent area 
					// 2=very silent area (silentArea1)
					// 3=most silent area (silentArea2)
					int32 tmpAvg2 = 0;
					tmpAvg2 = 5+tmpAvg1/4;
					if (tmpAvg1>100)  tmpAvg2 =  25+(tmpAvg1-100)/5;
					if (tmpAvg1>400)  tmpAvg2 =  85+(tmpAvg1-400)/6;
					if (tmpAvg1>1000) tmpAvg2 = 185+(tmpAvg1-1000)/40;
					if (tmpAvg1>3000) tmpAvg2 = 235+(tmpAvg1-3000)/100;
					if (tmpAvg1>5000) tmpAvg2 = 255;
					if (tmpAvg2<5) tmpAvg2=0;
					if (tmpAvg2>255) tmpAvg2=255;
					int16 centerValue = (uint8)tmpAvg2;
					
					int32 avgV=0;
					int32 minV=energy(left);
					int32 maxV=energy(left);
					for (int32 j=left; j<=right; j++) {
						avgV+=energy(j);
						if (minV>energy(j)) minV=energy(j);
						if (maxV<energy(j)) maxV=energy(j);
					} // for
					avgV/=(right-left+1);
					int32 silentArea1 = (avgV+maxV)/2;
					int32 silentArea2 = (avgV+minV)/2;
					
					// set values (silent area=1, very silent area=2, center of silence=centerValue)
					for (int32 j=left; j<=right; j++) {
						data.cuts(j) = 0;
						if (energy(j)<silentArea1) data.cuts(j) = 1;
						if (energy(j)<silentArea2) data.cuts(j) = 2;
					} // for
					data.cuts((left+right)/2) = centerValue;
					
					i=right+1;
				} // if
			} // if
		} // for
		
		
		  /*
		  // print values
		  cout << "silent areas" << nSilentAreas << endl;
		  cout << "start" << endl;
		  for (int32 i=0; i<tmpCutsSize[DIM_X]; i++) {
		  cout << (int32)data.cuts(i) << endl;
		  } // for
		  cout << "end" << endl;
		  exit(1);
		*/

		// count number of cuts
		data.numberOfCuts = 0;
		for (int32 i=0; i<energySize[DIM_X]; i++) {
			if (data.cuts(i) > 4) data.numberOfCuts++;
		} // for

		int32 increasedLowEnergy = (int32)(lowEnergyTh0*0.1);
		if (increasedLowEnergy<1) increasedLowEnergy=1;
		lowEnergyTh0 += increasedLowEnergy;
		lowEnergyTh1 = lowEnergyTh0 / 2;
		lowEnergyTh2 = lowEnergyTh0 / 4;

		int32 decreaseThSilent = (int32)(thSilent*0.1);
		if (decreaseThSilent<1) decreaseThSilent=1;
		thSilent-=decreaseThSilent;
		if (thSilent<=0) break;


	} while (data.numberOfCuts < minimumNumberOfCuts);
		
	return;	
} // readValues
