/*
    AudioEnergy.cpp     calculates the energy of an audio signal
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// MoCA includes
#include <AudioEnergy.h>


std::ostream& operator<< (ostream &os, const AudioEnergyData &data) 
{
	os << "version                    : " << data.version << endl;
	os << "filename                   : " << data.filename << endl;
	os << "blocklength in milliseconds: " << data.blockLengthMilliSeconds<< endl;
	os << "size[DIM_X]                : " << data.size[DIM_X]<< endl;
	os << "size[DIM_BAND]             : " << data.size[DIM_BAND]<< endl;
	os << endl;
	
	for (int32 i=0; i<data.energy.getSize()[DIM_X]; i++) {
		os << setfill(' ') << setw(5) << i << "    " << (int32)data.energy(i) << endl;
	} // for

	return os;
} // operator <<

OTagStream& operator<< (OTagStream& os, const AudioEnergyData &data) 
{
    TagFormat::writeStartTag (os, "AUDIO_ENERGY");
	os << endl;
	
    TagFormat::writeNodeValue(os, "version", data.version, 1);
    TagFormat::writeNodeValue(os, "filename", data.filename, 1);
    TagFormat::writeNodeValue(os, "blockLength", data.blockLengthMilliSeconds, 1);
    TagFormat::writeNodeValue(os, "size[DIM_X]", data.size[DIM_X], 1);
    TagFormat::writeNodeValue(os, "size[DIM_BAND]", data.size[DIM_BAND], 1);
	os << endl;

	TagFormat::writeStartTag (os, "ENERGY", 1);
	os << endl;

	for (int32 i=0; i<data.size[DIM_X]; i++) {
		os << setfill(' ') << setw(6) << i << "  ";
		for (int32 b=0; b<data.size[DIM_BAND]; b++) {
			os << setfill(' ') << setw(5) << data.energy(i,b) << "  ";
		} // for
		os << endl;;
	} // for
	
	TagFormat::writeEndTag (os, "ENERGY", 1);
	os << endl;

	
    TagFormat::writeEndTag (os, "AUDIO_ENERGY");
	os << endl;
	
    return os;
} // operator<<


ITagStream& operator>> (ITagStream& is, AudioEnergyData& data)
{
	
    char c;
    bool failFlag = true;
	
	if (is >> c) {
		failFlag = false;
		is.putback(c);

		TagFormat::readStartTag (is, "AUDIO_ENERGY");
		
		TagFormat::readNodeValue(is, "version", data.version);
		TagFormat::readNodeValue(is, "filename", data.filename);
		TagFormat::readNodeValue(is, "blockLength", data.blockLengthMilliSeconds);
		TagFormat::readNodeValue(is, "size[DIM_X]", data.size[DIM_X]);
		TagFormat::readNodeValue(is, "size[DIM_BAND]", data.size[DIM_BAND]);
		data.size[DIM_Y] = 1;
		data.size[DIM_Z] = 1;

		if (is.rdstate() & ios::failbit) {
			throw MocaException ("AudioEnergy: Illegal index file format!");
		} // if

		// initialize size of cuts
		data.energy.resize(data.size);
		for (int32 b=0; b<data.size[DIM_BAND]; b++) {
			for (int32 i=0; i<data.size[DIM_X]; i++) {
				data.energy(i,b) = 0;
			} // for
		} // for

		// read cuts
		if (is >> c) {
			is.putback(c);

			// read values
			TagFormat::readStartTag(is, "ENERGY");
			int32 f; // frame number
			int32 v; // value
			for (int32 i=0; i<data.size[DIM_X]; i++) {
				is >> f;
				for (int32 b=0; b<data.size[DIM_BAND]; b++) {
					is >> v;
					data.energy(i,b) = v;
				} // for
			} // for
			TagFormat::readEndTag(is, "ENERGY");

			if (is.rdstate() & ios::failbit) {
				throw MocaException ("AudioEnergy: Illegal index file format!");
			} // if
		} // if
		
		TagFormat::readEndTag (is, "AUDIO_ENERGY");
	} // if
	
	if (is.rdstate() & ios::failbit) {
		failFlag = true;
	} // if
    if (failFlag) {
		throw MocaException ("AudioEnergy: Illegal index file format!");
	} else {
        is.clear();
    } // else

    return is;
} // operator>> 





void 
AudioEnergy::process (AudioEnergyData &data)
{
	// read data
	ReadMocaData<AudioEnergyData> reader (pathFilename, AUDIO_ENERGY_FILENAME);
	bool dataValid = reader.process (AUDIO_ENERGY_VERSION, data); 
	
	if (!dataValid) {
		
		// create new index
		cerr << "\rAudioEnergy: Creating index file ...          " << endl;
		readValues(data);
		
		// save index file
		WriteMocaData<AudioEnergyData> writer (pathFilename, AUDIO_ENERGY_FILENAME);
		writer.process(data);
		cerr << "\rAudioEnergy: Creating index file ... done          " << endl;
	} // if

	return;
} // process




void 
AudioEnergy::readValues(AudioEnergyData &data)
{
	data.blockLengthMilliSeconds = blockLengthMilliSeconds;
	data.filename    = pathFilename;

	
	// open reader
	WaveHeader audioHeader;
	IWaveAudio reader (pathFilename);
	audioHeader = reader.getHeader();

	// get the number of samples in one block (within "blocklength" milliseconds)
	const int32 numberOfSamples = (int32) ((audioHeader.getSamplesPerSec() /1000.0) * data.blockLengthMilliSeconds); 
	data.size = reader.getSize();
	data.size[DIM_X]=(data.size[DIM_X]+numberOfSamples-1)/numberOfSamples;
	data.energy.resize(data.size);


	// find constant values at begin and end of the wav file (do they exist?)
	int32 firstConstantSampleEnd =-1;
	int32 lastConstantSampleBegin=-1;
	int32 firstValue = 0;
	int32 lastValue  = 0;
	bool  firstOtherValueFound = false;
	//bool  lastValueFound  = true;
	Signal<int16> rawData;
	int32 totalSamples=0;
	for (int32 i=0; i<data.size[DIM_X]; i++) {
		reader.read (rawData, numberOfSamples);
		
		// initialize values
		if (firstConstantSampleEnd==-1) {
			firstValue=(int32)((((ieee64)rawData(0,0))+rawData(0,1))/2.0);
			lastValue=firstValue;
			firstConstantSampleEnd =0;
			lastConstantSampleBegin=0;
		} // if
		for (int32 j=0; j<rawData.getSize()[DIM_X]; j++) {
			
			int32 v = (int32)((((ieee64)rawData(j,0))+rawData(j,1))/2.0);
			
			// find constant values at begin of wav file
			if (v == firstValue) {
				if (!firstOtherValueFound) {
					firstConstantSampleEnd=totalSamples+1;
				} // if
			} else {
				firstOtherValueFound=true;
			} // else
			
			// find constant values at begin of wav file
			if (v != lastValue) {
				lastValue = v;
				lastConstantSampleBegin=totalSamples;
			} // else
			
			totalSamples++;
		} // for
	} // for


	// analyse wav signal (get average values for each channel)
	reader.reset();
	totalSamples=0;
	ieee64 avgChannelLeft  =0.0;
	ieee64 avgChannelRight =0.0;
	for (int32 i=0; i<data.size[DIM_X]; i++) {
		reader.read (rawData, numberOfSamples);
		
		for (int32 j=0; j<rawData.getSize()[DIM_X]; j++) {
			
#if 0
			if ((totalSamples-2 == firstConstantSampleEnd) ||
				(totalSamples-1 == firstConstantSampleEnd) ||
				(totalSamples-0 == firstConstantSampleEnd) ||
				(totalSamples+1 == firstConstantSampleEnd) ||
				(totalSamples+2 == firstConstantSampleEnd) ||
				(totalSamples-2 == lastConstantSampleBegin) ||
				(totalSamples-1 == lastConstantSampleBegin) ||
				(totalSamples-0 == lastConstantSampleBegin) ||
				(totalSamples+1 == lastConstantSampleBegin) ||
				(totalSamples+2 == lastConstantSampleBegin)) {
			        //ieee64 v1=rawData(j,0);
			        //ieee64 v2=rawData(j,1);
				//ieee64 t=(int32)((v1+v2)/2.0);
			} // if
#endif
			
			if ((totalSamples >= firstConstantSampleEnd) && (totalSamples < lastConstantSampleBegin)) {
				avgChannelLeft  += rawData(j,0);
				avgChannelRight += rawData(j,1);
			} // if
			totalSamples++;
		} // for
	} // for
	if (totalSamples<1) totalSamples=1;
	avgChannelLeft  /= totalSamples;
	avgChannelRight /= totalSamples;
	
	// read energy
	reader.reset();
	totalSamples=0;
	ieee64 sampleLeft =0;
	ieee64 sampleRight=0;
	for (int32 i=0; i<data.size[DIM_X]; i++) {
		
		reader.read (rawData, numberOfSamples);
		
		ieee64 energyLeft =0.0;
		ieee64 energyRight=0.0;
		ieee64 samplesInBlock=0.0;
		for (int32 j=0; j<rawData.getSize()[DIM_X]; j++) {
			if ((totalSamples >= firstConstantSampleEnd) && (totalSamples < lastConstantSampleBegin)) {
				sampleLeft =rawData(j,0)-avgChannelLeft;
				sampleRight=rawData(j,1)-avgChannelRight;
				energyLeft  += sampleLeft*sampleLeft;
				energyRight += sampleRight*sampleRight;
			}// if
			samplesInBlock++;
			totalSamples++;
		} // for
		if (energyLeft<0) {
			energyLeft=0;
			cerr << "AudioEnergy: overflow error." << endl;
		} // if
		if (energyRight<0) {
			energyRight=0;
			cerr << "AudioEnergy: overflow error." << endl;
		} // if
		
		// normalize by number of samples in block
		if (samplesInBlock<1.0) samplesInBlock=1.0;
		if (energyLeft<0.0)  energyLeft=0.0;
		if (energyRight<0.0) energyRight=0.0;

		data.energy(i,0)=(int32)sqrt(energyLeft/samplesInBlock);
		data.energy(i,1)=(int32)sqrt(energyRight/samplesInBlock);
	} // for

	return;
} // process


