/*
    CameraModel.cpp     calculate the camera model
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// MoCA includes
#include <CameraModel.h>

// TODO:
//    - what happens, if no corners could be detected
//    - what happens, if the parameters of the model are incorrect? (use another model?)
//    - ...

std::ostream& operator<< (ostream& os, const CameraModelItem &item)
{	
	os << "frame: " << item.frame << "   valid: " << item.valid << 
		  "   number of matches: " << item.numberCornerMatchesAvailable << "   maxVectorLength: " << item.maxVectorLength << endl;
	os << "a[]: " << item.transData.a00 << "  " << item.transData.a01 << "  " 
	   << item.transData.a10 << "  " << item.transData.a11 << endl;
	os << "t[]: " << item.transData.tx << "  " << item.transData.ty << endl;
	os << "p[]: " << item.transData.px << "  " << item.transData.py << endl;
	os << endl;
	
	return os;
} // operator<<

OTagStream& operator<< (OTagStream& os, const CameraModelItem& item)
{
    TagFormat::writeNodeValue(os, "frame", item.frame);
    TagFormat::writeNodeValue(os, "valid", item.valid);
    TagFormat::writeNodeValue(os, "numberCornerMatchesAvailable", item.numberCornerMatchesAvailable);
    TagFormat::writeNodeValue(os, "maxVectorLength", item.maxVectorLength);
    TagFormat::writeNodeValue(os, "lastCorners", item.lastCorners);
    TagFormat::writeNodeValue(os, "currCorners", item.currCorners);
    TagFormat::writeNodeValue(os, "numberCornerMatched", item.numberCornerMatched);


	TagFormat::writeStartTag(os, "parameters");
	os << item.transData.a00 << " " << item.transData.a01 << " " << item.transData.a10 << " " << item.transData.a11 << " " 
	   << item.transData.tx  << " " << item.transData.ty  << " " << item.transData.px  << " " << item.transData.py << " ";
	TagFormat::writeEndTag(os, "parameters");

    return os;
} // operator<<

ITagStream& operator>> (ITagStream& is, CameraModelItem& item)
{
    char c;
    bool failFlag = false;
	
	if (is >> c) {
		is.putback(c);
		
		TagFormat::readNodeValue(is, "frame", item.frame);
		TagFormat::readNodeValue(is, "valid", item.valid);
		TagFormat::readNodeValue(is, "numberCornerMatchesAvailable", item.numberCornerMatchesAvailable);
		TagFormat::readNodeValue(is, "maxVectorLength", item.maxVectorLength);
		TagFormat::readNodeValue(is, "lastCorners", item.lastCorners);
		TagFormat::readNodeValue(is, "currCorners", item.currCorners);
		TagFormat::readNodeValue(is, "numberCornerMatched", item.numberCornerMatched);

		TagFormat::readStartTag(is, "parameters");
		is >> item.transData.a00;
		is >> item.transData.a01;
		is >> item.transData.a10;
		is >> item.transData.a11;
		is >> item.transData.tx;
		is >> item.transData.ty;
		is >> item.transData.px;
		is >> item.transData.py;
		TagFormat::readEndTag(is, "parameters");
	} // if
		
	if (is.rdstate() & ios::failbit) {
		failFlag = true;
	} // if
	if (failFlag) {
		throw MocaException ("CameraModel: Illegal file format!");
	} else {
		is.clear();
	} // else
	
	return is;
} // operator>> 




std::ostream& operator<< (ostream& os, const CameraModelData &cm)
{
	os << "version : " << cm.version << endl;
	os << "filename: " << cm.filename<< endl;
	
	int32 n = cm.data.size();
	for (int32 i=0; i<n; i++) {
		os << cm.data[i] << endl;
	} // for
	
	return os;
} // operator<<

OTagStream& operator<< (OTagStream& os, const CameraModelData& cm)
{
	TagFormat::writeNodeValue(os, "version", cm.version);
	TagFormat::writeNodeValue(os, "filename", cm.filename);
	os << endl;
	
	int32 n = cm.data.size();
	for (int32 i=0; i<n; i++) {
		os << cm.data[i] << endl;
	} // for

	return os;
} // operator<<

ITagStream& operator>> (ITagStream& is, CameraModelData& cm)
{
	char c;
	bool failFlag = false;
	
	if (is >> c) {
		is.putback(c);
		
		TagFormat::readNodeValue(is, "version", cm.version);
		TagFormat::readNodeValue(is, "filename", cm.filename);
		if (is.rdstate() & ios::failbit) {
			failFlag = true;
		} // if
		
		cm.data.clear();
		
		CameraModelItem item;
		while (is >> c) {
			is.putback(c);
			if (is >> item) {
				cm.data.push_back (item);
			} else {
				failFlag = true;
			} // else
		} // while
	} // if
	
	if (failFlag) {
		throw MocaException ("CameraModel: Illegal file format!");
	} else {
		is.clear();
	} // else
	
	return is;
} // operator>> 



CameraModel::CameraModel(const string &pathFilename, int32 cropleft, int32 cropright, int32 croptop, int32 cropbottom,
						 int32 defaultMaxMotionVectorLength, bool calcTranslationOnly) 
{
	processShot = false;
	defaultObjPos = POSITION_UNKNOWN;
	defaultPx0 = -1;
	defaultPx1 = -1;
	defaultPy0 = -1;
	defaultPy1 = -1;
	defaultMotionVectorLength = defaultMaxMotionVectorLength;
	useDefaultMotion = false;
	calcTranslation=calcTranslationOnly;


	this->pathFilename = pathFilename;
	path = "";
	{
		CreateMocaPath mocaPath (pathFilename, false);
		path = mocaPath.path;
	}

	this->cropleft   = cropleft;
	this->cropright  = cropright;
	this->croptop    = croptop;
	this->cropbottom = cropbottom;


//	cmd = LINUX_CAMERA_MODEL_EXECUTABLE;
#ifdef _WIN32
//	cmd = WINDOWS_CAMERA_MODEL_EXECUTABLE;
#endif

	// get video parameters
	VideoDecoder decoder (this->pathFilename);
	size = decoder.getSize();
	//MpegIndex indexReader(this->pathFilename);
	//indexReader.process(index);	
} 



void 
CameraModel::process(CameraModelData& data) 
{
	if (processShot) {
		cerr << "error in CameraModel:process: constructor and function do not match!" << endl;
		exit(1);
	} // if

	// read data
	ReadMocaData<CameraModelData> reader (pathFilename, CAMERA_MODEL_FILENAME);
	bool dataValid = reader.process (CAMERA_MODEL_VERSION, data); 


	if (dataValid) {
		int32 vidFrames  = size[DIM_Z];
		int32 dataLength = (int32)data.data.size();
		if ((dataLength != vidFrames) && (dataLength+1 != vidFrames)) {
			cerr << "CameraModel::process: wrong number of frames in index file" << endl;
			cerr << "data size : " << data.data.size() << endl;
			cerr << "video size: " << size[DIM_Z] << endl;
			dataValid=false;
		} // if
	} // if

	if (!dataValid) {

		// create new index
		cerr << "CameraModel: Creating index file ..." << endl;
		readValues(data);
			
		// save index file
		WriteMocaData<CameraModelData> writer (pathFilename, CAMERA_MODEL_FILENAME);
		writer.process(data);
		cerr << "\rCameraModel: Creating index file ...done                         " << endl;
	} // if

	return;
} // process




// calculate camera model parameters of two frames (based on the corners of these two frames)
void 
CameraModel::getCameraModelItem (const vector<CornerPosition> &currcorners, vector<CornerPositionInt> &currcorners_int,
							     const vector<CornerPosition> &lastcorners, const vector<CornerPositionInt> &lastcorners_int,
								 const Image<uint8> &curr, const Image<uint8> &last, 
							     CameraModelItem &item)
{
	int32 nlast = lastcorners_int.size();
	int32 ncurr = currcorners_int.size();
	item.clear();
	item.lastCorners=nlast;
	item.currCorners=ncurr;

	if ((nlast<4) || (ncurr<4)) {
		// at least 4 corners are necesarry to calculate camera model
		item.valid = 2;
	} else {
		
		// match corners
		matchCorners* matches = new matchCorners[nlast*ncurr];
		int* currcorr = new int[ncurr];
		int* lastcorr = new int[nlast];
		for (int i=0;i<ncurr;i++) currcorr[i]=-1;
		for (int i=0;i<nlast;i++) lastcorr[i]=-1;
		
		int i=0;
		int ncorr=0;
		for (int32 l = 0;l < (int32)lastcorners_int.size();l++) {
			for (int32 c=0;c<(int32)currcorners_int.size();c++) {
				int cx=currcorners_int[c].x;
				int lx=lastcorners_int[l].x;
				int cy=currcorners_int[c].y;
				int ly=lastcorners_int[l].y;
				
				int dx=cx-lx, dy=cy-ly;
				
				const int maxdist = defaultMotionVectorLength*defaultMotionVectorLength;
				if (dx*dx+dy*dy > maxdist) continue; // matches[i].score *= 5;
				
				matches[i].last = l;
				matches[i].curr = c;
				
				//SAD2_MMX::process (curr, last, cx-7, cy-7, lx-7, ly-7, matches[i].score);
				const uint8 * imgPtr1 = curr.getSamplePtr(cx-7, cy-7);
				const uint8 * imgPtr2 = last.getSamplePtr(lx-7, ly-7);
				long int yOffset1 = curr.getSampleOffsetY(cy-7);
				long int yOffset2 = last.getSampleOffsetY(cy-7);
				matches[i].score = processSAD2 (imgPtr1, imgPtr2, yOffset1, yOffset2);
				
				// keep match, if distance is not too large
				if (matches[i].score<5000) {
#if 0
					cerr << matches[i].score << endl;
#endif
					i++;
					ncorr++;
				} // if
			} // for
		} // for
		item.numberCornerMatchesAvailable = ncorr;
		qsort(matches, ncorr, sizeof(struct matchCorners), compMatchCorners);
		
		int32 thScore = 1000 + 4 * matches[0].score;
		if (thScore > 5000) thScore=5000;

		int found=0;
		i=0;
		while (found < min(ncurr,nlast) && i<ncorr) {
			if (found>20) thScore-=20;

			if (lastcorr[matches[i].last]<0 &&
				currcorr[matches[i].curr]<0 && matches[i].score < thScore) {
				lastcorr[matches[i].last] = matches[i].curr;
				currcorr[matches[i].curr] = matches[i].last;

#if 0
				CornerPositionInt tmpLast = lastcorners_int[matches[i].last];
				CornerPositionInt tmpCurr = currcorners_int[matches[i].curr];
				cerr << tmpCurr.x << "/" << tmpCurr.y  << "    \t" << tmpLast.x << "/" << tmpLast.y  << endl;
#endif
				found++;
			}
			i++;
		} // while
		item.numberCornerMatched = found;

	
		
		
#ifdef DISPLAY_RESULTS1
#if 0
		cerr << "debug" << endl;
		disp = curr;
		DrawCross cross (10, Color::WHITE);
		for (int32 i=0;i<currcorners_int.size();i++)
		{
			int32 x = currcorners_int[i].x;
			int32 y = currcorners_int[i].y;
			
			cross.process (ImagePosition (x,y), disp);
		}
		stringstream s;
		s << "F:/corners-neu" << setfill('0') << setw(2) << frameNo << ".pgm";
		string s1 = s.str();
		WriteImage(s1, disp);
#endif
#endif 	  
		
		
		// estimate motion model
		int nequs = 0;
		for (int c=0;c<ncurr;c++) 
		{
			if (currcorr[c]>=0) {
				nequs += 2;
			} // if
		} // for
		
		if (nequs == 0) {
			item.valid = 3;
		} else {
			
			// set valid
			item.valid = 0;
			
			//data.Create(9,nequs);
			vector<double> parameters;
			double **data = NULL; 
			data = new double* [nequs];
			for (int32 ii=0; ii<nequs; ii++) {
				data[ii] = new double[9];
			} // for
			
			vector<bool> outlier;
			vector<double> residual;
			outlier.resize (nequs);
			residual.resize (nequs);
			
			LinearEstimatorFeature lts;
			lts.SetOutlierInfo(&outlier);
			lts.SetResidualInfo(&residual);
			lts.SetNHTries(300);
			
			int equ=0;
			for (int c=0;c<ncurr;c++) {
				if (currcorr[c]>=0) {
					CornerPosition lastp,currp;
					currp = currcorners[c];
					lastp = lastcorners[currcorr[c]];
					
					// perspective
					data[equ][0] = lastp.x;
					data[equ][1] = lastp.y;
					data[equ][2] = 1;
					data[equ][3] = 0;
					data[equ][4] = 0;
					data[equ][5] = 0;
					data[equ][6] = -currp.x*lastp.x;
					data[equ][7] = -currp.x*lastp.y;
					data[equ][8] = currp.x;
					equ++;

					data[equ][0] = 0;
					data[equ][1] = 0;
					data[equ][2] = 0;
					data[equ][3] = lastp.x;
					data[equ][4] = lastp.y;
					data[equ][5] = 1;
					data[equ][6] = -currp.y*lastp.x;
					data[equ][7] = -currp.y*lastp.y;
					data[equ][8] = currp.y;
					equ++;
				}
			} // for
			
			//Assert(equ == nequs);
			lts.Estimate(parameters,data, 9, nequs);
			
			// delete data
			for (int32 ii=0; ii<nequs; ii++) {
				delete[] data[ii];
			} // for
			delete[] data;
			
			
			item.transData.a00 = parameters[0];
			item.transData.a01 = parameters[1];
			item.transData.a10 = parameters[3];
			item.transData.a11 = parameters[4];
			item.transData.tx  = parameters[2];
			item.transData.ty  = parameters[5];
			item.transData.px  = parameters[6];
			item.transData.py  = parameters[7];
			
#if 0
			cout << parameters[0] << " "
				<< parameters[1] << " "
				<< parameters[2] << " "
				<< parameters[3] << " "
				<< parameters[4] << " "
				<< parameters[5] << " "
				<< parameters[6] << " "
				<< parameters[7] << endl;
#endif
			
		} // else
				
		delete[] matches;
		delete[] currcorr;
		delete[] lastcorr;
	} // else
	item.getMaxVectorLength (size[DIM_X], size[DIM_Y], item.maxVectorLength);


	return;
}


// calculate camera parameters for two frames
void 
CameraModel::processTwoFrames (Image<uint8> &curr, Image<uint8> &last, CameraModelItem& item) 
{
	if (!processShot) {
		cerr << "error in CameraModel::processTwoFrames: constructor and function do not match!" << endl;
		exit(1);
	} // if

	ImageSize size = curr.getSize();
	if (size[DIM_BAND] != 1) {
		GrayscaleFilter<uint8> gray;
		gray.process(curr);
		gray.process(last);
		size = curr.getSize();
	} // if


	// normalize luminance?
	if (normalize) {
		ImageStatisticData dCurr, dLast;
		ImageStatistic stat;
		stat.processAvg(curr, dCurr);
		stat.processAvg(last, dLast);

		if (ABS(dCurr.avg[0]-128) > ABS(dLast.avg[0]-128)) {

			// modify dCurr
			int32 d = dLast.avg[0]-dCurr.avg[0];
			for (int32 y=0; y<size[DIM_Y]; y++) {
				for (int32 x=0; x<size[DIM_X]; x++) {
					int32 v = curr(x,y) + d;
					if (v<0) v=0;
					if (v>255) v=255;
					curr(x,y) = (uint8) v;
				} // for
			} // for
		} else {
			// modify dLast
			int32 d = dCurr.avg[0]-dLast.avg[0];
			for (int32 y=0; y<size[DIM_Y]; y++) {
				for (int32 x=0; x<size[DIM_X]; x++) {
					int32 v = last(x,y) + d;
					if (v<0) v=0;
					if (v>255) v=255;
					last(x,y) = (uint8) v;
				} // for
			} // for
		} // else
	} // if


	// corner detection
	vector<CornerPosition> lastcorners,currcorners;
	vector<CornerPositionInt> lastcorners_int,currcorners_int;
	getCorners (curr, currcorners, currcorners_int);
	getCorners (last, lastcorners, lastcorners_int);

	// initialize camera model
	item.frame = 0;
	item.numberCornerMatchesAvailable  = 0;
	item.maxVectorLength  = 0;
	item.valid = 1;

	// get parameters of camera model
	getCameraModelItem (currcorners, currcorners_int, lastcorners, lastcorners_int, curr, last, item);

	return;
}

void 
CameraModel::calculateTranslation (Image<uint8> &refImg, Image<uint8> &img, ieee64 &tx, ieee64 &ty)
{
	ImageSize s=refImg.getSize();
	int range=min(s[DIM_X],s[DIM_Y]) /4;
	if (range>32) range=32;
	if (range>defaultMotionVectorLength) range=defaultMotionVectorLength;


	tx=ty=0;
	BlockMotionFeature bmf (8, range);

	Image<int8> motionvectors;
	bmf.processFullSearch (refImg, img, tx,ty);

	return;
}


void 
CameraModel::calculateTranslationParameters (CameraModelData& cm) 
{
#if 0
	// get standard camera parameters
	calculateCameraParameters (cm);
#endif

	VideoDecoder reader (pathFilename, CODEC_Y);
	int32 numberOfFrames = reader.getSize()[DIM_Z];
	ImageSize size = reader.getImageSize();

	// initialize camera parameters
	cm.data.resize(numberOfFrames);

	// initialize features
	CropFilter<uint8> crop (size, cropleft, cropright, croptop, cropbottom);
	CameraTransformationPrecise ctp (CAMERA_TRANSFORM_TRANSLATORIAL);
	ClipFilter<uint8> clip (0,254);

	// initialize mask
	Image<uint8> mask;
	mask = reader.getImage(0);
	crop.process (mask);
	mask.clear(0);
	size = mask.getSize();

	cerr << "calculate motion" << endl; 
	Image<uint8> refImg, img;
	for (int32 frameNo=1; frameNo<numberOfFrames; frameNo++) {

		cerr << "\rframe: " << frameNo-1 << "   " << ends; 

		// read first image
		refImg = reader.getImage(frameNo-1);
		crop.process (refImg);
		clip.process (refImg);

		// read second image
		img = reader.getImage(frameNo);
		crop.process (img);
		clip.process (img);
		

		CameraModelItem item = cm.data[frameNo-1];
	
		// get transformation without previous transformation data
		ieee64 minErr0=0;
		TransformationData srcDestCameraParameters0;
#if 0
		ctp.process (refImg, img, mask, srcDestCameraParameters0, minErr0);
cout << frameNo-1 << "\t" << srcDestCameraParameters0.tx << "\t " << srcDestCameraParameters0.ty << endl;
cerr << frameNo-1 << "\t" << srcDestCameraParameters0.tx << "\t " << srcDestCameraParameters0.ty << endl;
continue;
#endif




		// get transformation with previous transformation data
		ieee64 minErr1=0;
		TransformationData srcDestCameraParameters1;
		srcDestCameraParameters1.tx = item.transData.tx;
		srcDestCameraParameters1.ty = item.transData.ty;
#if 0
		ctp.process (refImg, img, mask, srcDestCameraParameters1, minErr1);
cout << frameNo-1 << "\t" << srcDestCameraParameters1.tx << "\t " << srcDestCameraParameters1.ty << endl;
cerr << frameNo-1 << "\t" << srcDestCameraParameters1.tx << "\t " << srcDestCameraParameters1.ty << endl;
continue;
#endif

		// estimate pure translation
		ieee64 tx=0;
		ieee64 ty=0;
		calculateTranslation (refImg, img, tx, ty);
		ieee64 minErr2=0;
		TransformationData srcDestCameraParameters2;
		srcDestCameraParameters2.tx = -tx;
		srcDestCameraParameters2.ty = -ty;


#if 0
cout << frameNo-1 << "\t" << srcDestCameraParameters2.tx << "\t " << srcDestCameraParameters2.ty << endl;
cerr << frameNo-1 << "\t" << srcDestCameraParameters2.tx << "\t " << srcDestCameraParameters2.ty << endl;
continue;
#endif


#if 0
		ctp.process (refImg, img, mask, srcDestCameraParameters2, minErr2);
cout << frameNo-1 << "\t" << srcDestCameraParameters2.tx << "\t " << srcDestCameraParameters2.ty << endl;
cerr << frameNo-1 << "\t" << srcDestCameraParameters2.tx << "\t " << srcDestCameraParameters2.ty << endl;
continue;
#endif


#if 0

		if (minErr0 > minErr1) {
			minErr0 = minErr1;
			srcDestCameraParameters0 = srcDestCameraParameters1;
		} // if
		if (minErr0 > minErr2) {
			minErr0 = minErr2;
			srcDestCameraParameters0 = srcDestCameraParameters2;
		} // if
#endif

		
		// copy values
		srcDestCameraParameters0 = srcDestCameraParameters2;
		item.transData = srcDestCameraParameters0;
		item.frame = frameNo-1;
		item.numberCornerMatchesAvailable  = 100;
		item.valid = 0;
		ieee64 length=0;
		item.getMaxVectorLength (size[DIM_X], size[DIM_Y], length);
		item.maxVectorLength = (int32)length;
		cm.data[frameNo-1] = item;

#if 0
		if (item.valid == 0) {
			if (minErr0 > 20*20) {
				item.valid = 1;
				cerr << "frame: " << frameNo-1 << " (large difference) " << endl;
			} // if
		} // if
#endif
	} // while

	// validate parameters
	validateCameraParametersTranslation (cm);

	return;
}


void 
CameraModel::calculateCameraParameters (CameraModelData& cm) 
{

	// initialize parameters
	Image<uint8> last, curr;

#ifdef DISPLAY_RESULTS
	Image<uint8> disp;
#endif	

	// open reader
	VideoDecoder reader (pathFilename, CODEC_Y);
	ImageSize size = reader.getImageSize();
	int32 numberOfFrames = reader.getSize()[DIM_Z];
	
	// initialize features
	CropFilter<uint8> crop (size, cropleft, cropright, croptop, cropbottom);
	int32 w = size[DIM_X]-cm.cropleft-cm.cropright;
	int32 h = size[DIM_Y]-cm.croptop-cm.cropbottom;
	
	
	try
	{
		// read first image
		curr = reader.getImage(0);
		crop.process (curr);
		
		// corner detection
		vector<CornerPosition> lastcorners,currcorners;
		vector<CornerPositionInt> lastcorners_int,currcorners_int;
		getCorners (curr, currcorners, currcorners_int);

		
		for (int32 frameNo=1; frameNo<numberOfFrames; frameNo++) {
			
			cerr << "\rframe: " << (frameNo-1) << "     " << ends ;

#if 0
			// debug
			cout << "\rframe: " << item.frame << "     " << endl;
#endif
			
			last = curr;
			lastcorners = currcorners;
			lastcorners_int = currcorners_int;
			
			// read next image
			curr = reader.getImage(frameNo);
			crop.process (curr);
			//clip.process (curr);


			// calculate camera parameters of next two frames 
			CameraModelItem item;
			getCorners (curr, currcorners, currcorners_int);
			getCameraModelItem (currcorners, currcorners_int, lastcorners, lastcorners_int, curr, last, item);
			item.frame = frameNo-1;

			cm.data.push_back(item);
		} // while
	} catch (...) {
		cerr << "CameraModel::startCalculation: error" << endl;
	} // catch
	cerr << endl;


	// validate parameters
	validateCameraParameters (cm);

	return;
} // calculateCameraParameters


void 
CameraModel::getCorners (const Image<uint8> &img, vector<CornerPosition>& cornersFloat, vector<CornerPositionInt>& cornersInt) 
{
	int32 nCorners = 0;
	int32 cornerTh = 100;
	int32 it=0;
	int32 w = img.getSize()[DIM_X];
	int32 h = img.getSize()[DIM_Y];
	CornerDetection cornerDetect (cornerTh, true);

	while (nCorners<50 || nCorners>200) {
		it++;
		if (it>8) break;

		cornersFloat.clear();
		cornerDetect.setThreshold(cornerTh);
		cornerDetect.process (img, cornersFloat);
		cornerDetect.filterCorners (cornersFloat,7,7,w-1-8,h-1-8);
		if (defaultObjPos != POSITION_UNKNOWN) {
			cornerDetect.removeCornersRect (cornersFloat, defaultPx0,defaultPy0,defaultPx1,defaultPy1);
		} // if

		nCorners = cornersFloat.size();
		if (nCorners <100) {
			cornerTh -=5;
		} else {
			cornerTh +=5;
		} // if
	} // while

	cornerDetect.cornersToInt (cornersFloat, cornersInt);
	return;
}



void 
CameraModel::processOneShot (const Video<uint8> &vid, CameraModelData& cm) 
{
	if (!processShot) {
		cerr << "error in CameraModel:processOneShot: constructor and function do not match!" << endl;
		exit(1);
	} // if

	if (defaultObjPos != POSITION_UNKNOWN) {
		int32 w = vid.getSize()[DIM_X]; 
		int32 h = vid.getSize()[DIM_Y]; 

		switch (defaultObjPos) {
		case POSITION_CENTER:
			defaultPx0 = w/4;
			defaultPx1 = (3*w)/4;
			defaultPy0 = h/4;
			defaultPy1 = (3*h)/4;
			break;
		case POSITION_LEFT:
			defaultPx0 = 0;
			defaultPx1 = w/4;
			defaultPy0 = 0;
			defaultPy1 = h;
			break;
		case POSITION_RIGHT:
			defaultPx0 = (3*w)/4;
			defaultPx1 = w;
			defaultPy0 = 0;
			defaultPy1 = h;
			break;
		case POSITION_TOP:
			defaultPx0 = 0;
			defaultPx1 = w;
			defaultPy0 = 0;
			defaultPy1 = h/4;
			break;
		case POSITION_BOTTOM:
			defaultPx0 = 0;
			defaultPx1 = w;
			defaultPy0 = (3*h)/4;
			defaultPy1 = h;
			break;
		default:
			cerr << "error in CameraModel:processOneShot: unknown default image position!" << endl;
			break;
		} // switch
	} // if

	// initialize parameters
	Image<uint8> last, curr;

#ifdef DISPLAY_RESULTS
	Image<uint8> disp;
#endif	

	// open reader
	size = vid.getSize();
	ImageSize imgSize (size[DIM_X], size[DIM_Y], size[DIM_BAND]);
	int32 numberOfFrames = size[DIM_Z];
	
	// initialize features
	CropFilter<uint8> crop (imgSize, cropleft, cropright, croptop, cropbottom);
	int32 w = imgSize[DIM_X]-cm.cropleft-cm.cropright;
	int32 h = imgSize[DIM_Y]-cm.croptop-cm.cropbottom;
	
	
	try
	{
		// read first image
		curr = vid.getImage(0);
		crop.process (curr);
		
		// corner detection
		vector<CornerPosition> lastcorners,currcorners;
		vector<CornerPositionInt> lastcorners_int,currcorners_int;
		getCorners (curr, currcorners, currcorners_int);

#if 0
		Image<uint8> tmp = curr;
		CornerDetection c;
		c.drawCorners (tmp, currcorners_int);
		WriteImage(tmp, "D:/out.png");		
#endif
		
		for (int32 frameNo=1; frameNo<numberOfFrames; frameNo++) {
			
			CameraModelItem item;
			item.frame = frameNo-1;
			cerr << "\rframe: " << item.frame << "     " << ends ;

#if 0
			// debug
			cout << "\rframe: " << item.frame << "     " << endl;
#endif
			
			last = curr;
			lastcorners = currcorners;
			lastcorners_int = currcorners_int;
			
			// read next image
			curr = vid.getImage(frameNo);
			crop.process (curr);
			//clip.process (curr);


			// calculate camera parameters of next two frames 
			getCorners (curr, currcorners, currcorners_int);
			getCameraModelItem (currcorners, currcorners_int, lastcorners, lastcorners_int, curr, last, item);
			item.frame = frameNo-1;

#if 1
			cerr << item << endl;
#endif

#if 0
			{
				cerr << "debug:" << endl;
				CornerDetection cornerDetect (true);
				string fn = FileHandling::getNumberedFilename ("D:/debug/corners__", frameNo, ".png");
				cerr << "store image: " << fn << endl;
				Image<uint8> tmp = curr;
				cornerDetect.drawCorners (tmp, currcorners);
				WriteImage (tmp, fn);
			}
#endif

#if 0
			{
				cerr << "debug:" << endl;
				Image<uint8> tmp = curr;
				CornerDetection cd;
				cd.drawCorners (tmp, currcorners);
				WriteImage(tmp, "D:/debug/curr.png");
				tmp = last;
				cd.drawCorners (tmp, lastcorners);
				WriteImage(tmp, "D:/debug/last.png");
			}
#endif
			

#if 1
			{
				cerr << "debug:" << endl;
				string fn = FileHandling::getNumberedFilename ("D:/debug/mv__", frameNo, ".png");
				cerr << "store image: " << fn << endl;
				Image<uint8> tmp = curr;
				item.drawVectorImage (tmp);
				WriteImage (tmp, fn);
			}
#endif

			cm.data.push_back(item);
		} // while
	} catch (...) {
		cerr << "CameraModel::startCalculation: error" << endl;
	} // catch
	cerr << endl;


	// validate parameters
	validateCameraParameters (cm);

	return;
} // processVideo




void 
CameraModel::validateCameraParameters (CameraModelData& cm) 
{

	// check number of entries
	if ((int32)cm.data.size() != size[DIM_Z]) {
		CameraModelItem item;
		//cerr << "add camera model items (" << (size[DIM_Z] - cm.data.size()) << ")" << endl;
		while ((int32)cm.data.size() < size[DIM_Z]) {
			item.frame = cm.data.size();
			item.valid=0;
			cm.data.push_back(item);
		} // while
	} // if
	if ((int32)cm.data.size() != size[DIM_Z]) {
		cerr << "CameraModel::calculateCameraParameters: wrong number of frames" << endl;
		cerr << "data size : " << cm.data.size() << endl;
		cerr << "video size: " << size[DIM_Z] << endl;
		exit(1);
	} // if


	// compare frame number with index item (TODO: what happens, if they do not match???)
	for (int32 i=0; i<(int32)cm.data.size(); i++) {
		CameraModelItem item;
		item = cm.data[i];
		if (i != cm.data[i].frame) {
			cerr << "CameraModel::parameters of frame " << i << " not found." << endl;
		} // if
	} // for


#if 0
	// store images
	VideoDecoder reader(pathFilename);
	DrawLine<uint8> line;	
	Color c = Color::YELLOW;
	Image<uint8> img;
	for (int32 j=0; j<140; j++) {
		img = reader.getImage (j);
		stringstream oss;
		oss << "V:/temp/result/img_test_" << setw(6) << setfill('0') << (j) << ".jpg";
		cout << "save image: " << oss.str() << endl;
		OImageStream<uint8> ois(oss.str());
		ois.write (img);
	}
	exit(1);
#endif



	// analyse camera model items, remove parameters, which do not make sense:
	//   --> a motion vector should not be larger than 40% of the image
	const ieee64 allowedVectorLength = 
		(0.4*size[DIM_X])*(0.4*size[DIM_X]) + 
		(0.4*size[DIM_Y])*(0.4*size[DIM_Y]);
	for (int32 i=0; i<(int32)cm.data.size(); i++) {
		
		CameraModelItem tmp = cm.data[i];
		if (cm.data[i].valid == 0) {


#if 0
			// store images
			cout << cm.data[i] << endl;
			
			// read image, make it darker
			img = reader.getImage (cm.data[i].frame);
			for (int32 y=0; y<size[DIM_Y]; y++) {
				for (int32 x=0; x<size[DIM_X]; x++) {
					uint8 p = img(x,y,0);
					if (p>64) p-=64; else p=0;
					img(x,y,0) = p;
					
					p = img(x,y,1);
					if (p>64) p-=64; else p=0;
					img(x,y,1) = p;
					
					p = img(x,y,2);
					if (p>64) p-=64; else p=0;
					img(x,y,2) = p;
				} // for
			} // for
#endif


			ieee64 maxVectorLength = -10e10;
			ieee64 nnx, nny;
			for (int32 y=8; y<size[DIM_Y]-8; y+=16) {
				for (int32 x=8; x<size[DIM_X]-8; x+=16) 
				{
					cm.data[i].getVectors(x,y,nnx,nny);
					ieee64 vectorLength = (x-nnx)*(x-nnx) + (y-nny)*(y-nny);
					if (maxVectorLength < vectorLength) maxVectorLength = vectorLength;
				} // for
			} // for
			

			// a motion vector should not be larger than half the image
			if (maxVectorLength > allowedVectorLength) {
				//cerr << "invalid frame number: " << cm.data[i].frame << endl;
				cm.data[i].valid = 4;  // 4=parameters calculated, but they do not make sense (e.g. large object visible)
			} // if
		} // if
	} // for


	// analyse camera model of two following frames: set (valid==5 or valid==6), if the vectors change much:
	//   --> get vectors of 2 frames
	//   --> diff = maximum of difference of 2 corresponding vectors
	//   --> if (diff>largeVectorDifference) ==> valid=5
	//   --> if (diff>mediumVectorDifference) ==> valid=6
	const ieee64 mediumVectorDifference = 
		(0.06*size[DIM_X])*(0.06*size[DIM_X]) + 
		(0.06*size[DIM_Y])*(0.06*size[DIM_Y]);
	const ieee64 largeVectorDifference = 
		(0.12*size[DIM_X])*(0.12*size[DIM_X]) + 
		(0.12*size[DIM_Y])*(0.12*size[DIM_Y]);

	for (int32 i=0; i<(int32)cm.data.size()-1; i++) {
		if ((cm.data[i].valid == 0) && (cm.data[i+1].valid == 0)) {

			ieee64 maxVectorDifference = -10e10;
			ieee64 nnx1, nny1;
			ieee64 nnx2, nny2;
			bool firstFrameHasLargestVector=true;

			for (int32 y=8; y<size[DIM_Y]-8; y+=16) {
				for (int32 x=8; x<size[DIM_X]-8; x+=16) {
					cm.data[i].getVectors  (x,y,nnx1,nny1);
					cm.data[i+1].getVectors(x,y,nnx2,nny2);
					ieee64 vectorDifference = (nnx1-nnx2)*(nnx1-nnx2) + (nny1-nny2)*(nny1-nny2);
					ieee64 vectorLength1 = (x-nnx1)*(x-nnx1) + (y-nny1)*(y-nny1);
					ieee64 vectorLength2 = (x-nnx2)*(x-nnx2) + (y-nny2)*(y-nny2);
					if (maxVectorDifference < vectorDifference) {
						maxVectorDifference = vectorDifference;
						firstFrameHasLargestVector = (vectorLength1 > vectorLength2);
					} // if
				} // for
			} // for

			if (maxVectorDifference > mediumVectorDifference) {
				if (firstFrameHasLargestVector) {
					cm.data[i].valid = 6;    // 6=parameters calculated, but the motion vectors of 2 following frames are different (may not be an error)
				} else {
					cm.data[i+1].valid = 6;  // 6=parameters calculated, but the motion vectors of 2 following frames are different (may not be an error)
				} // else
			} // if
			if (maxVectorDifference > largeVectorDifference) {
				if (firstFrameHasLargestVector) {
					cm.data[i].valid = 5;    // 5=parameters calculated, but the motion vectors of 2 following frames are very different
				} else {
					cm.data[i+1].valid = 5;  // 5=parameters calculated, but the motion vectors of 2 following frames are very different
				} // else
			} // if

		} // if
	} // for

	return;
}



void 
CameraModel::validateCameraParametersTranslation (CameraModelData& cm) 
{

	// check number of entries
	if ((int32)cm.data.size() != size[DIM_Z]) {
		CameraModelItem item;
		//cerr << "add camera model items (" << (size[DIM_Z] - cm.data.size()) << ")" << endl;
		while ((int32)cm.data.size() < size[DIM_Z]) {
			item.frame = cm.data.size();
			item.valid=0;
			cm.data.push_back(item);
		} // while
	} // if
	if ((int32)cm.data.size() != size[DIM_Z]) {
		cerr << "CameraModel::calculateCameraParameters: wrong number of frames" << endl;
		cerr << "data size : " << cm.data.size() << endl;
		cerr << "video size: " << size[DIM_Z] << endl;
		exit(1);
	} // if


	// compare frame number with index item (TODO: what happens, if they do not match???)
	for (int32 i=0; i<(int32)cm.data.size(); i++) {
		CameraModelItem item;
		item = cm.data[i];
		if (i != cm.data[i].frame) {
			cerr << "CameraModel::parameters of frame " << i << " not found." << endl;
		} // if
	} // for


	// translation only, remove other parameters
	for (int32 i=0; i<(int32)cm.data.size(); i++) {
		TransformationData data;
		data = cm.data[i].transData;
		data.a00 = data.a11 = 1.0; 
		data.a01 = data.a10 = 0.0; 
		data.px  = data.py  = 0.0; 
		cm.data[i].transData = data;
	} // for
	


	// normalize length of translation
	for (int32 i=0; i<(int32)cm.data.size(); i++) {
		ieee64 tx = cm.data[i].transData.tx;
		ieee64 ty = cm.data[i].transData.ty;

		if (tx>=0) {
			tx = ((int)(100*tx+0.5)) / 100.0;
		} else {
			tx = ((int)(100*tx-0.5)) / 100.0;
		} // else
		if (ty>=0) {
			ty = ((int)(100*ty+0.5)) / 100.0;
		} else {
			ty = ((int)(100*ty-0.5)) / 100.0;
		} // else

		cm.data[i].transData.tx = tx;
		cm.data[i].transData.ty = ty;
	} // for


	// analyse camera model items, remove parameters, which do not make sense:
	//   --> a motion vector should not be larger than 40% of the image
	const ieee64 allowedVectorLength = 
		(0.4*size[DIM_X])*(0.4*size[DIM_X]) + 
		(0.4*size[DIM_Y])*(0.4*size[DIM_Y]);
	for (int32 i=0; i<(int32)cm.data.size(); i++) {
		
		//CameraModeltem item = cm.data[i];;
		if (cm.data[i].valid == 0) {


			ieee64 maxVectorLength = -10e10;
			ieee64 nnx, nny;
			for (int32 y=8; y<size[DIM_Y]-8; y+=16) {
				for (int32 x=8; x<size[DIM_X]-8; x+=16) 
				{
					cm.data[i].getVectors(x,y,nnx,nny);
					ieee64 vectorLength = (x-nnx)*(x-nnx) + (y-nny)*(y-nny);
					if (maxVectorLength < vectorLength) maxVectorLength = vectorLength;
				} // for
			} // for
			

			// a motion vector should not be larger than half the image
			if (maxVectorLength > allowedVectorLength) {
				//cerr << "invalid frame number: " << cm.data[i].frame << endl;
				cm.data[i].valid = 4;  // 4=parameters calculated, but they do not make sense (e.g. large object visible)
			} // if
		} // if
	} // for


	// analyse camera model of two following frames: set (valid==5 or valid==6), if the vectors change much:
	//   --> get vectors of 2 frames
	//   --> diff = maximum of difference of 2 corresponding vectors
	//   --> if (diff>largeVectorDifference) ==> valid=5
	//   --> if (diff>mediumVectorDifference) ==> valid=6
	const ieee64 mediumVectorDifference = 
		(0.06*size[DIM_X])*(0.06*size[DIM_X]) + 
		(0.06*size[DIM_Y])*(0.06*size[DIM_Y]);
	const ieee64 largeVectorDifference = 
		(0.12*size[DIM_X])*(0.12*size[DIM_X]) + 
		(0.12*size[DIM_Y])*(0.12*size[DIM_Y]);

	for (int32 i=0; i<(int32)cm.data.size()-1; i++) {
		if ((cm.data[i].valid == 0) && (cm.data[i+1].valid == 0)) {

			ieee64 maxVectorDifference = -10e10;
			ieee64 nnx1, nny1;
			ieee64 nnx2, nny2;
			bool firstFrameHasLargestVector=true;

			for (int32 y=8; y<size[DIM_Y]-8; y+=16) {
				for (int32 x=8; x<size[DIM_X]-8; x+=16) {
					cm.data[i].getVectors  (x,y,nnx1,nny1);
					cm.data[i+1].getVectors(x,y,nnx2,nny2);
					ieee64 vectorDifference = (nnx1-nnx2)*(nnx1-nnx2) + (nny1-nny2)*(nny1-nny2);
					ieee64 vectorLength1 = (x-nnx1)*(x-nnx1) + (y-nny1)*(y-nny1);
					ieee64 vectorLength2 = (x-nnx2)*(x-nnx2) + (y-nny2)*(y-nny2);
					if (maxVectorDifference < vectorDifference) {
						maxVectorDifference = vectorDifference;
						firstFrameHasLargestVector = (vectorLength1 > vectorLength2);
					} // if
				} // for
			} // for

			if (maxVectorDifference > mediumVectorDifference) {
				if (firstFrameHasLargestVector) {
					cm.data[i].valid = 6;    // 6=parameters calculated, but the motion vectors of 2 following frames are different (may not be an error)
				} else {
					cm.data[i+1].valid = 6;  // 6=parameters calculated, but the motion vectors of 2 following frames are different (may not be an error)
				} // else
			} // if
			if (maxVectorDifference > largeVectorDifference) {
				if (firstFrameHasLargestVector) {
					cm.data[i].valid = 5;    // 5=parameters calculated, but the motion vectors of 2 following frames are very different
				} else {
					cm.data[i+1].valid = 5;  // 5=parameters calculated, but the motion vectors of 2 following frames are very different
				} // else
			} // if

		} // if
	} // for

	return;
}



void 
CameraModel::readValues (CameraModelData& cm) 
{

	// initialize cm
	cm.data.clear();
	cm.version = CAMERA_MODEL_VERSION;
	cm.cropleft   = cropleft;
	cm.cropright  = cropright;
	cm.croptop    = croptop;
	cm.cropbottom = cropbottom;
	
	// get path of filename, do not create new directory
	CreateMocaPath mocaPath (pathFilename, false);
	cm.filename = mocaPath.file;


	// get camera parameters 
	if (!calcTranslation) {
		calculateCameraParameters (cm);
	} else {
		calculateTranslationParameters (cm);
	} // else
	return;
} // readValues

