/*
   This file is part of the MoCA video library.
   Name: CameraOperation.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  
   \author    Stephan Kopf
   
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de
	
	 (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	 
*/


#ifndef CAMERA_OPERATION_H
#define CAMERA_OPERATION_H

// C++ includes
#include <iomanip>
#include <sstream>
#include <iostream>
using namespace std;


// MoCA includes
#include <CameraModel.h>
#include <CameraTransformation.h>
#include <MedianFilter.h>
#include <GaussFilter.h>



static const string CAMERA_OPERATION_VERSION  = "Version-0.42";
static const string CAMERA_OPERATION_FILENAME = "cameraoperation.idx";


/*
	CameraOperation: get a textual description of the camera operation, 
	                 e.g., between frame 80-100: left pan 
*/

enum cameraOperationType {
	CAMERA_OPERATION_UNKNOWN      = 0, 
	CAMERA_OPERATION_STATIC       = 1, 
	CAMERA_OPERATION_ZOOM_IN      = 2, 
	CAMERA_OPERATION_ZOOM_OUT     = 3, 
	CAMERA_OPERATION_PAN_LEFT     = 4, 
	CAMERA_OPERATION_PAN_RIGHT    = 5, 
	CAMERA_OPERATION_TILT_UP      = 6, 
	CAMERA_OPERATION_TILT_DOWN    = 7,
	CAMERA_OPERATION_ROTATE_LEFT  = 8,
	CAMERA_OPERATION_ROTATE_RIGHT = 9
};


struct CameraOperationItem
{
	CameraOperationItem () {
		firstFrame = -1;
		lastFrame  = -1;
		length     = 0;
		strength   = 0;
		cot        = CAMERA_OPERATION_UNKNOWN;
	} 
    CameraOperationItem& operator=(const CameraOperationItem &other) {
		if (this != &other) {
			firstFrame = other.firstFrame;
			lastFrame  = other.lastFrame;
			length     = other.length;
			strength   = other.strength;
			cot        = other.cot;
			trans      = other.trans;
		} // if
        return *this;
    } // operator=
	
	int32 firstFrame;				// first frame 
	int32 lastFrame;				// last frame
	ieee32 length;					// e.g. 10 pixels to the left
	ieee64 strength;				// aggregated parameters
	cameraOperationType cot;		// camera operation

	TransformationData trans;
	//ieee64 a00, a01, a10, a11;   // affine transformation
	//ieee64 tx, ty;               // pan, tilt (horizontal and vertical movement)
	//ieee64 px, py;               // perscpective parameters
	//
	//        a00*x + a01*y + tx           a10*x + a11*y + ty 
	//  x' = --------------------    y' = --------------------  
	//        px*x + py*y + 1              px*x + py*y + 1
	//
};
std::ostream& operator<< (ostream&,    const CameraOperationItem&);
OTagStream&   operator<< (OTagStream&, const CameraOperationItem&);
ITagStream&   operator>> (ITagStream&, CameraOperationItem&);



struct CameraOperationData
{
	CameraOperationData () 
	{
		version = CAMERA_OPERATION_VERSION;
		filename = "";
		dataSize = 0;
		data.clear();
	} 
    CameraOperationData& operator=(const CameraOperationData &other) {
		if (this != &other) {
			version = other.version;
			filename= other.filename;
			dataSize= other.dataSize;
			data    = other.data;
		} // if
        return *this;
    } // operator=
	
	string version;
	string filename;
	int32 dataSize;
	vector<CameraOperationItem> data;
};

std::ostream& operator<< (ostream&,    const CameraOperationData&);
OTagStream&   operator<< (OTagStream&, const CameraOperationData&);
ITagStream&   operator>> (ITagStream&, CameraOperationData&);



/*
Constructor: initializes variables

	process(): 
	if status file does not exist:
	calculate the camera operation from mpeg-file, save status file
	else:
	read camera operation from status file
*/
class CameraOperation
{
public:
	
	CameraOperation(const string &pathFilename);
	~CameraOperation() {};
	
	void process (CameraOperationData&);
	void printStatistics (string &fn, CameraOperationData& op);

private:
	void readValues (CameraOperationData& cod);
	void smoothValues (const Signal<ieee64> &v0, Signal<ieee64> &v1, ieee64 th, ieee64 avg);
	void getAverage (const CameraModelData &cm, TransformationData& t, int32 left, int32 right);

	void detectStatic (const CameraModelData &cm, CameraOperationData& op);
	void detectPan    (const CameraModelData &cm, CameraOperationData& op);
	void detectTilt   (const CameraModelData &cm, CameraOperationData& op);
	void detectZoom   (const CameraModelData &cm, CameraOperationData& op);

#if 0
	void detectRotateLeft  (const CameraModelData &cm, CameraOperationData& op);
	void detectRotateRight (const CameraModelData &cm, CameraOperationData& op);
#endif

	string pathFilename;
};

#endif
