/*
    CssVideoAnalysis.cpp     analyse a video, extract CSS features
    Copyright (C) 2004 Gerald Kuehne

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <CssVideoAnalysis.h>

OTagStream& operator<< (OTagStream& os, const CssVideoAnalysisPeak& cm)
{
    TagFormat::writeStartTag(os, "CssPeak");
	TagFormat::writeStartTag(os, "position");   os << cm.position;   TagFormat::writeEndTag(os, "position");
	TagFormat::writeStartTag(os, "height");     os << cm.height;     TagFormat::writeEndTag(os, "height");
	TagFormat::writeStartTag(os, "width");      os << cm.width;      TagFormat::writeEndTag(os, "width");
	TagFormat::writeStartTag(os, "isConcave");  os << cm.isConcave;  TagFormat::writeEndTag(os, "isConcave");
    TagFormat::writeEndTag (os, "CssPeak");

    return os;
}

ITagStream& operator>> (ITagStream& is, CssVideoAnalysisPeak& cm)
{
    TagFormat::readStartTag(is, "CssPeak");
	TagFormat::readStartTag(is, "position");   is >> cm.position;   TagFormat::readEndTag(is, "position");
	TagFormat::readStartTag(is, "height");     is >> cm.height;     TagFormat::readEndTag(is, "height");
	TagFormat::readStartTag(is, "width");      is >> cm.width;      TagFormat::readEndTag(is, "width");
	TagFormat::readStartTag(is, "isConcave");  is >> cm.isConcave;  TagFormat::readEndTag(is, "isConcave");
    TagFormat::readEndTag (is, "CssPeak");
	cm.visited=false;
    return is;
}

OTagStream& operator<< (OTagStream& os, const CssVideoAnalysisFrame& src)
{
	TagFormat::writeNodeValue(os, "frame",src.frameNo);
	TagFormat::writeNodeValue(os, "xyRation",src.xyRatio);
    for (int32 i=0; i<CSS_MAX_PEAK_NUMBERS; i++) {
        os << src.peak[i];
    }
    return os;
}

ITagStream& operator>> (ITagStream& is, CssVideoAnalysisFrame& dest)
{
    for (int32 i=0; i<CSS_MAX_PEAK_NUMBERS; i++) {
        dest.peak[i].clear();
    } // for

    char c;
    bool failFlag = false;
    if (is >> c) {
        is.putback(c);

		TagFormat::readNodeValue(is, "frame",dest.frameNo);
		TagFormat::readNodeValue(is, "xyRation",dest.xyRatio);

		for (int32 i=0; i<CSS_MAX_PEAK_NUMBERS; i++) {
			if (!(is >> dest.peak[i])) {
				failFlag = true;
				break;
			} // if
		} // for
	} // if

    if (!failFlag) {
        is.clear(is.rdstate() & ~ios::failbit);
    }

    return is;
}


OTagStream& operator<< (OTagStream& os, const CssVideoAnalysisData& src)
{
	TagFormat::writeNodeValue(os, "version",src.version);
	TagFormat::writeNodeValue(os, "filename",src.filename);

	int32 numberOfFrames = src.frameData.size();
	TagFormat::writeNodeValue(os, "numberOfFrames",numberOfFrames);

	list<CssVideoAnalysisFrame>::const_iterator it;
    for (it=src.frameData.begin(); it!=src.frameData.end(); it++) {
        os << (*it);
    } // for
    return os;
}

ITagStream& operator>> (ITagStream& is, CssVideoAnalysisData& dest)
{
	dest.frameData.clear();

    char c;
    bool failFlag = false;
    if (is >> c) {
        is.putback(c);

		TagFormat::readNodeValue(is, "version",dest.version);
		TagFormat::readNodeValue(is, "filename",dest.filename);
		TagFormat::readNodeValue(is, "numberOfFrames",dest.numberOfFrames);

		CssVideoAnalysisFrame frame;
		for (int32 i=0; i<dest.numberOfFrames; i++) {
			if (is >> frame) {
				dest.frameData.push_back(frame);
			} else {
				failFlag = true;
				break;
			} // if
		} // for
	} // if

    if (!failFlag) {
        is.clear(is.rdstate() & ~ios::failbit);
    }

    return is;
}




void
CssVideoAnalysis::
process (CssVideoAnalysisData &data)
{
	// read data
	ReadMocaData<CssVideoAnalysisData> reader (pathFilename, CSS_VIDEO_ANALYSIS_FILENAME);
	bool dataValid = reader.process (CSS_VIDEO_ANALYSIS_VERSION, data); 

	if (!dataValid) {

		// create new index
		cerr << "CssVideoAnalysis: Creating index file ..." << endl;
		readValues(data);
			
		// save index file
		WriteMocaData<CssVideoAnalysisData> writer (pathFilename, CSS_VIDEO_ANALYSIS_FILENAME);
		writer.process(data);
		cerr << "\rCssVideoAnalysis: Creating index file ...done                         " << endl;
	} // if

	return;

}


void
CssVideoAnalysis::
calculateCssData (CssVideoAnalysisData &data)
{
	// initialize data
	data.filename = pathFilename;
	data.version = CSS_VIDEO_ANALYSIS_VERSION;
	data.numberOfFrames = 0;
	data.frameData.clear();

	// get object mask of the video
	ObjectSegmentationData segmentedObjectsData;
	{
		ObjectSegmentation segmentedObjects(pathFilename, cropLeft, cropRight, cropTop, cropBottom);
		segmentedObjects.process(segmentedObjectsData);
		cerr << segmentedObjectsData << endl;
	}

	// initialize css variables
	list<ObjectSegmentationContour>::iterator it;
    CssImage css;
	Image<uint8> cssImg;
    CssMaximumData cmList;
	vector<CssMaximumItem>::iterator cssItemIt; 
    CssMaximaFeature cmf;

	// for each object mask:
	//   - calculate css peaks
	for (it=segmentedObjectsData.data.begin(); it!=segmentedObjectsData.data.end(); it++) {
		ObjectSegmentationContour objectData = (*it);

		// calculate css image
        css.process((*it).contourParameter, cssImg);

		WriteImage (cssImg, "F:/img-css-1.png");


		// calculate css peaks
        cmf.processBottomWidth (cssImg, cmList);

		// set cssData
		CssVideoAnalysisFrame cssData;
		cssData.frameNo = (*it).frame;
		int32 n=cmList.data.size();
		if (n>CSS_MAX_PEAK_NUMBERS) n=CSS_MAX_PEAK_NUMBERS;
	
		cssItemIt = cmList.data.begin();
		for (int32 i=0; i<n; i++) {
			cssData.peak[i].position  = (*cssItemIt).position;
			cssData.peak[i].height    = (*cssItemIt).height;
			cssData.peak[i].width     = (*cssItemIt).width;
			cssData.peak[i].isConcave = (*cssItemIt).isConcave;
			cssData.peak[i].visited   = false;
			cssItemIt++;
		} // for

		// get xyRation
		int32 pxMin=10000000;
		int32 pxMax=-1;
		int32 pyMin=10000000;
		int32 pyMax=-1;
		for (int32 i=0; i< CSS_MAX_PEAK_NUMBERS; i++) {
			int32 px = (*it).contourParameter(i,0);
			int32 py = (*it).contourParameter(i,1);
			if(pxMin > px) pxMin = px;
			if(pxMax < px) pxMax = px;
			if(pyMin > py) pyMin = py;
			if(pyMax < py) pyMax = py;
		} // for
		cssData.xyRatio = (pyMax-pyMin);
		if (cssData.xyRatio > 0.0) {
			cssData.xyRatio = (pxMax-pxMin) / cssData.xyRatio;
		} else {
			cssData.xyRatio = 1.0;
		} //

		data.frameData.push_back(cssData);
		data.numberOfFrames++;
	} // for

	return;
}

void
CssVideoAnalysis::
readValues (CssVideoAnalysisData &data)
{
	// get css data
	calculateCssData(data);
	data.numberOfFrames = data.frameData.size();

	// read database
	


	// matching


	return;
}

