/*
   This file is part of the MoCA video library.
   Name: EcrFeature.h
   Copyright (C) 2004 Markus Beier, Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: beier@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  \author    Markus Beier, Stephan Kopf
  
   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: beier@pi4.informatik.uni-mannheim.de
   
	(c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	
*/


#ifndef ECR_FEATURE_H
#define ECR_FEATURE_H


//C++ includes
#include <vector>
#include <string>
#include <sstream>
#include <iostream>
using namespace std;

// MoCA includes
#include <MocaTypes.h>
#include <VideoDecoder.h>
#include <MocaTypes.h>
#include <CannyFilter.h>
#include <DilationFilter.h>
#include <GrayscaleFilter.h>
#include <BinarizeFilter.h>
#include <InverseFilter.h>
#include <CropFilter.h>
#include <OImageStream.h>
#include <CameraModel.h>
#include <CameraTransformation.h>
#include <EdgeDetectionFilter.h>




//! Class to calculate ECR (edge change ratios)
/*!
  This class will calculate the ECR values of the frames of a video.
  The difference of consecutive ECR values will be calculated and 
  stored in a vector provided as parameter.
*/

const static string ECR_FEATURE_VERSION  ="Version-0.05";
const static string ECR_FEATURE_FILENAME ="ecr.idx";



struct EcrData
{
	EcrData (string &pathFilename) 
	{
		version = ECR_FEATURE_VERSION;
		filename = "";

		size              = 0;
		minEcrValue       = 1e10;
		maxEcrValue       = -1 ;
		avgEcrValue       = 0;
		ecrFilename       = "";
		pinFilename       = "";
		poutFilename      = "";
		ecr.clear();
		pin.clear();
		pout.clear();

		this->pathFilename = pathFilename;
		CreateMocaPath pathReader (pathFilename, false);
		mocaPath = pathReader.destinationPath;
	} 
    EcrData& operator=(const EcrData &other) {
		if (this != &other) {
			version            = other.version;
			filename           = other.filename;

			size               = other.size;
			minEcrValue        = other.minEcrValue;
			maxEcrValue        = other.maxEcrValue;
			avgEcrValue        = other.avgEcrValue;
			ecrFilename        = other.ecrFilename;
			pinFilename        = other.pinFilename;
			poutFilename       = other.poutFilename;
			ecr                = other.ecr;
			pin                = other.pin;
			pout               = other.pout;
			mocaPath           = other.mocaPath;
			pathFilename       = other.pathFilename;
		} // if
        return *this;
    } // operator=
	
	
	ieee64 minEcrValue;
	ieee64 maxEcrValue;
	ieee64 avgEcrValue;
	string ecrFilename;
	string pinFilename;
	string poutFilename;

	int32 size;
    vector<ieee64> ecr;
    vector<ieee64> pin;
    vector<ieee64> pout;

	string version;
	string filename;
	string mocaPath;
	string pathFilename;
};

std::ostream& operator<< (ostream&,    const EcrData&);
OTagStream&   operator<< (OTagStream&, const EcrData&);
ITagStream&   operator>> (ITagStream&, EcrData&);




class EcrFeature
{
public:
	static const string DEFAULT_ECR_FEATURE_FILE_NAME;

	
    //! Contructor initializing the class with a filename of the MPEG file 
    /*!
	\param fullname of Type string& provides the filename of the mpeg file to be processed
    */
    EcrFeature ();
    EcrFeature (uint8 edgeThreshold, 
				uint8 dilateRadius, uint8 radiusmult, ieee64 sumreg, 
				uint8 sumradius, uint8 maxRadius, uint8 transradMax, ieee64 transThreshold);
	
    //! Default destructor
    ~EcrFeature() {};
	
	
    //! Calculating the ECR differences of consecutive frames in the 
	//! given video, store the ECR values in a file and then returning these values.
    /*!
	\param item of Type EcrData& is the structure wherein the resulting data will be stored. 
    */
	void process (EcrData& ecr);

	
private:
	void readValues (EcrData& ecr);
	uint32 countEdgePixel (const Image<uint8>& img);
	ieee64 getOutgoingPixel(const Image<uint8>& edgeImage, const Image<uint8>& dilatedImage);

	void processImages (
			const Image<uint8> &oldImg, const Image<uint8> &oldDilated, ieee64 oldEdgeCount,
			const Image<uint8> &newImg, const Image<uint8> &newDilated, ieee64 newEdgeCount,
			ieee64 &ecrValue, ieee64 &rhoOut, ieee64 &rhoIn);


private:
	string pathFilename;

	uint8 edgeThreshold;
	uint8 dilateRadius;
	uint8 radiusmult;
	ieee64 sumreg;
	uint8 sumradius;
	uint8 maxRadius;
	uint8 transradMax;
	ieee64 transThreshold;

	VideoSize size;
}; //class EcrFeature


#endif
