/*
    FaceDetection.cpp     detect faces in videos
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "FaceDetection.h"



FaceDetection::FaceDetection(FaceDetectionParameter& p)
{
	if (p.inputData == FACE_INPUT_UNKNOWN) {
		cerr << "Error in FaceDetection: specify inputData in constructor (image or video)" << endl << endl;
		exit(1);
	} // if

    lightingCorrectionAllocatedSize = -1;
    lightingCorrectionX  = -1;
    lightingCorrectionX2 = -1;
    networkUint8ToFloat[0] = -2.0f;

    video=NULL;
    nameOfOutFile=NULL;
    nnInput=NULL;
    nnOutput=NULL;
    mask=NULL;
    pixel=NULL;
    columnSum = NULL;
    rowSum = NULL;

    param = p;
	shiftX=0;
	shiftY=0;
    initializeParameters(param.srcFileFirst,param.srcFileLast,param.stepWidth,param.analyseOne);

} // FaceDetection::FaceDetection()






FaceDetection::~FaceDetection() {
	int i;

	if (pixel!=NULL) {
		for (i = 0; i < param.size; i++) {
			if (pixel[i]!=NULL) delete[] pixel[i];
			pixel[i]=NULL;
		} // for
		delete[] pixel;
	} // if
	pixel=NULL;

	if (mask!=NULL) {
		for (i = 0; i < param.size; i++) {
			if (mask[i]!=NULL) delete[] mask[i];
			mask[i]=NULL;
		} // for
		delete[] mask;
	} // if
	mask=NULL;

    if (nnOutput!=NULL) {
		for (i = 0; i < FACE_DETECTION_DEFAULT_NUMBER_OF_NETS; i++) {
			if (nnOutput[i] != NULL) delete[] nnOutput[i];
			nnOutput[i]=NULL;
		} // if
		delete[] nnOutput;
	} // if
	nnOutput=NULL;

    if (nnInput !=NULL) delete[] nnInput;
	nnInput=NULL;

	if (video!=NULL) delete video;
	video=NULL;

	// ???
	for (i = 0; i < FACE_DETECTION_DEFAULT_NUMBER_OF_NETS; i++) {
		if (facesThroughNet[i]!=NULL) {
			facesThroughNet[i]->clear();
			delete facesThroughNet[i];
		} // if
		facesThroughNet[i]=NULL;
	} // for

	
	if (columnSum!=NULL) {
		delete[] columnSum;
	} // if
	columnSum=NULL;

	if (rowSum!=NULL) {
		delete[] rowSum;
	} // if
	rowSum=NULL;

} // FaceDetection::~FaceDetection()


void 
FaceDetection::initializeParameters(int srcFileFirst, int srcFileLast, int stepWidth, int32 analyseOne)
{
	mask=NULL;
	video=NULL;
	totalNumberOfFaces=0;

    scalingFactor = FACE_DETECTION_SCALE_FACTOR;
    colorCheckPos = colorCheckNeg = 0;
    frameCounter = 0;
    inverseFactor = 1.0;

    neuralNet[0] = nn1; neuralNet[1] = nn2; neuralNet[2] = nn3; neuralNet[3] = nn4; neuralNet[4] = nn5;
    nnInput = new float[nn1REC.NoOfInput];
    nnOutput = new float*[FACE_DETECTION_DEFAULT_NUMBER_OF_NETS];
    nnOutput[0] = new float[nn1REC.NoOfOutput]; nnOutput[1] = new float[nn2REC.NoOfOutput];
    nnOutput[2] = new float[nn3REC.NoOfOutput]; nnOutput[3] = new float[nn4REC.NoOfOutput];
    nnOutput[4] = new float[nn5REC.NoOfOutput];

	if ((param.mpgFilename == "") && (param.inputData == FACE_INPUT_VIDEO)) {
        cerr << "Error in FaceDetection: MPEG file name not specified" << endl;
        cerr << endl;
        exit(1);		
	} // if


	if (param.inputData == FACE_INPUT_IMAGE) {
		if (param.defaultFaceSize!=FACE_DETECTION_SIZE_DEFAULT) {

			switch (param.defaultFaceSize)
			{
			case FACE_DETECTION_SIZE_MEDIUM:
				param.fastSize = (int32) (1.5 * FACE_DETECTION_PARAMETER_SIZE);
				break;
			case FACE_DETECTION_SIZE_LARGE:
				param.fastSize = (int32) (2.0 * FACE_DETECTION_PARAMETER_SIZE);
				break;
			case FACE_DETECTION_SIZE_HUGE:
				param.fastSize = (int32) (3.0 * FACE_DETECTION_PARAMETER_SIZE);
				break;
			} // switch

		} // if
	} // if


	if (param.inputData == FACE_INPUT_VIDEO) {

		{
			VideoStatisticData data;
			VideoStatistic vidStat (param.mpgFilename);
			vidStat.process(data);
			param.useColorInfo= (data.color==TYPE_COLOR);
		}


		video = new VideoDecoder (param.mpgFilename); 
		//if (!video->isColor()) param.useColorInfo=false;

		// initialize framelist
		if (stepWidth <=0) stepWidth=1;
		int32 tmp  = video->getSize()[DIM_Z] - 1;
		if ((srcFileLast == -1) || (srcFileLast > tmp)) {
			srcFileLast = tmp;
		} // if		
		if (srcFileFirst > srcFileLast) {
			srcFileFirst = srcFileLast;
		} // if
		if (srcFileFirst < 0) {
			srcFileFirst = 0;
		} // if

		// get minimal dimension of the video and set size for large faces
		int32 minDimension = video->getSize()[DIM_X];
		if (minDimension>video->getSize()[DIM_Y]) minDimension=video->getSize()[DIM_Y];

		if (param.defaultFaceSize!=FACE_DETECTION_SIZE_DEFAULT) {
			// medium sized face
			if (param.defaultFaceSize==FACE_DETECTION_SIZE_MEDIUM) {
				param.fastSize=(int32)(0.15*minDimension); 
			} // if
			
			// large face
			if (param.defaultFaceSize==FACE_DETECTION_SIZE_LARGE) {
				param.fastSize=(int32)(0.25*minDimension); 
			} // if
			
			// huge face
			if (param.defaultFaceSize==FACE_DETECTION_SIZE_HUGE) {
				param.fastSize=(int32)(0.35*minDimension); 
			} // if
			if (param.fastSize<FACE_DETECTION_PARAMETER_SIZE) {
				param.fastSize=FACE_DETECTION_PARAMETER_SIZE;
			} // if
		} // if
	
		param.analyseFrame.resize(SignalSize(srcFileLast+1));
		int32 i;
		for (i=0; i < param.analyseFrame.getSize()[DIM_X]; i++) {
			param.analyseFrame(i) = false;
		} // for
		if (analyseOne !=-1) {
			// analyse only one frame
			param.analyseFrame(analyseOne) = true;
		} else {
			for (i=srcFileFirst; i<srcFileLast; i+=stepWidth) {
				param.analyseFrame(i) = true;
			} // for
		} // else
	}


    for (int32 i = 0; i < FACE_DETECTION_DEFAULT_NUMBER_OF_NETS; i++) {
		facesThroughNet[i] = NULL;
        facesThroughNet[i] = new list<Face>();
    } // for
    pixel = new int*[param.size];
    for (int32 i = 0; i < param.size; i++) {
        pixel[i] = new int[param.size];
    } // for
    makeMask();
} // initializeParameters


void 
FaceDetection::makeMask() 
{
    int hide = 0;
    int maxHide = 3*param.size/20;
    mask = new int*[param.size];
    for (int i = 0; i < param.size; i++)
        mask[i] = new int[param.size];
  
    for (int y = 0; y < param.size; y++) {
        if (y < maxHide)
            hide = maxHide - y;
        else if (y >= (param.size - maxHide))
            hide = y - param.size + maxHide + 1;
        else
            hide = 1;
    
        for (int x = 0; x < param.size; x++) {
            if ((x < hide) || (x >= (param.size - hide)))
                mask[y][x] = 0;
            else
                mask[y][x] = 1;
        } // for
    } // for
    return;
} // FaceDetection::makeMask()



void 
FaceDetection::scaleImage(double scaleFactor) 
{
    //! ScaleImageObj<uint8> scaleObj;
    //! scaleObj.inPlace(image, scaleFactor, scaleFactor);
    ScaleFilter<uint8> sf (scaleFactor, scaleFactor);
    sf.process (image);
    
    // get the pixel values
    inverseFactor *= 1.0/scaleFactor;
    if (param.useColorInfo) {
        //!  scaleObj.inPlace(breakMask, scaleFactor, scaleFactor);
        sf.process (breakMask);
    } // if
    return;
} // FaceDetection::scaleImage()


void 
FaceDetection::rotateImage(int angle) 
{
    //! image = RotateImage(orgImage, image, angle * PI/180.0);
    RotationFilter<uint8> rf ((ieee64)angle * FACE_DETECTION_PI/180.0);
    rf.process (orgImage, image);

    // get the pixel values
    inverseFactor = 1.0;
    if (param.useColorInfo) {
        //! colorImage = RotateImage(orgColorImage, colorImage, angle*PI/180.0);
        rf.process (orgColorImage, colorImage);
    } // if
    return;
} // FaceDetection::rotateImage()


inline bool 
FaceDetection::checkColor(const int x, const int y) 
{
    //! return (breakMask[0][y][x] & 128);
    return (breakMask(x,y,0) & 128);
} // FaceDetection::checkColor()


void 
FaceDetection::cropImage(int x, int y) 
{
    x -= param.size/2;
    y -= param.size/2;
    for (int j = 0; j < param.size; j++) {
        for (int i = 0; i < param.size; i++) {
            //! pixel[j][i] = image[0][y + j][x + i];
            pixel[j][i] = image (x + i, y + j, 0);
        } // for
    } // for
    return;
} // FaceDetection::cropImage()


bool
FaceDetection::lightingCorrection() 
{
    // -1 = nothing allocated yet
    // > 0  some space already allocated
    if (lightingCorrectionAllocatedSize != param.size) {

        if (columnSum!=NULL) delete[] columnSum;
		columnSum=NULL;

		if (rowSum!=NULL) delete[] rowSum;
		rowSum=NULL;

        columnSum = new int[param.size];
        rowSum = new int[param.size];
        lightingCorrectionAllocatedSize = param.size;
        
        lightingCorrectionX  = param.size * (param.size+1) / 2;  // Sum (1, ... , size) 
        lightingCorrectionX2 = param.size * (param.size+1) * (2*param.size+1) / 6; // Sum (1^2, ... , size^2)
        //cerr << "Allocate Lighting correction temps" << endl;
    } // if

    int f[3] = {0, 0, 0};
    
    // extract grayvalues
	int i;
	for (i = 0; i < param.size; i++) {
        columnSum[i] = 0;
        rowSum[i] = 0;
        for (int j = 0; j < param.size; j++) {
            columnSum[i] += pixel[j][i];
            rowSum[i] += pixel[i][j];
        } // for
    } // for
  
    // prepare values
    for (i = 0; i < param.size; i++) {
        f[0] += columnSum[i];
        f[1] += (i + 1)*columnSum[i];
        f[2] += (i + 1)*rowSum[i];
    } // for
  
    float h = (float)f[0] * (float)lightingCorrectionX / (float)param.size;
    float a[3] = {0, 0, 0};
    a[2] = ( (float)f[2] - h) / (float)(lightingCorrectionX2*param.size - lightingCorrectionX*lightingCorrectionX);
    a[1] = ( (float)f[1] - h) / (float)(lightingCorrectionX2*param.size - lightingCorrectionX*lightingCorrectionX);
    a[0] = (float)f[0]/(float)(param.size*param.size) - (((float)lightingCorrectionX*a[1] + (float)lightingCorrectionX*a[2])/(float)param.size);
    
    int maxCorr = (int) a[0];
    // maxCorr kann wesentlich schneller bestimmt werden !
    if (a[1] > 0) maxCorr += (int) a[1] * (param.size-1);
    if (a[2] > 0) maxCorr += (int) a[2] * (param.size-1);
        
    int minV = 255;
    int maxV = 0;
  
    // correct pixel values
    float correctionTmp = a[0]; 
	int j;
    for (j = 0; j < param.size; j++) {
        float correctionFloat = correctionTmp;
        
        for (int i = 0; i < param.size; i++) {
            // Auch das geht schneller
            // correction = maxCorr - (int)(a[0] + a[1]*(float)i + a[2]*(float)j);
            int value = pixel[j][i] + maxCorr - int(correctionFloat);
            if (value < 0) value = 0;
            if (value > 255) value = 255;

            // resample values
            pixel[j][i] = value;
            // get min and max values
            if (mask[j][i] == 1) {
                if (value < minV) minV = value;
                if (value > maxV) maxV = value;
            }
            correctionFloat += a[1];
        }
        correctionTmp += a[2];
    } // for
  
    // Falls zu wenig Kontrast in dem Bildausschnitt vorkommt, 
    // Suche erst gar nicht nach einem Gesicht.
    // Fuehrt bestimmt zu einer falschen Erkennung
    if (maxV-minV < 30) {
        //cerr << min << "\t" << max << endl;
        return false;
    } // if

    // histogram equalisation
    float c2 = 0;
    c2 = 255.0/(float)(maxV - minV);  
    float c1 = (-(float)minV)*c2;

    for (j = 0; j < param.size; j++) {
        for (int i = 0; i < param.size; i++) {
            int value = (int)(c2*pixel[j][i] + c1);
            if (value < 0) value = 0;
            if (value > 255) value = 255;
            pixel[j][i] = value;
        } // for
    } // for
    return true;
} // FaceDetection::lightingCorrection()


void 
FaceDetection::network(float *nnInput, float **nnOutput) 
{

    // pre-compute table
    if (networkUint8ToFloat[0] == -2.0f) {
        float *puint8ToFloat = networkUint8ToFloat;
        //cerr << "Allocate uint8 to float table." << endl;
        for (int i=0 ; i<255 ; i++) {
            int value = i * 2 - 255;
            *puint8ToFloat++ = value / 255.0f;
        } // for
    } // if

    // Auch das geht ist schneller
    float *nnInputPtr = nnInput;
    for (int y = 0; y < param.size; y++) {
        for (int x = 0; x < param.size; x++) {
            // get single grayvalue
            // convert from 0..255 to -1..1 and save in netInput array
            *nnInputPtr = networkUint8ToFloat[pixel[y][x]];
            nnInputPtr++;
        } // for
    } // for
  
    // activate net and get result
    for (int i = 0; i < param.numberOfNets; i++)
        neuralNet[i](nnInput, nnOutput[i], 0);
    return;
} // FaceDetection::network()


void 
FaceDetection::reverseRotate(Face *face, int32 angle) 
{
    // rotates the vector r back to its former position
    // convert to local coordinate
    
    ImageSize size = colorImage.getSize();
    int centerX = face->x - size[DIM_X] /2;
    int centerY = face->y - size[DIM_Y] /2;

    double alpha = double(angle) / 180.0 * FACE_DETECTION_PI;
    face->x = (int32)(centerX * cos(alpha) - centerY * sin(alpha) + size[DIM_X]/2 - shiftX);
    face->y = (int32)(centerX * sin(alpha) + centerY * cos(alpha) + size[DIM_Y]/2 - shiftY);

    return;
} // FaceDetection::reverseRotate()


void 
FaceDetection::searchFaces(list<Face> **faceList, int32 angle) 
{
    if (param.verbose) {
        cout << param.mpgFilename << ": " << frameCounter << endl;
        cout << "width   height  #pat    #faces1 #faces2 #faces3 #faces4 #faces5 time(sec)" << endl;
    }
    
    int faceCounter[FACE_DETECTION_DEFAULT_NUMBER_OF_NETS];
  
    // generate break mask; 0 = do not investigate position
    if (param.useColorInfo) {
		int mode = FaceMaskFilter::FACE_MASK_COLOR;
        FaceMaskFilter fmf (mode);
        fmf.process (colorImage, breakMask); 
#if 0
		// debug
		string filename = "F:/temp/mask.png";
		WriteImage(breakMask, filename);
		filename = "F:/temp/img.png";
		WriteImage(colorImage, filename);
#endif
    } // if


    // define minimum size
    if (param.fastSize != param.size) {
        scaleImage( double(param.size)/double(param.fastSize) );
        scalingFactor = 0.83;
    }

    ImageSize imgSize = image.getSize();
    while ( (imgSize[DIM_X] > param.size) && (imgSize[DIM_Y] > param.size) ) {
        if (param.verbose) {
            cout << imgSize[DIM_X] << "\t" << imgSize[DIM_Y] << "\t";
        } // if
        
        long int t = time(NULL);
        int patternCounter = 0;
        for (int i = 0; i < FACE_DETECTION_DEFAULT_NUMBER_OF_NETS; i++)
            faceCounter[i] = 0;
        // compute max steps in x and y direction
        int distance = param.size/2;
        int maxX = imgSize[DIM_X] - param.size + distance;
        int maxY = imgSize[DIM_Y] - param.size + distance;

        for (int y=distance ; y<maxY ; y++) {
            for (int x=distance ; x<maxX ; x++) {
                if (!param.useColorInfo || checkColor(x, y)) {
                    colorCheckPos++;
                    cropImage(x, y);

                    // image processing (lighting correction, histogram equalisation)                    
                    if (!lightingCorrection()) continue;

                    // do NN detection
                    network(nnInput, nnOutput);
                    
                    for (int i = 0; i < param.numberOfNets; i++) {
                        if (nnOutput[i][0] >= param.threshold) {
                            // add faces to list
                            Face face(frameCounter, (int)(x*inverseFactor),
                                      (int)(y*inverseFactor),	(int)(param.size*inverseFactor),
                                      angle, nnOutput[i][0]);
                            if (angle != 0) {
								// invert rotation
                                reverseRotate(&face, angle);
                            } // if
                            faceList[i]->push_back(face);
                            faceCounter[i]++;
                        } // if
                    } // for
                } else {
                    colorCheckNeg++;
                }    
                patternCounter++;
            }
        }

        scaleImage(scalingFactor);
        imgSize = image.getSize();
        
        if (param.verbose) {
            int totalFaces = 0;
            cout << colorCheckPos << "\t";
            for (int i = 0; i < FACE_DETECTION_DEFAULT_NUMBER_OF_NETS; i++) {
                cout << faceCounter[i] << "\t";
                totalFaces += faceCounter[i];
            } // for
            cout << (time(NULL) - t) << endl;
        } // if        
    }
    return;
} // FaceDetection::searchFaces()


void 
FaceDetection::thresholding(int distance, int thres, list<Face> &fl) 
{
    // for each face, count neighbours
    list<Face>::iterator face;
    for (face = fl.begin(); face != fl.end(); face++)
        countNeighbours(*face, distance, fl);
    
	// apply thresholding, only faces with enough neighbours should survive
    face = fl.begin();
    while (face != fl.end()) {
        if ((*face).neighbourCounter < thres)
            fl.erase(face++);
        else
            face++;
    } // while
    return;
} // FaceDetection::thresholding()


void 
FaceDetection::countNeighbours(Face &f, const int distance, list<Face> &fl) 
{
    // counts the neighbours in the specified neighbourhood (distance)
    f.neighbourCounter = 0;
    for (list<Face>::iterator neighbour = fl.begin(); neighbour != fl.end(); neighbour++) {
        if ((f != *neighbour) &&
            (f.x - distance) <= ((*neighbour).x) &&
            (f.y - distance) <= ((*neighbour).y) &&
            (f.x + distance) >= ((*neighbour).x) &&
            (f.y + distance) >= ((*neighbour).y)) {
            // its a neighbour of the face
            f.neighbourCounter++;
            f.neighbourWeight += (*neighbour).weight;
        } // if
    } // for
    return;
} // FaceDetection::countNeighbours()


void 
FaceDetection::overlapDetection(list<Face> &fl) 
{
    // tests for overlapping faces and deletes the "smaller" one
    fl.sort();
    fl.reverse();
    for (list<Face>::iterator f = fl.begin(); f != fl.end(); f++) {
        if (param.verbose)
            cout << "\t" << (*f).frame << "\t" << (*f).x << "\t" << (*f).y
                 << "\t" << (*f).width << "\t" << ((*f).neighbourCounter + 1) << "\t"
                 << (*f).angle << "\t" << (*f).weight << endl;
        list<Face>::iterator neighbour = f;
        neighbour++;
        while (neighbour != fl.end()) {
            if ((*f).nextTo(*neighbour))
                fl.erase(neighbour++);
            else
                neighbour++;
        } // while
    } // for
    return;
} // FaceDetection::overlapDetection()


/*
void 
FaceDetection::drawRect(list<Face> &faceList) 
{
    // copy color Image;
    Image<uint8> outputImage = orgColorImage;
  
    // marks the faces in the image
    list<Face>::iterator f;
    for (f = faceList.begin(); f != faceList.end(); f++) {
        
        int r = (*f).width/2; // radius
        int x = (*f).x;
        int y = (*f).y;
        if ((*f).angle == 0) {
            DrawRectangle rect (Color::RED);
            ImagePosition p1 (x-r,y-r);
            ImagePosition p2(x+r,y+r);
            rect.process (p1, p2, outputImage);
        } else {

            
            // TODO !!!!!
            r +=3;
            DrawRectangle rect (Color::RED);
            ImagePosition p1 (x-r,y-r);
            ImagePosition p2(x+r,y+r);
            rect.process (p1, p2, outputImage);
            
        } // else
    } // for
  

    // save jpeg image
    stringstream ostr;
    //ostr << reportPath << setw(6) << setfill('0') << frameCounter << ".jpg" << ends;
    ostr << facePath << setw(6) << setfill('0') << frameCounter << ".jpg";
    string nameOfJpgFile = ostr.str();
    
    cerr << "save image: " << nameOfJpgFile << endl;
    OImageStream<uint8> ois (nameOfJpgFile.c_str());
    ois.write(outputImage);

    return;
} // FaceDetection::drawRect()
*/

void 
FaceDetection::doReport(list<Face> &faceList, int no) 
{
    if (param.verbose) {
        cout << "total\tx\ty\twidth\tfaces\trotation\tweight" << endl;
        int counter = 0;
        for (list<Face>::iterator f = faceList.begin(); f != faceList.end(); f++) {
            counter++;
            cout << counter << "\t" << (*f).x << "\t" << (*f).y
                 << "\t" << (*f).width << "\t" << ((*f).neighbourCounter + 1) << "\t"
                 << (*f).angle << "\t" << (*f).weight << endl;
        } // for
        cout << "total faces = " << faceList.size() << endl;
        if (param.useColorInfo)
            cout << "color check " << colorCheckPos << " times positive, "
                 << colorCheckNeg << " times negative." << endl;
    } // if

    if (param.report) {
		stringstream ostr;
		ostr << facePath << setw(6) << setfill('0') << no << ".txt";
		string nameOfReportFile = ostr.str();
		
		ofstream reportFile(nameOfReportFile.c_str());    

        //! reportFile << film->getPath() << film->getName() << endl;
        reportFile << facePath << param.mpgFilename << endl;
        reportFile << faceList.size() << endl;
        int counter = 0;
        for (list<Face>::iterator f = faceList.begin() ; 
             f != faceList.end(); 
             f++) {
            counter++;
            reportFile << counter << "\t" << (*f).x << "\t" << (*f).y << "\t"
                       << (*f).width << "\t" << (*f).angle << endl;
            cout << counter << "\t" << (*f).x << "\t" << (*f).y << "\t"
                 << (*f).width << "\t" << (*f).angle << endl;
        } // for

	    reportFile.close();
    } //if
    return;
} // FaceDetection::doReport()


void FaceDetection::resetValues() {
    inverseFactor = 1.0;
    colorCheckPos = colorCheckNeg = 0;

    for (int i = 0; i < FACE_DETECTION_DEFAULT_NUMBER_OF_NETS; i++) {
		//???
        //facesThroughNet[i]->erase(facesThroughNet[i]->begin(), facesThroughNet[i]->end());        
		if (facesThroughNet[i]!=NULL) {
			facesThroughNet[i]->clear();
			delete facesThroughNet[i];
		} // if
		facesThroughNet[i]=NULL;
        facesThroughNet[i] = new list<Face>();
    } // for
    return;
} // FaceDetection::resetValues()


/*
void FaceDetection::searchInRotatedImage(int angle, int step) 
{
    list<Face> *rotatedFaces[FACE_DETECTION_DEFAULT_NUMBER_OF_NETS];

    for (int j = (-angle*step); j <= (angle*step); j += angle) {
        if (j == 0) continue;
        rotateImage(j);

        int i;
        for (i = 0; i < FACE_DETECTION_DEFAULT_NUMBER_OF_NETS; i++) {
			rotatedFaces[i] = NULL;
            rotatedFaces[i] = new list<Face>();
		} // for
        
        searchFaces(rotatedFaces, j);
        
        for (i = 0; i < FACE_DETECTION_DEFAULT_NUMBER_OF_NETS; i++) {
            for (list<Face>::iterator rf = rotatedFaces[i]->begin(); rf != rotatedFaces[i]->end(); rf++)
                facesThroughNet[i]->push_back(*rf);
            
			// ???
            //rotatedFaces[i]->erase(rotatedFaces[i]->begin(), rotatedFaces[i]->end());
			if (rotatedFaces[i]!=NULL) {
				rotatedFaces[i]->clear();            
				delete rotatedFaces[i];
			} // if
			rotatedFaces[i]=NULL;
        } // for
    } // for
    return;
} // FaceDetection::searchInRotatedImage()
*/

void 
FaceDetection::detectFaces(FaceData& data) 
{

	// store output file
	string path1 = "";
	string path2 = "";
	if (param.outFile) {
		// get path of filename
		CreateMocaPath mocaPath (param.mpgFilename);
		path1 = mocaPath.destinationPath + "face_orig/";
		path2 = mocaPath.destinationPath + "face_box/";
		FileHandling::createDirectory (path1, false);
		FileHandling::createDirectory (path2, false);
	} // if
	Image<uint8> outImg;


	// analyze video
	list<Face> faces;
	cout << endl;
	int32 lastNumberOfFaces=-1;
    for (frameCounter=0; frameCounter <param.analyseFrame.getSize()[DIM_X]; frameCounter++) {

		if (param.analyseFrame(frameCounter)) {

			// report processing start time for each video frame
			if (param.verbose) {
				cout << "\rframe: " << frameCounter << endl;
			} // if
			if (lastNumberOfFaces != totalNumberOfFaces || frameCounter%100==1) {
				cerr << "\rframe: " << frameCounter 
					 << "    number of faces: " << totalNumberOfFaces 
					 << "    " << endl;
			} else {
				cerr << "\rframe: " << frameCounter << "    " << ends;
			} // else
			lastNumberOfFaces = totalNumberOfFaces;
			
			// analyze next image
			faces.clear();
			orgColorImage = video->getImage(frameCounter);
			processImage(orgColorImage, faces);


			FaceItem item;
			if (faces.size() == 0) {
				FaceItem emptyFace;
				emptyFace.frame=frameCounter;
				emptyFace.p1[DIM_X]=-1;
				emptyFace.p1[DIM_Y]=-1;
				emptyFace.p2[DIM_X]=-1;
				emptyFace.p2[DIM_Y]=-1;
				data.faces.push_back(emptyFace);
				data.numberOfFaces++;
			} else {
				
				// store original frame
				if (param.outFile) {
					outImg = orgColorImage;
					string fn = path1 + FileHandling::getNumberedFilename ("img", frameCounter, ".png");
					WriteImage (outImg, fn);
				} // if

				// store faces
				list<Face>::iterator it;
				for (it = faces.begin(); it != faces.end(); it++) {			
					item.angle = (*it).angle;
					item.frame = frameCounter;
					item.p1 [DIM_X] = (*it).x - (*it).width / 2;
					item.p1 [DIM_Y] = (*it).y - (*it).width / 2;
					item.p2 [DIM_X] = item.p1 [DIM_X] + (*it).width;
					item.p2 [DIM_Y] = item.p1 [DIM_Y] + (*it).width;
					data.faces.push_back(item);
					data.numberOfFaces++;
					
					if (param.outFile) {
						markFace (item, outImg);
					} // if
				} // for

				// store frame with face regions
				if (param.outFile) {
					string fn = path2 + FileHandling::getNumberedFilename ("img", frameCounter, ".png");
					WriteImage (outImg, fn);
				} // if
			} // else

			resetValues();
		} // if
    } // for
	cerr << "                         " << endl;
	return;
} // detectFaces

void 
FaceDetection::processVideo(FaceData& data) 
{

	// get path of filename, do not create new directory
	CreateMocaPath mocaPath (param.mpgFilename, false);
	data.filename = mocaPath.file;

	// read data
	data.clear();
	ReadMocaData<FaceData> reader (param.mpgFilename, FACE_DETECTION_FILENAME);
	bool dataValid = reader.process (FACE_DETECTION_VERSION, data); 

	bool dataError=true;
	if (dataValid) {

		// some frames may be already analysed and stored
		dataError = false;
		list<FaceItem>::iterator it;
		for (it=data.faces.begin(); it!=data.faces.end(); it++) {
			param.analyseFrame((*it).frame) = false;
		} // for

		// all frames already analysed?
		for (int32 i=0; i<param.analyseFrame.getSize()[DIM_X]; i++) {
			if (param.analyseFrame(i)) {
				dataValid = false;
				break;
			} // if
		} // for
	} // if

	if (!dataValid) {

		// create new index
		cerr << "FaceDetection: Creating index file ..." << endl;
		if (dataError) data.clear();
		data.filename = param.mpgFilename;
		detectFaces(data);
		data.faces.sort();
		data.numberOfFaces=data.faces.size();
			
		// save index file
		WriteMocaData<FaceData> writer (param.mpgFilename, FACE_DETECTION_FILENAME);
		writer.process(data);
		cerr << "\rFaceDetection: Creating index file ...done" << endl;
	} // if

    return;
} // FaceDetection::process



// the current angle is used for storage, the image is already rotated 
void 
FaceDetection::analyseImage (list<Face> &frameList, int32 currentAngle)
{
	// initialize
	if (param.verbose) {
		cout << endl << "rotation angle: " << currentAngle << endl;
	} // if
	frameList.clear();

	// search for faces
	searchFaces(facesThroughNet, currentAngle);

	for (int i = 0; i < param.numberOfNets; i++) {
		thresholding(2, 2, *(facesThroughNet[i]));
		if (param.verbose) {
			cout << "Net " << i << "\tno\tx\ty\twidth\tfaces\trotation\tweight" << endl;
		} // if
		overlapDetection(*(facesThroughNet[i]));
		for (list<Face>::iterator f = facesThroughNet[i]->begin(); f != facesThroughNet[i]->end(); f++) {
			frameList.push_back(*f);
		} // for
	} // for
	if (param.numberOfNets > 1) {
		thresholding(2, 1, frameList);
		if (param.verbose) {
			cout << "total\tx\ty\twidth\tfaces\trotation\tweight" << endl;
		} // if
		overlapDetection(frameList);
	} // if
	doReport(frameList, frameCounter);


	// store faces
	FaceItem item;
	if (frameList.size() > 0) {

		list<Face>::iterator it;
		for (it = frameList.begin(); it != frameList.end(); it++) {			
			item.angle = currentAngle;
			item.frame = frameCounter;
			item.p1 [DIM_X] = (*it).x - (*it).width / 2;
			item.p1 [DIM_Y] = (*it).y - (*it).width / 2;
			item.p2 [DIM_X] = item.p1 [DIM_X] + (*it).width;
			item.p2 [DIM_Y] = item.p1 [DIM_Y] + (*it).width;
			totalNumberOfFaces++;
		} // for
	} // else


	return;
}


void 
FaceDetection::processImage(const Image<uint8> & src, list<Face> &faces)
{
	// initialize
	faces.clear();
	time_t tt = time(NULL);
	if (param.verbose) {
		cout << "current time: " << ctime(&tt) << endl;
	} // if

	// store original image
	orgColorImage = src;

	// store original grayscale image
	GrayscaleFilter<uint8> gf;
	gf.process (orgColorImage, orgImage);


	// detect rotated faces 
	list<Face> frameList;
	list<Face>::iterator it;
    for (int32 currentAngle  = (-param.degree * param.rotationStep); 
	           currentAngle <= ( param.degree * param.rotationStep); 
		       currentAngle +=   param.degree) {
		
        // initialize
		frameList.clear();


		// nothing to do if rotation is zero
		shiftX=0;
		shiftY=0;
		if (currentAngle == 0) {
			image = orgImage;
			if (param.useColorInfo) {
				colorImage = orgColorImage;
			} // if
		} else {
			// rotate the image
			RotationFilter<uint8> rf (currentAngle * FACE_DETECTION_PI/180.0);
			rf.process (orgImage, image);

			// calculate shift factor
			ImageSize s0=orgImage.getSize();
			ImageSize s1=image.getSize();
			shiftX=(s1[DIM_X]-s0[DIM_X])/2;
			shiftY=(s1[DIM_Y]-s0[DIM_Y])/2;
			
			if (param.useColorInfo) {
				rf.process (orgColorImage, colorImage);
			} // if
		} // if
		
		// detect faces 
		analyseImage (frameList, currentAngle);
		for (it = frameList.begin(); it != frameList.end(); it++) {
			Face f = (*it);
			faces.push_back(f);
		} // for
		resetValues();
	} // for
	
	if (param.verbose) {
		cout << "overall time = " << (time(NULL) - tt) << " sec" << endl << endl;
	} // if
			
    return;
} // FaceDetection::processImg


void 
FaceDetection::markFace (FaceItem &item, Image<uint8> &img) 
{
	if (item.frame>-1 && item.p2[DIM_X]>0 && item.p2[DIM_Y]>0) {

		DrawLine line(Color::RED, 3);
		ieee64 angle = 	item.angle / 180.0 * MOCA_PI;

		int32 x0=(item.p1[DIM_X]+item.p2[DIM_X])/2;
		int32 y0=(item.p1[DIM_Y]+item.p2[DIM_Y])/2;
		
		
		// first corner
		ieee64 tx = item.p1[DIM_X] - x0;
		ieee64 ty = item.p1[DIM_Y] - y0;
		int32 x1 = (int32) (tx*cos(angle)-ty*sin(angle) + x0);
		int32 y1 = (int32) (tx*sin(angle)+ty*cos(angle) + y0);
		
		// second corner
		tx = item.p1[DIM_X] - x0;
		ty = item.p2[DIM_Y] - y0;
		int32 x2 = (int32) (tx*cos(angle)-ty*sin(angle) + x0);
		int32 y2 = (int32) (tx*sin(angle)+ty*cos(angle) + y0);
		
		// third corner
		tx = item.p2[DIM_X] - x0;
		ty = item.p1[DIM_Y] - y0;
		int32 x3 = (int32) (tx*cos(angle)-ty*sin(angle) + x0);
		int32 y3 = (int32) (tx*sin(angle)+ty*cos(angle) + y0);
		
		// forth corner
		tx = item.p2[DIM_X] - x0;
		ty = item.p2[DIM_Y] - y0;
		int32 x4 = (int32) (tx*cos(angle)-ty*sin(angle) + x0);
		int32 y4 = (int32) (tx*sin(angle)+ty*cos(angle) + y0);
		
		// mark face border
		line.process (x1,y1,x2,y2, img);
		line.process (x1,y1,x3,y3, img);
		line.process (x4,y4,x2,y2, img);
		line.process (x4,y4,x3,y3, img);
	} // if

	return;
}


void 
FaceDetection::storeFaces (FaceData& data, string &filename)
{
	// get path of filename, do not create new directory
	CreateMocaPath mocaPath (filename);
	string path1 = mocaPath.destinationPath+ "face_orig/";
	string path2 = mocaPath.destinationPath + "face_box/";
	FileHandling::createDirectory (path1, false);
	FileHandling::createDirectory (path2, false);

	VideoDecoder decoder (filename); 
	Image<uint8> img;

	int32 n = data.faces.size();
	int32 cnt=0;
	int32 lastFrame=-1;
	list<FaceItem>::iterator it;
	for (it=data.faces.begin(); it!=data.faces.end(); it++) {
		cout << "\rstore face: " << cnt << " / " << n << "  " << ends;

		FaceItem item = (*it);
		if (item.frame>-1 && item.p2[DIM_X]>0 && item.p2[DIM_Y]>0) {

			// get image
			int32 frame = item.frame;
			if (lastFrame != frame) {
				img = decoder.getImage (frame);
				lastFrame = frame;
			} // if

			// store original frame
			string fn = path1 + FileHandling::getNumberedFilename ("img", cnt, ".png");
			WriteImage (img, fn);

			// mark face and store frame 
			markFace (item, img);
			fn = path2 + FileHandling::getNumberedFilename ("img", cnt, ".png");
			WriteImage (img, fn);

		} // if
		cnt++;
	} // for

	return;
}

