/*
    GeometricFeature.cpp     calculate geometric features of an object in a binary image
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "GeometricFeature.h"

std::ostream& operator<< (ostream& os, const GeometricFeatureData &item)
{
	os << "compactness            : " << item.compactness << endl;
	os << "eccentricity           : " << item.eccentricity  << endl;
	os << "number of object pixels: " << item.numberObjectPixel<< endl;
	os << "contour length         : " << item.contourLength<< endl;

	os << "average x              : " << item.avgX << endl;
	os << "average y              : " << item.avgY << endl;
	os << "moment m20             : " << item.m20 << endl;
	os << "moment m02             : " << item.m02 << endl;
	os << "moment m11             : " << item.m11 << endl;
	
	return os;
} // operator<<

OTagStream& operator<< (OTagStream& os, const GeometricFeatureData& cm)
{
    TagFormat::writeStartTag(os, "GeometricFeature");
	os << endl;
	TagFormat::writeStartTag(os, "compactness",1);   os << cm.compactness;   TagFormat::writeEndTag(os, "compactness");
	TagFormat::writeStartTag(os, "eccentricity",1);     os << cm.eccentricity;     TagFormat::writeEndTag(os, "eccentricity");
	TagFormat::writeStartTag(os, "numberObjectPixel",1);      os << cm.numberObjectPixel;      TagFormat::writeEndTag(os, "numberObjectPixel");
	TagFormat::writeStartTag(os, "contourLength",1); os << cm.contourLength; TagFormat::writeEndTag(os, "contourLength");
	TagFormat::writeStartTag(os, "avgX",1);   os << cm.avgX;   TagFormat::writeEndTag(os, "avgX");
	TagFormat::writeStartTag(os, "avgY",1);   os << cm.avgY;   TagFormat::writeEndTag(os, "avgY");
	TagFormat::writeStartTag(os, "m20",1);  os << cm.m20;  TagFormat::writeEndTag(os, "m20");
	TagFormat::writeStartTag(os, "m02",1);  os << cm.m02;  TagFormat::writeEndTag(os, "m02");
	TagFormat::writeStartTag(os, "m11",1);  os << cm.m11;  TagFormat::writeEndTag(os, "m11");
    TagFormat::writeEndTag (os, "GeometricFeature");

    return os;
}

ITagStream& operator>> (ITagStream& is, GeometricFeatureData& cm)
{
    TagFormat::readStartTag(is, "GeometricFeature");
	TagFormat::readStartTag(is, "compactness");   is >> cm.compactness;   TagFormat::readEndTag(is, "compactness");
	TagFormat::readStartTag(is, "eccentricity");     is >> cm.eccentricity;     TagFormat::readEndTag(is, "eccentricity");
	TagFormat::readStartTag(is, "numberObjectPixel");      is >> cm.numberObjectPixel;      TagFormat::readEndTag(is, "numberObjectPixel");
	TagFormat::readStartTag(is, "contourLength"); is >> cm.contourLength; TagFormat::readEndTag(is, "contourLength");
	TagFormat::readStartTag(is, "avgX");   is >> cm.avgX;   TagFormat::readEndTag(is, "avgX");
	TagFormat::readStartTag(is, "avgY");   is >> cm.avgY;   TagFormat::readEndTag(is, "avgY");
	TagFormat::readStartTag(is, "m20");  is >> cm.m20;  TagFormat::readEndTag(is, "m20");
	TagFormat::readStartTag(is, "m02");  is >> cm.m02;  TagFormat::readEndTag(is, "m02");
	TagFormat::readStartTag(is, "m11");  is >> cm.m11;  TagFormat::readEndTag(is, "m11");
    TagFormat::readEndTag (is, "GeometricFeature");

    return is;
}




void 
GeometricFeature::process(const Image<uint8>& src, GeometricFeatureData& data) 
{

	data.compactness = 0.0;
	data.eccentricity  = 0.0;
	data.numberObjectPixel = 0;
	
	// binarize image
	Image<uint8> img;
	BinarizeFilter<uint8> bf(128);
	bf.process (src,img);
	
	// get largest object
	Image<uint8> largestObject;
	bool objectFound;
	RegionFeature<uint8> rf;
	rf.process(img, largestObject, objectFound);
	uint8 objectColor     = rf.getForegroundColor();
	uint8 backgroundColor = rf.getBackgroundColor();

	if (objectFound) {

		// count number of object pixels
		ImageSize size = img.getSize();
		int32 objPixel = 0;
		for (int32 y=0; y<size[DIM_Y]; y++) {
			for (int32 x=0; x<size[DIM_X]; x++) {
				if (img(x,y) == objectColor) {
					objPixel++;
				} // if
			} // for
		} // for
		data.numberObjectPixel = objPixel;


		// get contour
        vector<ImageEdgePosition> iep;
        ContourEdgeList contourEdge (backgroundColor);
        contourEdge.process (img, iep);
		data.contourLength =  contourEdge.getLength (iep);

		if (data.numberObjectPixel > 0) {
			data.compactness = (data.contourLength * data.contourLength) / 
				               (4 * MOCA_PI * data.numberObjectPixel);
		} // if

		// calculate center
		ieee64 avgX = 0.0;
		ieee64 avgY = 0.0;
		int32 n = iep.size();
		for (int32 i=0; i<n; i++) {
			avgX +=(iep[i])[DIM_X]; 
			avgY +=(iep[i])[DIM_Y]; 
		} // for
		if (n>0) {
			avgX /= ((ieee64)n);
			avgY /= ((ieee64)n);
		} // if
		data.avgX = avgX;
		data.avgY = avgY;

		// calculate moments
		ieee64 m20 = 0;
		ieee64 m02 = 0;
		ieee64 m11 = 0;
		for (int32 i=0; i<n; i++) {
			ieee64 dx = avgX - (iep[i])[DIM_X];
			ieee64 dy = avgY - (iep[i])[DIM_Y];

			m20 += dx*dx;
			m02 += dy*dy;
			m11 += dx*dy;
		} // for
		data.m02 = m02;
		data.m20 = m20;
		data.m11 = m11;

		// calculate eccentricity (the eigenvalues of matrix of the moments)
		// use pq equation
		ieee64 v0 = (m02+m20) / 2.0;
		ieee64 v1 = v0*v0 - m02*m20 + m11*m11;
		if (v1 >= 0) {
			ieee64 p0 = v0;
			ieee64 p1 = v0;
			if (v1 > 0) {
				v1 = sqrt (v1);
				p0 += v1;
				p1 -= v1;
			} // if
	
			ieee64 ecc = max(p0,p1) / min(p0,p1);
			if (ecc > 0) ecc = sqrt (ecc);
			data.eccentricity = ecc;
		} // if

	} // if


	return;
} // process


