/*
    LinearEstimatorFeature.cpp     LTS estimator
    Copyright (C) 2002 Dirk Farin

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



#include <LinearEstimatorFeature.h>

struct residual_data
{
	double residual;
	int    data_idx;
};


static int compdouble(const void* a,const void* b)
{
	residual_data* da = (residual_data*)a;
	residual_data* db = (residual_data*)b;
	
	if (da->residual < db->residual) return -1;
	else return 1;
}


LinearEstimatorFeature::LinearEstimatorFeature()
	: d_htries(100),
	  d_csteps(2),
	  d_trim_fact(0.6),
	  d_outlier(NULL),
	  d_residual(NULL)
{
}


void 
LinearEstimatorFeature::Estimate(vector<double>& parameters, double **data, int32 dataWidth, int32 dataHeight)
{
	int dim = dataWidth-1;
	int ndata = dataHeight;
	int trim_fact = (int)(ndata *d_trim_fact);
	double best_residue = -1;

	gsl_vector* rhs      = gsl_vector_alloc(dim);
	gsl_vector* solution = gsl_vector_alloc(dim);
	gsl_matrix* matrix   = gsl_matrix_alloc(dim,dim);
	
	gsl_matrix* cstep_matrix  = gsl_matrix_alloc(trim_fact,dim);
	gsl_vector* cstep_rhs     = gsl_vector_alloc(trim_fact);
	
	gsl_matrix* svd_Q  = gsl_matrix_alloc(dim,dim);
	gsl_vector* svd_S  = gsl_vector_alloc(dim);

	parameters.resize(dim);
	residual_data* residues = new residual_data[dataHeight];
	
	gsl_error_handler_t* old_err_handler = gsl_set_error_handler_off();


	for (int step=0;step<d_htries;step++) {
		for (int i=0;i<dim;i++) {
			int line = rand()%dataHeight;
			
			for (int j=0;j<dim;j++)
				gsl_matrix_set(matrix,i,j, data[line][j]);
			
			gsl_vector_set(rhs,i,data[line][dim]);
		} // for
		
		int gsl_err = gsl_linalg_HH_solve(matrix,rhs,solution);
		if (gsl_err) {
			// if matrix does not have full rank, try again
			
			if (gsl_err == GSL_ESING)
				continue;
			
			//AssertDescr(0, "unknown gsl error detected");
			cerr << "LinearEstimatorFeature::Estimate:unknown gsl error detected" << endl;
			exit(1);
		} // if


		for (int cstep=0;cstep<d_csteps;cstep++) {
			for (int i=0;i<ndata;i++) {
				residues[i].residual = 0;
				for (int c=0;c<dim;c++)
					residues[i].residual += gsl_vector_get(solution,c)*data[i][c];
				residues[i].residual -= data[i][dim];
				residues[i].residual *= residues[i].residual;
				residues[i].data_idx = i;
			} // for
			
			qsort(residues,ndata,sizeof(residual_data),compdouble);

			for (int i=0;i<ndata-1;i++) {
				//Assert(residues[i].residual <= residues[i+1].residual);
				if (!(residues[i].residual <= residues[i+1].residual)) {
					cerr << "LinearEstimatorFeature::Estimate:Assert1" << endl;
					exit(1);
				} // if
			} // for
			
			for (int i=0;i<trim_fact;i++) {
				for (int n=0;n<dim;n++)
					gsl_matrix_set(cstep_matrix, i,n, data[ residues[i].data_idx ][n]);
				
				gsl_vector_set(cstep_rhs, i, data[ residues[i].data_idx ][dim]);
			} // for
			
			gsl_linalg_SV_decomp_jacobi(cstep_matrix, svd_Q, svd_S);
			gsl_linalg_SV_solve(cstep_matrix, svd_Q, svd_S, cstep_rhs, solution);
		} // for

		for (int i=0;i<ndata-1;i++) {
			//Assert(residues[i].residual <= residues[i+1].residual);
			if (!(residues[i].residual <= residues[i+1].residual)) {
				cerr << "LinearEstimatorFeature::Estimate:Assert2" << endl;
				exit(1);
			} // if
		}

		double residue = 0.0;
		for (int i=0;i<trim_fact;i++)
			residue += residues[i].residual;

		
		if (best_residue < 0 || residue < best_residue) {
			best_residue = residue;
			
			for (int i=0;i<dim;i++)
				parameters[i] = gsl_vector_get(solution,i);
			
			if (d_outlier) {
				for (int i=0;i<ndata;i++)
					(*d_outlier)[i]=true;
				
				for (int i=0;i<trim_fact;i++) {
					(*d_outlier)[ residues[i].data_idx ] = false;
				} // for
			} // if
			
			if (d_residual) {
				for (int i=0;i<ndata;i++) {
					(*d_residual)[ residues[i].data_idx ]  = residues[i].residual;
				} // for
			} // if
		} // if
    } // for

  delete[] residues;

  gsl_vector_free(rhs);
  gsl_vector_free(solution);
  gsl_matrix_free(matrix);

  gsl_matrix_free(cstep_matrix);
  gsl_vector_free(cstep_rhs);

  gsl_matrix_free(svd_Q);
  gsl_vector_free(svd_S);

  gsl_set_error_handler(old_err_handler);
}

