/*
   This file is part of the MoCA video library.
   Name: LogoDetection.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  
   \author    Stephan Kopf
   
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de
	
	 (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	 
*/


#ifndef LOGO_DETECTION_H
#define LOGO_DETECTION_H

// C includes
#include <math.h>

// C++ includes
#include <iomanip>
#include <sstream>
#include <iostream>
using namespace std;


// MoCA includes
#include <MocaSignal.h>
#include <WriteMocaData.h>
#include <HistogramDifferenceFeature.h>
#include <VideoStatistic.h>
#include <GaussFilter.h>
#include <CannyFilter.h>
#include <BinarizeFilter.h>
#include <EcrFeature.h>
#include <FileHandling.h>
#include <MotionVector.h>
#include <DrawRectangle.h>


// windows includes
#ifdef COMPILER_MSVC
#include <process.h>
#endif


const static string LOGO_DETECTION_VERSION = "Version-0.16";
const static string LOGO_DETECTION_FILENAME= "logo.idx";

struct LogoDetectionBox
{
	LogoDetectionBox ()
	{
		minX = -1;
		maxX = -1;
		minY = -1;
		maxY = -1;
		rel  = 0.0; // reliability
	}
    LogoDetectionBox& operator=(const LogoDetectionBox &other) {
		if (this != &other) {
			minX = other.minX;
			maxX = other.maxX;
			minY = other.minY;
			maxY = other.maxY;
			rel  = other.rel;
		} // if
        return *this;
    } // operator=

	int32 minX, maxX, minY, maxY;
	ieee64 rel;
};

struct LogoDetectionData
{
	LogoDetectionData () 
	{
		version = LOGO_DETECTION_VERSION;
		filename = "";
		maxEdgeValue = 0;
	} 
    LogoDetectionData& operator=(const LogoDetectionData &other) {
		if (this != &other) {
			version         = other.version;
			filename        = other.filename;

			box[0]=other.box[0];
			box[1]=other.box[1];
			box[2]=other.box[2];
			box[3]=other.box[3];

			maxEdgeValue = other.maxEdgeValue;
		} // if
        return *this;
    } // operator=

	string version;
	string filename;

	// store up to 4 bounding boxes with test, logos, ...
	int32 maxEdgeValue;
	LogoDetectionBox box[4];
};

std::ostream& operator<< (ostream &, const LogoDetectionData&);
OTagStream& operator<< (OTagStream&, const LogoDetectionData&);
ITagStream& operator>> (ITagStream&, LogoDetectionData&);





/*
	Constructor: initializes variables
	
	process(): 
		if status file does not exist:
		   calculate the action in mpeg-file, save status file
		else:
		   read action data from status file
*/
class LogoDetection
{
public:

	LogoDetection(const string &fullName, bool storeFrame=false)
	{
		pathFilename = fullName;
		store = storeFrame;

		VideoDecoder decoder (pathFilename);
		size = decoder.getSize();
	}
	~LogoDetection() {};

	void process (LogoDetectionData&);

	// find the nearest border the sysmbol is attached to
	// int border: 0:left, 1:right, 2: top, 3: bottom
	void getBorder(LogoDetectionBox &box, int &borderNumber, int32 &distanceToBorder);

private:
	
	void readValues (LogoDetectionData& data);
	void summarizeLines(Image<uint8> &tmp, 
						Signal<int32> &sigX, Signal<int32> &sigSumX, Signal<int32> &sigY, Signal<int32> &sigSumY,
						int32 &lineMaxX, int32 &lineMaxY);
	void getColumnX (Image<uint8> &img, int32 lineMaxX, Signal<int32> &sigX, Signal<int32> &sigSumX, int32 avg, 
				     int32 &minX, int32 &minY, int32 &maxX, int32 &maxY);
	void getRowY (Image<uint8> &img, int32 lineMaxY, Signal<int32> &sigY, Signal<int32> &sigSumY, int32 avg, 
			      int32 &minX, int32 &minY, int32 &maxX, int32 &maxY);
	ieee64 getRatio (Image<uint8> &img, int32 minX, int32 minY, int32 maxX, int32 maxY);
	void storeFrame(Image<uint8> &img, LogoDetectionData& data);


	string pathFilename;
	bool store;
	VideoSize size;
};

#endif
