/*
   This file is part of the MoCA video library.
   Name: MocaFeatures.h
   Copyright (C) 2006 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#ifndef MOCA_FEATURES_H
#define MOCA_FEATURES_H

//
// just include this file, if you have to include many different features (headers) in your file
//
// include all header files in directory "features"
#include <ActionDetection.h>
#include <AudioCut.h>
#include <AudioEnergy.h>
#include <BlockMotionFeature.h>
#include <CameraBackgroundFeature.h>
#include <CameraModel.h>
#include <CameraOperation.h>
#include <CameraParameterEstimation.h>
#include <CameraTransformation.h>
#include <CameraTransformationPrecise.h>
#include <ContourEdgeList.h>
#include <ContourFeature.h>
#include <ContourParametrization.h>
#include <CornerDetection.h>
#include <CssDatabase.h>
#include <CssDisplayResult.h>
#include <CssImage.h>
#include <CssMatching.h>
#include <CssMaximaFeature.h>
#include <CssVideoAnalysis.h>
#include <CssVideoObject.h>
#include <CSS_DB_Echo.h>
#include <CSS_DB_Schema.h>
#include <CutDetectionFeature.h>
#include <EcrFeature.h>
#include <EdgeBasedContrastFeature.h>
#include <EdgebasedDistance.h>
#include <Face.h>
#include <FaceDetection.h>
#include <FaceRecognition.h>
#include <GeometricFeature.h>
#include <HistogramDifferenceFeature.h>
#include <HistogramFeature.h>
#include <HistoricVideoOptimization.h>
#include <ImageStatistic.h>
#include <LinearEstimatorFeature.h>
#include <LogoDetection.h>
#include <MinMaxFeature.h>
#include <MotionVector.h>
#include <ObjectMaskFeature.h>
#include <ObjectMaskShot.h>
#include <ObjectSegmentationFeature.h>
#include <ProjectionProfile.h>
#include <RegionbasedDistance.h>
#include <RegionFeature.h>
#include <SadFeature.h>
#include <SceneDetection.h>
#include <SceneDetectionDefaultScenes.h>
#include <SimpleFeatures.h>
#include <StructureMatchingFeature.h>
#include <TextDetection.h>
#include <TextFont.h>
#include <TextRegionDetection.h>
#include <TextSegmentation.h>
#include <TrackingFeature.h>
#include <VideoAbstractFeature.h>
#include <VideoStatistic.h>

// additional C includes
extern "C" {
#include "FaceData1.h"
#include "FaceData2.h"
#include "FaceData3.h"
#include "FaceData4.h"
#include "FaceData5.h"
}

void testFeatures ()
{
	// define path with test images
	string libMocaPath="../";  // linux path
#ifdef WIN32
	libMocaPath="../../../../libmoca/";  // windows path
#endif

	// global variables
	string fn=libMocaPath + "doc/lena.ppm";
	Image<uint8> img;
	ReadImage (fn, img);


	// get corners of an image
	vector<CornerPosition> corners;
	CornerDetection cd;
	cd.process (img, corners);

	// get action of a video
	fn=libMocaPath + "doc/test.avi";
	ActionData aData;
	ActionDetection ad (fn);
	ad.process(aData);

	// get audio cuts
	fn=libMocaPath + "doc/test.wav";
	int32 minimumNumberOfCuts = 1;
	int32 blockLengthMilliSeconds = 20;
	AudioCut ac (fn, minimumNumberOfCuts, blockLengthMilliSeconds);
	AudioCutData acData;
	ac.process (acData);

	// get audio energy
	fn=libMocaPath + "doc/test.wav";
	AudioEnergy ae (fn);
	AudioEnergyData aeData;
	ae.process (aeData);

	// get motion of a block
	int32 blockSize=8;
	int32 searchRange=16;
	Image<int8> motionvectors;
	BlockMotionFeature bmf(blockSize, searchRange);
	bmf.process (img, img, motionvectors);

	// construct an background camera image
	VideoSize srcVideoSize (100,100,3,10);
	int32 bgSizeX=120;
	int32 bgSizeY=120;
	int32 bgShiftX=10;
	int32 bgShiftY=10;
	bool calculateObjects=false;
	bool precise=false;
	bool estimateObjectLocation=false;
	Video<uint8> vid (srcVideoSize);
	vid.clear(255);
	Video<uint8> motionMaskVid;
	Image<uint8> bgImg;
	vector<TransformationData> vidTrans;
	CameraBackgroundFeature cbf (srcVideoSize, bgSizeX, bgSizeY, bgShiftX, bgShiftY, 
								 calculateObjects, precise, estimateObjectLocation);
	cbf.process(vid, motionMaskVid, bgImg, vidTrans);

	// calculate camera model
	CameraModelData cmData;
	CameraModel cm;
	cm.process (cmData);


	// calculate camera operation
	CameraOperationData coData;
	CameraOperation co(fn);
	co.process(coData);


	// calculate camera parameters
	Image<ieee64> srcVectors;
	Vector<ieee64> parameters (10);
	CameraParameterEstimation cpe;
	cpe.process(srcVectors, parameters);

	// calculate camera transformation
	TransformationData trans;
	PerspectiveTransformation pt;
	pt.processColorImg (trans, img, img);

	// calculate precise camera transformation
	TransformationData srcDestCameraParameters;
	ieee64 minErr;
	CameraTransformationPrecise ctp;
	ctp.process(img, img, img, srcDestCameraParameters, minErr);

	// calculate contour
	vector<ImageEdgePosition> iepos;
	ContourEdgeList cel;
	cel.process(img, iepos);

	// calculate contour
	ContourFeature cf;
	ContourFeatureData cfData;
	cf.process(img, cfData);

	// parametrize contour
	vector<ImageEdgePosition> iep;
	vector<ImagePosition> ip;
	ContourParametrization cp (200);
	cp.process(iep, ip);

	// open css database
	vector<CssStoredDatabaseItem> cssData;
	vector<string> objectClassName;
	CssDatabaseSchemaType schema=ECHO_DB_TYPE;
	CssDatabase cssD (schema);
	cssD.getDatabase (cssData, objectClassName);

	// dusplay css results
	vector<FileInfo> databaseList;
	vector<FileInfo> sequenceList;
	bool showStatus=true;
	bool extended=false;
	CssDisplayResult cssDR (databaseList, sequenceList, showStatus, extended);
	cssDR.showAll(5);

	// TODO: add all other classes 

#if 0
#include <CssImage.h>
#include <CssMatching.h>
#include <CssMaximaFeature.h>
#include <CssVideoAnalysis.h>
#include <CssVideoObject.h>
#include <CSS_DB_Echo.h>
#include <CSS_DB_Schema.h>
#include <CutDetectionFeature.h>
#include <EcrFeature.h>
#include <EdgeBasedContrastFeature.h>
#include <EdgebasedDistance.h>
#include <Face.h>
#include <FaceDetection.h>
#include <FaceRecognition.h>
#include <GeometricFeature.h>
#include <HistogramDifferenceFeature.h>
#include <HistogramFeature.h>
#include <HistoricVideoOptimization.h>
#include <ImageStatistic.h>
#include <LinearEstimatorFeature.h>
#include <LogoDetection.h>
#include <MinMaxFeature.h>
#include <MotionVector.h>
#include <ObjectMaskFeature.h>
#include <ObjectMaskShot.h>
#include <ObjectSegmentationFeature.h>
#include <ProjectionProfile.h>
#include <RegionbasedDistance.h>
#include <RegionFeature.h>
#include <SadFeature.h>
#include <SceneDetection.h>
#include <SceneDetectionDefaultScenes.h>
#include <SimpleFeatures.h>
#include <StructureMatchingFeature.h>
#include <TextDetection.h>
#include <TextFont.h>
#include <TextRegionDetection.h>
#include <TextSegmentation.h>
#include <TrackingFeature.h>
#include <VideoAbstractFeature.h>
#include <VideoStatistic.h>

// additional C includes
extern "C" {
#include "FaceData1.h"
#include "FaceData2.h"
#include "FaceData3.h"
#include "FaceData4.h"
#include "FaceData5.h"
#endif

	return;
}
    

#endif

