/*
    MotionVector.cpp     detects motion vectors
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// MoCA includes
#include <MotionVector.h>



std::ostream& operator<< (ostream& os, const MotionVectorStatus &data)
{
	os << "version    : " << data.version << endl;
	os << "filename   : " << data.filename<< endl;
	os << "image size : " << data.size[DIM_X] << "x" << data.size[DIM_Y] << endl;
	os << "frames     : " << data.size[DIM_Z] << endl;
	os << "blocksize  : " << data.blockSize << endl;
	os << endl;

	return os;
} // operator<<

OTagStream& operator<< (OTagStream& os, const MotionVectorStatus& data)
{
	TagFormat::writeNodeValue(os, "version", data.version);
	TagFormat::writeNodeValue(os, "filename", data.filename);
	TagFormat::writeNodeValue(os, "DIM_X", data.size[DIM_X]);
	TagFormat::writeNodeValue(os, "DIM_Y", data.size[DIM_Y]);
	TagFormat::writeNodeValue(os, "numberOfFrames", data.size[DIM_Z]);
	TagFormat::writeNodeValue(os, "blocksize", data.blockSize);

	return os;
} // operator<<


ITagStream& operator>> (ITagStream& is, MotionVectorStatus& data)
{
	char c;
	//bool failFlag = false;
	
	if (is >> c) {
		is.putback(c);
		
		TagFormat::readNodeValue(is, "version", data.version);
		TagFormat::readNodeValue(is, "filename", data.filename);
		TagFormat::readNodeValue(is, "DIM_X", data.size[DIM_X]);
		TagFormat::readNodeValue(is, "DIM_Y", data.size[DIM_Y]);
		TagFormat::readNodeValue(is, "numberOfFrames", data.size[DIM_Z]);
		TagFormat::readNodeValue(is, "blocksize", data.blockSize);
		if (is.rdstate() & ios::failbit) {
			throw MocaException ("MotionVectorStatus: Illegal index file format!");
		} // if
		data.size[DIM_BAND]=3;
	} // if
	
	if (is.rdstate() & ios::failbit) {
		throw MocaException ("MotionVectorStatus: Illegal index file format!");
	} else {
		is.clear();
	} // else
	
	return is;
} // operator>> 




std::ostream& operator<< (ostream& os, const MotionVectorBuffer &data)
{
	os << "data:" << endl;
	for (int32 i=0; i<data.size[DIM_Z]; i++) {
		for (int32 y=0; y<data.size[DIM_Y]; y++) {
			for (int32 x=0; x<data.size[DIM_X]; x++) {
				os << "(" << x << "," << y << "," << i << ")  " 
				   << data.data(x,y,i,0) << "\t" 
				   << data.data(x,y,i,1) << "\t" 
				   << data.data(x,y,i,2) << "\t" << endl;
			} // for
		} // for
	} // for
	
	return os;
} // operator<<

OTagStream& operator<< (OTagStream& os, const MotionVectorBuffer& data)
{
	for (int32 i=0; i<data.size[DIM_Z]; i++) {
		for (int32 y=0; y<data.size[DIM_Y]; y++) {
			for (int32 x=0; x<data.size[DIM_X]; x++) {
				os << setfill(' ')  << setw(7) << i+data.offset << "     " 
				   << setfill(' ')  << setw(5) << (int)data.data(x,y,i,0) << "     " 
				   << setfill(' ')  << setw(5) << (int)data.data(x,y,i,1) << "     " 
				   << setfill(' ')  << setw(5) << (int)data.data(x,y,i,2) << endl;
			} // for
		} // for
		os << endl;
	} // for

	return os;
} // operator<<


ITagStream& operator>> (ITagStream& is, MotionVectorBuffer& data)
{
	char c;
	//bool failFlag = false;
	
	if (is >> c) {
		is.putback(c);
		
		for (int32 i=0; i<data.size[DIM_Z]; i++) {
			for (int32 y=0; y<data.size[DIM_Y]; y++) {
				for (int32 x=0; x<data.size[DIM_X]; x++) {
					int32 tmp;
					is >> tmp;  // frame number
					/*
					is >> data.data(x,y,i,0);
					is >> data.data(x,y,i,1);
					is >> data.data(x,y,i,2);
					*/
					int tmp1;
					is >> tmp1; data.data(x,y,i,0) = tmp1;
					is >> tmp1; data.data(x,y,i,1) = tmp1;
					is >> tmp1; data.data(x,y,i,2) = tmp1;
				} // for
			} // for
			if (is.rdstate() & ios::failbit) {
				throw MocaException ("MotionVectorBuffer: Illegal index file format!");
			} // if
		} // for
	} // if
	
	if (is.rdstate() & ios::failbit) {
		throw MocaException ("MotionVectorBuffer: Illegal index file format!");
	} else {
		is.clear();
	} // else
	
	return is;
} // operator>> 





MotionVector::MotionVector(const string &pathFilename, int32 searchRange) 
{
	this->pathFilename = pathFilename;
	this->searchRange = searchRange;
	bufferOffset = -1;
	motionIts = NULL;
	lastOffset = -1;

	// read data
	process(status);
} 


void 
MotionVector::getMotionVector(int32 frame, Image<ieee64> &data)
{
	if (frame<0) {
		frame=0;
	} // if
	if (frame>=status.size[DIM_Z]) {
		frame=status.size[DIM_Z]-1;
	} // if

	bool readValues=false;
	if (bufferOffset==-1) readValues=true;
	if (frame < bufferOffset) readValues=true;
	if (frame >= bufferOffset+DEFAULT_BUFFER_SIZE) readValues=true;

	if (readValues) {
		bufferOffset = frame - frame%DEFAULT_BUFFER_SIZE;

		// initialize buffer
		if (buffer.size[DIM_X]==0) {
			buffer.size[DIM_X]=status.size[DIM_X];
			buffer.size[DIM_Y]=status.size[DIM_Y];
			buffer.size[DIM_Z]=DEFAULT_BUFFER_SIZE;
			buffer.size[DIM_BAND]=3;
			buffer.data.resize(buffer.size);
		} // if

		if ((motionIts==NULL) || (frame<lastOffset) || (lastOffset==-1)) {
			if (motionIts!=NULL) {
				delete motionIts;
			} // if
			motionIts = NULL;
			CreateMocaPath mocaPath (pathFilename);
			motionIts = new ITagStream (mocaPath.destinationPath + MOTION_VECTOR_DATA_FILENAME);
			lastOffset = 0;
		} // if

		while (lastOffset<=frame) {
			(*motionIts) >> buffer;
			lastOffset+=DEFAULT_BUFFER_SIZE;
		} // while
	} // if

	int32 index=frame%DEFAULT_BUFFER_SIZE;
	ImageSize dataSize(buffer.size[DIM_X], buffer.size[DIM_Y], 3);
	data.resize(dataSize);
	for (int32 b=0; b<3; b++) {
		for (int32 y=0; y<dataSize[DIM_Y]; y++) {
			for (int32 x=0; x<dataSize[DIM_X]; x++) {
				data(x,y,b) = buffer.data(x,y,index,b);
			} // for
		} // for
	} // for

	return;
} // getMotionVector


void 
MotionVector::process(MotionVectorStatus& data) 
{
	// read data
	ReadMocaData<MotionVectorStatus> reader (pathFilename, MOTION_VECTOR_STATUS_FILENAME);
	bool dataValid = reader.process (MOTION_VECTOR_VERSION, data); 

	if (!dataValid) {

		// create new index
		cerr << "\rMotionVector: Creating index file ...           " << endl;
		readValues(data);
			
		// save index file
		WriteMocaData<MotionVectorStatus> writer (pathFilename, MOTION_VECTOR_STATUS_FILENAME);
		writer.process(data);
		cerr << "\rMotionVector: Creating index file ...done           " << endl;
	} // if

	return;
} // process


// compare the motion vectors of 2 following frames of a video:
// summarize absolute difference of motion vectors
void 
MotionVector::getSAD (Signal<ieee64> &data)
{
	SignalSize size (status.size[DIM_Z]);
	data.resize (size);
	data.clear(0);

	Image<ieee64> mv1;
	Image<ieee64> mv2;
	getMotionVector(0, mv1);
	ImageSize mvSize = mv1.getSize();
	ieee64 factor = mvSize[DIM_X] * mvSize[DIM_Y];
	if (factor < 1.0) factor = 1.0;
	factor = 100.0 / (64.0 * factor);

	for (int32 i=0; i<size[DIM_X]-1; i++) {

		getMotionVector(i, mv1);
		getMotionVector(i+1, mv2);

		ieee64 diff=0.0;
		for (int32 y=0; y<mvSize[DIM_Y]; y++) {
			for (int32 x=0; x<mvSize[DIM_X]; x++) {
				//diff += abs (mv1(x,y,0) - mv2(x,y,0)) + abs (mv1(x,y,1) - mv2(x,y,1));
				diff += fabs (mv1(x,y,0) - mv2(x,y,0));
				diff += fabs (mv1(x,y,1) - mv2(x,y,1));
			} // for
		} // for
		data(i+1) = (diff * factor);
		if (data(i+1) < 0)   {
			cerr << "error in MotionVector::getSAD(): (data(i+1) < 0)" << endl << endl;
			data(i+1) = 0;
		}
		if (data(i+1) > 255) {
			data(i+1) = 255;
			cerr << "error in MotionVector::getSAD(): (data(i+1) > 255)" << endl << endl;
		} // if

	} // for

	return;
} // getSAD 



void 
MotionVector::readValues (MotionVectorStatus& status) 
{
	// get video size
	VideoCodec codec = CODEC_Y; 
	int cacheSize=3;
	int32 lastFrameToDecode=-1;
	VideoDecoder reader (pathFilename, cacheSize, lastFrameToDecode, codec);

	// set values
	status.version = MOTION_VECTOR_VERSION;
	status.size[DIM_X]=0;
	status.size[DIM_Y]=0;
	status.size[DIM_BAND]=3;
	status.size[DIM_Z]=reader.getSize()[DIM_Z] - 1;
	status.blockSize = 16;

	// get path of filename
	CreateMocaPath mocaPath (pathFilename);
	status.filename = mocaPath.file;


	Image<uint8> img[2];
	int firstImg = 0;
	int nextImg  = 1;
	img[firstImg] = reader.getImage(0);
	cerr << "\rframe: 0     " << flush;
	int32 bufOffset=0;
	buffer.offset=bufOffset;

	// store motion vectors
	OTagStream ots (mocaPath.destinationPath + MOTION_VECTOR_DATA_FILENAME);


	Image<int8> vec;
	//MotionVectorSearchMode searchMode=THREE_STEP_SEARCH_SCALED;
	BlockMotionFeature bmf (status.blockSize, searchRange/*, searchMode*/);
	for (int i=1; i< status.size[DIM_Z]+1; i++) {
		
		// store already calculated motion vectors
		if (i-1-bufOffset>=DEFAULT_BUFFER_SIZE) {

			// store buffer
			ots << buffer;
			bufOffset+=DEFAULT_BUFFER_SIZE;
			buffer.offset=bufOffset;

			// clear buffer 
			for (int32 tz=0; tz<DEFAULT_BUFFER_SIZE; tz++) {
				for (int32 ty=0; ty<buffer.size[DIM_Y]; ty++) {
					for (int32 tx=0; tx<buffer.size[DIM_X]; tx++) {
						buffer.data(tx,ty,tz,0)=0;
						buffer.data(tx,ty,tz,1)=0;
						buffer.data(tx,ty,tz,2)=0;
					} // for
				} // for
			} // for

		} // if

		// read next image
		img[nextImg] = reader.getImage(i);
		cerr << "\rframe: " << i << "   " << flush;

		// get motion vectors
		bmf.process(img[firstImg], img[nextImg], vec);

		// data size initialized?
		if (status.size[DIM_X]==0) {
			status.size[DIM_X]=vec.getSize()[DIM_X];
			status.size[DIM_Y]=vec.getSize()[DIM_Y];

			//data.data.resize(data.size);
			buffer.size[DIM_X]=status.size[DIM_X];
			buffer.size[DIM_Y]=status.size[DIM_Y];
			buffer.size[DIM_Z]=DEFAULT_BUFFER_SIZE;
			buffer.size[DIM_BAND]=3;
			buffer.data.resize(buffer.size);

			for (int32 tz=0; tz<DEFAULT_BUFFER_SIZE; tz++) {
				for (int32 ty=0; ty<buffer.size[DIM_Y]; ty++) {
					for (int32 tx=0; tx<buffer.size[DIM_X]; tx++) {
						buffer.data(tx,ty,tz,0)=0;
						buffer.data(tx,ty,tz,1)=0;
						buffer.data(tx,ty,tz,2)=0;
					} // for
				} // for
			} // for
		} // if

		for (int32 y=0; y<status.size[DIM_Y]; y++) {
			for (int32 x=0; x<status.size[DIM_X]; x++) {
				buffer.data(x,y,i-1-bufOffset,0) = vec(x,y,0);
				buffer.data(x,y,i-1-bufOffset,1) = vec(x,y,1);

				// (read scaled value)  scale value (average block distance) to 0..1
				buffer.data(x,y,i-1-bufOffset,2) = vec(x,y,2); // / (255.0*status.blockSize*status.blockSize);
			} // for
		} // for


		// update image position
		firstImg=nextImg;
		nextImg =1-firstImg;
	} // for

	// store buffer
	ots << buffer;
	bufOffset+=DEFAULT_BUFFER_SIZE;
	buffer.offset=bufOffset;
	
	// clear buffer 
	for (int32 tz=0; tz<DEFAULT_BUFFER_SIZE; tz++) {
		for (int32 ty=0; ty<buffer.size[DIM_Y]; ty++) {
			for (int32 tx=0; tx<buffer.size[DIM_X]; tx++) {
				buffer.data(tx,ty,tz,0)=0;
				buffer.data(tx,ty,tz,1)=0;
				buffer.data(tx,ty,tz,2)=0;
			} // for
		} // for
	} // for

	//cerr << endl;

	// close motion vector data file
	ots.close();

	return;
} // readValues




std::ostream& operator<< (ostream& os, const MotionParametrizationData &data)
{
//cout.setf(ios::fixed);
//cout.precision(2);

	os << "a:"  << ios::fixed << setprecision(4) << data.a  << "   ";
	os << "b:"  << ios::fixed << setprecision(4) << data.b  << "   ";
	os << "c:"  << ios::fixed << setprecision(4) << data.c  << "   ";
	os << "d:"  << ios::fixed << setprecision(4) << data.d  << "   ";
	os << "dx:" << ios::fixed << setprecision(3) << data.dx << "  ";
	if (data.dx >=0.0) os << " ";
	os << "dy:" << ios::fixed << setprecision(3) << data.dy << endl;

	return os;
} // operator<<



MotionParametrization::MotionParametrization(int32 macroBlockSize, const ImageSize &size) 
{
	this->macroBlockSize=macroBlockSize;
	imgSize = size;
} 


void leastSquareLinear (Signal<ieee64> &data, ieee64 &a, ieee64 &b, ieee64 &error)
{	
	// data stores values: data(i,0) stores x values, data(i,1) stores y values
	// 
	// function, where the parameters should be approximated: y=a+b*x
	// 
	// X = (SUM x) / n
	// Y = (SUM y) / n
	//
	// B= ((SUM y*y - n*Y*Y) - (SUM x*x - n*X*X))  / 2*(n*X*Y - SUM x*y)
	// 
	// b=-B +- SQRT(B*B+1)
	// a = Y - bX
	//
	// b1=-B + SQRT(B*B+1)
	// b2=-B - SQRT(B*B+1)
	// a1 = Y - b1*X
	// a2 = Y - b2*X
	a=0.0;
	b=0.0;
	error=0.0;
	int32 n = data.getSize()[DIM_X];
	if (n<=1) return;

	ieee64 avgX = 0.0;
	ieee64 avgY = 0.0;
	ieee64 sumXX= 0.0;
	ieee64 sumYY = 0.0;
	ieee64 sumXY = 0.0;
	for (int32 i=0; i<n; i++) {
		ieee64 x = data(i,0);
		ieee64 y = data(i,1);

		avgX+=x;
		avgY+=y;
		sumXX+=x*x;
		sumYY+=y*y;
		sumXY+=x*y;
	} // for
	avgX/=n;
	avgY/=n;

	ieee64 B1 = (sumYY - n*avgY*avgY) - (sumXX - n*avgX*avgX);
	ieee64 B2 = 2*(n*avgX*avgY - sumXY);
	if (B2 == 0.0) return;
	ieee64 B= B1 / B2;
	
	ieee64 b1 = -B + sqrt(B*B+1);
	ieee64 b2 = -B - sqrt(B*B+1);
	ieee64 a1 = avgY - b1*avgX;
	ieee64 a2 = avgY - b2*avgX;

	ieee64 diff1=0.0;
	ieee64 diff2=0.0;
	for (int32 i=0; i<n; i++) {
		ieee64 x = data(i,0);
		ieee64 y = data(i,1);
		ieee64 v1 = a1 + x*b1;
		ieee64 v2 = a2 + x*b2;

		diff1 += abs(v1-y);
		diff2 += abs(v2-y);
	} // for
	if (diff1<diff2) {
		a=a1;
		b=b1;
		error=diff1;
	} else {
		a=a2;
		b=b2;
		error=diff2;
	} // else

	return;
} // leastSquareLinear 


void 
MotionParametrization::getMotionParameters (Image<ieee64> &motionVectors, MotionParametrizationData& data)
{
	ImageSize motionVectorSize = motionVectors.getSize();


	// get translation (pan,tilt)
	ieee64 realibility = 0.0;
	ieee64 totalRealibility = 0.0;
	ieee64 avgX=0;
	ieee64 avgY=0;
	for (int32 y=0; y< motionVectorSize[DIM_Y]; y++) {
		for (int32 x=0; x< motionVectorSize[DIM_X]; x++) {
			realibility = motionVectors (x,y,2);
			totalRealibility += realibility;
			avgX += realibility*motionVectors (x,y,0);
			avgY += realibility*motionVectors (x,y,1);
		} // for
	} // for
	if (totalRealibility < 1.0) totalRealibility = 1.0;
	avgX /= totalRealibility;
	avgY /= totalRealibility;
	data.dx = avgX;
	data.dy = avgY;


	// remove translation from data
	for (int32 y=0; y< motionVectorSize[DIM_Y]; y++) {
		for (int32 x=0; x< motionVectorSize[DIM_X]; x++) {
			//cout << "old: " << motionVectors (x,y,0) << "  " << motionVectors (x,y,1) << "  ";
			motionVectors (x,y,0) -= avgX;
			motionVectors (x,y,1) -= avgY;
			//cout << "new: " << motionVectors (x,y,0) << "  " << motionVectors (x,y,1) << endl;
		} // for
	} // for
	


	// get zoom in/out
	// get average values for collumns 
	int32 motionVectorSizeX = motionVectorSize[DIM_X];
	int32 motionVectorSizeY = motionVectorSize[DIM_Y];
	Image<ieee64> tmpMotionVectors = motionVectors;
	ieee64 a1,b1;    // y=a+bx
	ieee64 e1=10e10;
	int32 c=0;
	while (e1>2.0) {
		c++;
		if (c>motionVectorSizeX/2) break;

		// put values to sigX
		Signal<ieee64> sigX (motionVectorSizeX);
		for (int32 x=0; x< motionVectorSizeX; x++) {
			sigX(x) = 0.0;
			totalRealibility = 0.0;
			for (int32 y=0; y< motionVectorSizeY; y++) {
				realibility = motionVectors (x,y,2);
				totalRealibility += realibility;
				sigX(x) += realibility*motionVectors (x,y,0);
			} // for
			if (totalRealibility < 1.0) totalRealibility = 1.0;
			sigX(x) /= totalRealibility;
		} // for

		// calculate linear function and get error e1
		SignalSize fSize (motionVectorSizeX,2);
		Signal<ieee64> f (fSize);
		for (int32 i=0; i<fSize[DIM_X];i++) {
			f(i,0)=i;
			f(i,1)=sigX(i);
		} // for
		leastSquareLinear (f, a1, b1,e1);
		e1/=motionVectorSizeX;

		// check, if error is too large
		if (e1>2.0) {

			// get largest difference from function
			int32 maxPosition=0;
			ieee64 maxDiff=0.0;
			ieee64 diff;
			for (int32 i=0; i<motionVectorSizeX; i++) {
				ieee64 x = f(i,0);
				ieee64 y = f(i,1);
				ieee64 v = a1 + x*b1;
				
				diff = abs(v-y);
				if (maxDiff<diff) {
					maxDiff=diff;
					maxPosition=i;
				} // if
			} // for

			// remove worst motion vector
			motionVectors(maxPosition,2) =0.0;
		} // if
	} // while



	// get zoom in/out
	// get average values for rows
	motionVectors = tmpMotionVectors;
	ieee64 a2,b2;    // y=a+bx
	ieee64 e2=10e10;
	c=0;
	while (e2>2.0) {
		c++;
		if (c>motionVectorSizeY/2) break;

		// put values to sigY
		Signal<ieee64> sigY (motionVectorSizeY);
		for (int32 y=0; y< motionVectorSizeY; y++) {
			sigY(y) = 0.0;
			totalRealibility = 0.0;
			for (int32 x=0; x< motionVectorSizeX; x++) {
				realibility = motionVectors (x,y,2);
				totalRealibility += realibility;
				//cout << motionVectors (x,y,1) << "  " << ends;
				sigY(y) += realibility*motionVectors (x,y,1);
			} // for
			if (totalRealibility < 1.0) totalRealibility = 1.0;
			sigY(y) /= totalRealibility;
			//cout << "sigY: " << sigY(y) << endl;
		} // for

		// calculate linear function and get error e2
		SignalSize fSize (motionVectorSizeY,2);
		Signal<ieee64> f (fSize);
		for (int32 i=0; i<fSize[DIM_X];i++) {
			f(i,0)=i;
			f(i,1)=sigY(i);
		} // for
		leastSquareLinear (f, a2, b2,e2);
		e2/=motionVectorSizeY;

		// check, if error is too large
		if (e2>2.0) {

			// get largest difference from function
			int32 maxPosition=0;
			for (int32 i=0; i< motionVectorSizeY; i++) {
				cout << f(i,0) << "  " << f(i,1) << "  " << sigY(i) << endl;
			} // for
			


			ieee64 maxDiff=0.0;
			ieee64 diff;
			for (int32 i=0; i<motionVectorSizeY; i++) {
				ieee64 x = f(i,0);
				ieee64 y = f(i,1);
				ieee64 v = a1 + x*b1;
				
				diff = abs(v-y);
				if (maxDiff<diff) {
					maxDiff=diff;
					maxPosition=i;
				} // if
			} // for

			// remove worst motion vector
			motionVectors(maxPosition,2) =0.0;
		} // if
	} // while


	// exact parameters based on image
	a1-=b1/2.0;
	a2-=b2/2.0;
	b1/= macroBlockSize;
	b2/= macroBlockSize;
	

	// average error is more than 3 pixel
	bool zoomValid=true;
	if (e1>3.0) zoomValid=false;
	if (e2>3.0) zoomValid=false;


	// calculate data.a and data.d
	ieee64 n_a = a1 + b1 * imgSize[DIM_X] + (imgSize[DIM_X] / 2.0);
	n_a /= (imgSize[DIM_X] / 2.0);
	ieee64 n_d = a2 + b2 * imgSize[DIM_Y] + (imgSize[DIM_Y] / 2.0);
	n_d /= (imgSize[DIM_Y] / 2.0);
	
	// how much motion is in the real image with these values?
	// get maximum zoom-motion
	ieee64 maxMotion = 0.0;
	if (maxMotion < fabs((1.0-n_a)*imgSize[DIM_X]/2.0)) maxMotion = fabs((1.0-n_a)*imgSize[DIM_X]/2.0);
	if (maxMotion < fabs((1.0-n_d)*imgSize[DIM_Y]/2.0)) maxMotion = fabs((1.0-n_d)*imgSize[DIM_Y]/2.0);
	if (maxMotion<3.0) zoomValid=false;

	// zoom in (x-direction) and zoom out (y-direction)?
	if (b1*b2 <= 0.0) zoomValid=false;
	ieee64 th = 0.002 / macroBlockSize;
	if (fabs(b1) < th) zoomValid=false;
	if (fabs(b2) < th) zoomValid=false;

	// store parameters
	if (zoomValid) {

		// zoom in
		if ((b1>th) && (b2>th)) {
			//cout << "zoom in: " << n1 << "   "  << n2 << endl;
			data.a=(2.0*n_a+1.0*n_d)/3.0;
			data.d=(1.0*n_a+2.0*n_d)/3.0;
		} // if

		// zoom out
		if ((b1<-th) && (b2<-th)) {
			//cout << "zoom out: " << n1 << "    "  << n2 << endl;
			data.a=(2.0*n_a+1.0*n_d)/3.0;
			data.d=(1.0*n_a+2.0*n_d)/3.0;
		} // if
	} // if


	return;
} // getMotionParameters

void 
MotionParametrization::process (Image<ieee64> &motionVectors, MotionParametrizationData& data)
{
	data.a = 1.0;
	data.b = 0.0;
	data.c = 0.0;
	data.d = 1.0;
	data.dx= 0.0;
	data.dy= 0.0;

	// calculate motion vectors
	Image<ieee64> tmpData = motionVectors;
	getMotionParameters (tmpData, data);
	cerr <<"param[1]: " << data;


	// optimize and validate parameters
	// check, if each motion vector fits to the modell, remove the motion vectors, which do not fit 
	// (1) vectors at the border of the image
	// (2) vectors with significant difference
	// ==> set tmpData(x,y,2)=0
	// set threshold to 25 (mv in x or y direction has 5 or more pixels difference)
	tmpData = motionVectors;
	ImageSize motionVectorSize=motionVectors.getSize();
	bool valuesModified=false;
	for (int32 y=0; y< motionVectorSize[DIM_Y]; y++) {
		for (int32 x=0; x< motionVectorSize[DIM_X]; x++) {

	//int32 t1 = (int32)tmpData(x,y,0);
	//int32 t2 = (int32)tmpData(x,y,1);
	//int32 t3 = (int32)tmpData(x,y,2);
	//int32 t4 = (int32)tmpData(x+1,y,0);
	//int32 t5 = (int32)tmpData(x+1,y,1);
	//int32 t6 = (int32)tmpData(x+1,y,2);
	//int32 t7 = (int32)tmpData(x,y+1,0);
	//int32 t8 = (int32)tmpData(x,y+1,1);
	//int32 t9 = (int32)tmpData(x,y+1,2);


			int32 xApprox = (int32)tmpData(x,y,0);
			int32 yApprox = (int32)tmpData(x,y,1);

			int32 xBlockCenter = x*macroBlockSize + macroBlockSize/2;
			int32 yBlockCenter = y*macroBlockSize + macroBlockSize/2;

			int32 px = xBlockCenter -imgSize[DIM_X]/2;
			int32 py = yBlockCenter -imgSize[DIM_Y]/2;
			int32 xNewBlockCenter = (int32)(data.a*px + data.b*py + data.dx + imgSize[DIM_X]/2.0 + 0.5);
			int32 yNewBlockCenter = (int32)(data.c*px + data.d*py + data.dy + imgSize[DIM_Y]/2.0 + 0.5);

			int32 xReal = xNewBlockCenter - xBlockCenter;
			int32 yReal = yNewBlockCenter - yBlockCenter;

			ieee64 dist = (xReal-xApprox)*(xReal-xApprox) + (yReal-yApprox)*(yReal-yApprox);
			if (dist>=45.0) {
				tmpData(x,y,2) = 0;
				valuesModified=true;
			}
		} // for
	} // for
	if (valuesModified) {
		getMotionParameters (tmpData, data);
	} // if
	cerr << "param[2]: " << data;



	// optimize and validate parameters
	// check, if each motion vector fits to the modell, remove the motion vectors, which do not fit 
	// (1) vectors at the border of the image
	// (2) vectors with significant difference
	// ==> set tmpData(x,y,2)=0
	// set threshold to 9 (mv in x or y direction has 3 or more pixels difference)
	tmpData = motionVectors;
	valuesModified=false;
	for (int32 y=0; y< motionVectorSize[DIM_Y]; y++) {
		for (int32 x=0; x< motionVectorSize[DIM_X]; x++) {

	//int32 t1 = tmpData(x,y,0);
	//int32 t2 = tmpData(x,y,1);
	//int32 t3 = tmpData(x,y,2);
	//int32 t4 = tmpData(x+1,y,0);
	//int32 t5 = tmpData(x+1,y,1);
	//int32 t6 = tmpData(x+1,y,2);
	//int32 t7 = tmpData(x,y+1,0);
	//int32 t8 = tmpData(x,y+1,1);
	//int32 t9 = tmpData(x,y+1,2);


			int32 xApprox = (int32)tmpData(x,y,0);
			int32 yApprox = (int32)tmpData(x,y,1);

			int32 xBlockCenter = x*macroBlockSize + macroBlockSize/2;
			int32 yBlockCenter = y*macroBlockSize + macroBlockSize/2;

			int32 px = xBlockCenter -imgSize[DIM_X]/2;
			int32 py = yBlockCenter -imgSize[DIM_Y]/2;
			int32 xNewBlockCenter = (int32)(data.a*px + data.b*py + data.dx + imgSize[DIM_X]/2.0 + 0.5);
			int32 yNewBlockCenter = (int32)(data.c*px + data.d*py + data.dy + imgSize[DIM_Y]/2.0 + 0.5);

			int32 xReal = xNewBlockCenter - xBlockCenter;
			int32 yReal = yNewBlockCenter - yBlockCenter;

			ieee64 dist = (xReal-xApprox)*(xReal-xApprox) + (yReal-yApprox)*(yReal-yApprox);
			if (dist>=20.0) {
				tmpData(x,y,2) = 0;
				valuesModified=true;
			}
		} // for
	} // for
	if (valuesModified) {
		getMotionParameters (tmpData, data);
	} // if
	cerr << "param[3]: " << data << endl;


	// optimize and validate parameters
	// check, if each motion vector fits to the modell, remove the motion vectors, which do not fit 
	// (1) vectors at the border of the image
	// (2) vectors with significant difference
	// ==> set tmpData(x,y,2)=0
	// set threshold to 9 (mv in x or y direction has 3 or more pixels difference)
	tmpData = motionVectors;
	valuesModified=false;
	for (int32 y=0; y< motionVectorSize[DIM_Y]; y++) {
		for (int32 x=0; x< motionVectorSize[DIM_X]; x++) {

	//int32 t1 = tmpData(x,y,0);
	//int32 t2 = tmpData(x,y,1);
	//int32 t3 = tmpData(x,y,2);
	//int32 t4 = tmpData(x+1,y,0);
	//int32 t5 = tmpData(x+1,y,1);
	//int32 t6 = tmpData(x+1,y,2);
	//int32 t7 = tmpData(x,y+1,0);
	//int32 t8 = tmpData(x,y+1,1);
	//int32 t9 = tmpData(x,y+1,2);


			int32 xApprox = (int32)tmpData(x,y,0);
			int32 yApprox = (int32)tmpData(x,y,1);

			int32 xBlockCenter = x*macroBlockSize + macroBlockSize/2;
			int32 yBlockCenter = y*macroBlockSize + macroBlockSize/2;

			int32 px = xBlockCenter -imgSize[DIM_X]/2;
			int32 py = yBlockCenter -imgSize[DIM_Y]/2;
			int32 xNewBlockCenter = (int32)(data.a*px + data.b*py + data.dx + imgSize[DIM_X]/2.0 + 0.5);
			int32 yNewBlockCenter = (int32)(data.c*px + data.d*py + data.dy + imgSize[DIM_Y]/2.0 + 0.5);

			int32 xReal = xNewBlockCenter - xBlockCenter;
			int32 yReal = yNewBlockCenter - yBlockCenter;

			ieee64 dist = (xReal-xApprox)*(xReal-xApprox) + (yReal-yApprox)*(yReal-yApprox);
			if (dist>=9.0) {
				tmpData(x,y,2) = 0;
				valuesModified=true;
			}
		} // for
	} // for
	if (valuesModified) {
		getMotionParameters (tmpData, data);
	} // if
	cerr << "param[3]: " << data << endl;


	return;
} // process


void 
MotionParametrization::compensateForwardMotion (Image<uint8> &src, Image<uint8> &dest, MotionParametrizationData& data)
{
	dest=src;

	for (int32 y=0; y<imgSize[DIM_Y]; y++) {
		for (int32 x=0; x<imgSize[DIM_X]; x++) {
			
			ieee64 px=x-imgSize[DIM_X]/2.0;
			ieee64 py=y-imgSize[DIM_Y]/2.0;
			
			int32 destX = (int32)(data.a*px + data.b*py + data.dx + imgSize[DIM_X]/2.0);
			int32 destY = (int32)(data.c*px + data.d*py + data.dy + imgSize[DIM_Y]/2.0);
			if (destX < 0) destX=0;
			if (destX >= imgSize[DIM_X]) destX=imgSize[DIM_X]-1;
			if (destY < 0) destY=0;
			if (destY >= imgSize[DIM_Y]) destY=imgSize[DIM_Y]-1;
			
			for (int32 b=0; b<imgSize[DIM_BAND]; b++) {
				dest(x, y, b) = src(destX, destY, b);
			} // for
		} // for
	} // for
	return;
}


void 
MotionParametrization::compensateBackwardMotion (Image<uint8> &src, Image<uint8> &dest, MotionParametrizationData& data)
{
	dest=src;

	for (int32 y=0; y<imgSize[DIM_Y]; y++) {
		for (int32 x=0; x<imgSize[DIM_X]; x++) {
			
			ieee64 px=x-imgSize[DIM_X]/2.0;
			ieee64 py=y-imgSize[DIM_Y]/2.0;
			bool valuesValid=true;
			
			ieee64 nx1 = data.d*px - data.b*py - data.d*data.dx + data.b*data.dy;
			ieee64 nx2 = data.a*data.d - data.b*data.c;
			if (nx2 == 0.0) {
				valuesValid=false;
				nx2=1.0;
			} // if
			int32 destX = (int32)(nx1/nx2 + imgSize[DIM_X]/2.0 + 0.5);
			if (destX < 0) destX=0;
			if (destX >= imgSize[DIM_X]) destX=imgSize[DIM_X]-1;
			
			ieee64 ny1 = data.c*px - data.a*py - data.c*data.dx + data.a*data.dy;
			ieee64 ny2 = data.b*data.c - data.a*data.d;
			if (ny2 == 0.0) {
				valuesValid=false;
				ny2=1.0;
			} // if
			int32 destY = (int32)(ny1/ny2 + imgSize[DIM_Y]/2.0 + 0.5);
			if (destY < 0) destY=0;
			if (destY >= imgSize[DIM_Y]) destY=imgSize[DIM_Y]-1;
			
			if (destX != x) {
				cout << "";
			}
			if (destY != y) {
				cout << "";
			}

			if (valuesValid) {
				for (int32 b=0; b<imgSize[DIM_BAND]; b++) {
					dest(x, y, b) = src(destX, destY, b);
				} // for
			} // if
		} // for
	} // for
	return;
}

