/*
    VideoAbstractFeature.cpp     generate video abstracts (summaries)
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// MoCA includes
#include <VideoAbstractFeature.h>


std::ostream& operator<< (ostream &os, const SceneDescription &data) 
{
	os << setw(5) << setfill(' ') << data.firstFrame << "..";
	os << setw(5) << setfill(' ') << data.lastFrame << "   ";
	os << setw(3) << setfill(' ') << (int)data.priorityLength << "    ";
	os << setw(3) << setfill(' ') << (int)data.priorityAction << "    ";
	os << setw(3) << setfill(' ') << (int)data.priorityContrast << "     ";
	os << setw(3) << setfill(' ') << (int)data.priorityFaces << "    ";
	os << setw(3) << setfill(' ') << (int)data.priorityCluster << "    ";
	os << setw(3) << setfill(' ') << (int)data.priorityPosition << "     ";
	
	os << setw(4) << setfill(' ') << (int)data.summarizedPriority << "   ";
	if (data.getSelected() == SELECTED) os << "+++";
	if (data.getSelected() == IGNORED) os << "---";
	os << endl;

	return os;
} // operator <<




std::ostream& operator<< (ostream &os, const VideoDescription &data) 
{
	os << "framerate         : " << data.frameRate       << endl;
	os << "number of scenes  : " << data.numberOfScenes  << endl;
	os << "number of frames  : " << data.numberOfFrames  << endl;
	os << "number of seconds : " << data.numberOfSeconds << endl;
	os << "selected scenes   : " << data.numberOfSelectedScenes  << endl;
	os << "selected frames   : " << data.numberOfSelectedFrames  << endl;
	os << "selected seconds  : " << data.numberOfSelectedSeconds << endl;
	os << endl;

	os << "priority values" << endl;
	os << "    frames    length action contrast faces cluster position sum" << endl;     

	list<SceneDescription>::const_iterator it;
	for (it=data.scenes.begin(); it!=data.scenes.end(); it++) {
		os << (*it);
	} // for

	return os;
} // operator <<




OTagStream& operator<< (OTagStream& os, const VideoDescription &data) 
{
    TagFormat::writeStartTag (os, "VIDEO_DESCRIPTION");
	os << endl;

    TagFormat::writeNodeValue(os, "version", data.version, 1);
    TagFormat::writeNodeValue(os, "framerate", data.frameRate, 1);
	os << endl;

    TagFormat::writeNodeValue(os, "numberOfScenes", data.numberOfScenes, 1);
    TagFormat::writeNodeValue(os, "numberOfFrames", data.numberOfFrames, 1);
    TagFormat::writeNodeValue(os, "numberOfSeconds", data.numberOfSeconds, 1);
    TagFormat::writeNodeValue(os, "numberOfSelectedScenes", data.numberOfSelectedScenes, 1);
    TagFormat::writeNodeValue(os, "numberOfSelectedFrames", data.numberOfSelectedFrames, 1);
    TagFormat::writeNodeValue(os, "numberOfSelectedSeconds", data.numberOfSelectedSeconds, 1);
	os << endl;


	if (data.numberOfScenes > 0) {
		TagFormat::writeStartTag (os, "SCENES", 1);
		os << endl;


		list<SceneDescription>::const_iterator it;
		for (it=data.scenes.begin(); it!=data.scenes.end(); it++) {

			TagFormat::writeStartTag (os, "firstframe", 2);
				os << setfill(' ') << setw(7) << (*it).firstFrame; 
			TagFormat::writeEndTag (os, "firstframe",-1);

			TagFormat::writeStartTag (os, "lastframe", 2);
				os << setfill(' ') << setw(7) << (*it).lastFrame; 
			TagFormat::writeEndTag (os, "lastframe",-1);

			TagFormat::writeStartTag (os, "numberOfFrames", 2);
				os << setfill(' ') << setw(7) << (*it).numberOfFrames; 
			TagFormat::writeEndTag (os, "numberOfFrames",-1);

			TagFormat::writeStartTag (os, "lengthSeconds", 2);
				os << setfill(' ') << setw(7) << (*it).lengthSeconds; 
			TagFormat::writeEndTag (os, "lengthSeconds",-1);

			TagFormat::writeStartTag (os, "histogramDifferences", 2);
				os << setfill(' ') << setw(7) << (*it).histogramDifferences; 
			TagFormat::writeEndTag (os, "histogramDifferences",-1);


			TagFormat::writeStartTag (os, "priorityLength", 2);
				os << setfill(' ') << setw(7) << (*it).priorityLength; 
			TagFormat::writeEndTag (os, "priorityLength",-1);

			TagFormat::writeStartTag (os, "priorityAction", 2);
				os << setfill(' ') << setw(7) << (*it).priorityAction; 
			TagFormat::writeEndTag (os, "priorityAction",-1);

			TagFormat::writeStartTag (os, "priorityContrast", 2);
				os << setfill(' ') << setw(7) << (*it).priorityContrast; 
			TagFormat::writeEndTag (os, "priorityContrast",-1);

			TagFormat::writeStartTag (os, "priorityCluster", 2);
				os << setfill(' ') << setw(7) << (*it).priorityCluster; 
			TagFormat::writeEndTag (os, "priorityCluster",-1);

			TagFormat::writeStartTag (os, "priorityPosition", 2);
				os << setfill(' ') << setw(7) << (*it).priorityPosition; 
			TagFormat::writeEndTag (os, "priorityPosition",-1);

			string s = "SELECTED";
			if ((*it).getSelected() == IGNORED) s = " IGNORED";
			if ((*it).getSelected() == SELECTION_UNKNOWN) s = " UNKNOWN";
			TagFormat::writeNodeValue(os, "selected", s,2);
		} // for

		TagFormat::writeEndTag (os, "SCENES", 1);
		os << endl;
	} // if

    TagFormat::writeEndTag (os, "VIDEO_DESCRIPTION");
	os << endl;
	
    return os;
} // operator<<



VideoAbstract::~VideoAbstract()
{
	deleteMetadataFiles();
	deleteTemporaryFiles();
} 


void
VideoAbstract::deleteMetadataFiles()
{
	if (deleteMetadata) {

		// delete source audio file
		if (srcAudioWavFile != "") {
			FileHandling::deleteFile (srcAudioWavFile, false);
		} // if

		// get all files in index directory
		CreateMocaPath mocaPath (srcFilename, false);
		string path     = mocaPath.destinationPath;
		string allFiles = path + "*.*";
		
		// delete all metadata files
		FileHandling::deleteFile (allFiles, false);
		
		// delete metadata directory
		FileHandling::deleteDirectory (path, false);
	} // if
	return;
} 


void
VideoAbstract::deleteTemporaryFiles()
{
	if (deleteTemporaryData) {
		cerr << "deleting temporary files..." << endl;
		FileHandling::deleteFile (destMv1File, true);
		FileHandling::deleteFile (srcAudioWavFile, true);
		FileHandling::deleteFile (destWavFile, true);
		FileHandling::deleteFile (destMp2File, true);
	} // if
	return;
} 


void 
VideoDescription::updateAudioSelectedSeconds (AudioCutData &audioCuts)
{

	// initialize array (set to: "not selected")
	int32 numberOfFrames = audioCuts.cuts.getSize()[DIM_X];
	SignalSize size(numberOfFrames);
	audioSelectedFramesList.resize(size);
	for (int32 i=0; i<numberOfFrames; i++) audioSelectedFramesList(i)=2; // not selected


	list<SceneDescription>::iterator it = scenes.begin();
	for (it=scenes.begin(); it!=scenes.end(); it++) {
		if ((*it).getSelected() == SELECTED) {

			// get best position for cut (search silent area)

			// search in the area near firstFrame, store value in firstFrame
			int32 pos1 = (*it).firstFrame;
			while ((pos1>0) && (audioCuts.cuts(pos1)<3)) pos1--;
			int32 pos2 = (*it).firstFrame;
			while ((pos2<numberOfFrames-1) && (audioCuts.cuts(pos2)<3)) pos2++;
			int32 posCenter=((*it).firstFrame+(*it).lastFrame)/2;
			int32 dist1 = (*it).firstFrame-pos1;
			int32 dist2 = pos2-(*it).firstFrame;
			int32 firstFrame = pos1;
			if (dist1>dist2) {
				firstFrame = pos2;
			} // if
			if (pos2>=posCenter) firstFrame = pos1;

			// search in the area near lastFrame, store value in lastFrame
			pos1 = (*it).lastFrame;
			while ((pos1>0) && (audioCuts.cuts(pos1)<3)) pos1--;
			pos2 = (*it).lastFrame;
			while ((pos2<numberOfFrames-1) && (audioCuts.cuts(pos2)<3)) pos2++;
			dist1 = (*it).lastFrame-pos1;
			dist2 = pos2-(*it).lastFrame;
			int32 lastFrame = pos1;
			if (dist1>dist2) {
				lastFrame = pos2;
			} // if
			if (pos1<=posCenter) lastFrame = pos2;

			if ((*it).firstFrame == 0) firstFrame=0;
			if ((*it).lastFrame == numberOfFrames-1) lastFrame=numberOfFrames-1;

			for (int32 i=firstFrame; i<=lastFrame; i++) audioSelectedFramesList(i)=1; // selected
		} // if
	} // for

	// count number of selected frames 
	audioSelectedFrames  = 0;
	audioSelectedSeconds = 0;
	for (int32 i=0; i<numberOfFrames; i++) {
		if (audioSelectedFramesList(i) == 1) audioSelectedFrames++;
	} // for
	audioSelectedSeconds = audioSelectedFrames/frameRate;

	return;
} // updateSelectedSeconds






VideoAbstract::VideoAbstract(const string &srcFile, const string &destFile, VideoAbstractParameters &param)
{

	// initialize parameters
	analyseFaces = param.faces;
	createAbstract = param.createAbstract;
	numberOfCluster = 20;
	noAudio = false;
	srcAudioWavFile="";
	createGrayscale=false;

	deleteMetadata      = param.deleteMetadata;
	deleteTemporaryData = param.deleteTemporaryData;

	// set source filenames
	CreateMocaPath mocaPathSrc (srcFile, false);
	srcFilename = mocaPathSrc.pathFile;
	destMv1File = mocaPathSrc.path + "tmp_abstract.mv1";
	destWavFile = mocaPathSrc.path + "tmp_abstract_dest.wav";
	destMp2File = mocaPathSrc.path + "tmp_abstract_dest.mp2";

	// does source file exist?
	if (!FileHandling::fileExists (srcFilename)) {
		cerr << endl << "Error: Could not open source file \"" << srcFilename << "\"" << endl;
		exit(1);
	} // if

	// set destination filenames
	CreateMocaPath mocaPathDest (destFile, false);
	destFilename = mocaPathDest.pathFile;

	// create directories for metadata
	FileHandling::createDirectory (mocaPathSrc.destinationPath, false);
	if (mocaPathDest.path != "./") {
		FileHandling::createDirectory (mocaPathDest.path, false);
	} // if


	videoFrames = -1;
	lastFrameToDecode=-1;

	// read video statistics
	try {
		// open reader
		VideoDecoder reader (srcFilename);

		// get size
		srcSize = reader.getSize();
		imgSize[DIM_X] = srcSize[DIM_X];
		imgSize[DIM_Y] = srcSize[DIM_Y];
		imgSize[DIM_Z] = 1;
		imgSize[DIM_BAND] = srcSize[DIM_BAND];
		numberOfFrames = srcSize[DIM_Z];

		// get frame rate
		frameRate = reader.getFrameRate();
		videoScenes.setFrameRate(frameRate);
		srcSeconds = srcSize[DIM_Z]/frameRate;
		srcMinutes = srcSeconds/60.0;

		// close reader
		//reader.close();
    } catch (...) {
        cerr << endl << "VideoAbstract: could not read video parameters." << endl;
        exit (1);
    } // try


	// do not create an abstract, if the length of the video is shorter than 0.5 minutes
	if (srcMinutes < 0.5) {

		// stop, if no explicit length was defined! 
		if ((param.lengthMinutes==-1.0) && (param.lengthPercent==-1.0)) {
			
			// set error message
			int32 secV = ((int32)(srcSeconds)) %60;
			int32 minV = (int32)(srcMinutes);
			string s = "";
			if (secV<10) s="0";
			string eDescription;
			{
				stringstream oss;
				oss << "The length of the source video is too short (" << minV << ":" << s << secV << ")";
				eDescription = oss.str();
			}
			eDescription += "\nNo abstract is created!\n";
			
			throw MocaException(eDescription);
		} // if
	} // if



	destSeconds=-1.0;
	destPercent=-1.0;
	if ((param.lengthMinutes==-1.0) && (param.lengthPercent==-1.0)) {

		// the length of the abstract of a very short video (the original video is shorter 
		// than 3 minutes), is 33.3 percent of the original video
		destPercent=33.3; 
		if (srcMinutes >   3) destPercent=32.0;
		if (srcMinutes >   4) destPercent=31.0;
		if (srcMinutes >   5) destPercent=30.0;
		if (srcMinutes >   6) destPercent=28.0;
		if (srcMinutes >   7) destPercent=26.0;
		if (srcMinutes >   8) destPercent=24.0;
		if (srcMinutes >   9) destPercent=22.0;
		if (srcMinutes >  10) destPercent=20.0;
		if (srcMinutes >  13) destPercent=18.0;
		if (srcMinutes >  16) destPercent=16.0;
		if (srcMinutes >  20) destPercent=15.0;
		if (srcMinutes >  25) destPercent=14.0;
		if (srcMinutes >  30) destPercent=13.0;
		if (srcMinutes >  40) destPercent=12.0;
		if (srcMinutes >  50) destPercent=11.0;
		if (srcMinutes >  60) destPercent=10.0;
		if (srcMinutes >  90) destPercent= 9.0;
		if (srcMinutes > 120) destPercent= 8.0;
	} // if
	if (param.lengthMinutes!=-1.0) {
		destSeconds=60.0*param.lengthMinutes;
	} // if
	if (param.lengthPercent!=-1.0) {
		destPercent=param.lengthPercent;
	} // if

	if (destSeconds!=-1.0) {
		if (destSeconds>=srcSeconds) {
			cerr << endl;
			cerr << "Warning: destination video is larger than source video!" << endl;

			int32 tmp=(int32)srcSeconds/3600;
			cerr << "length of source video     : ";
			if (tmp<10) cerr << "0";
			cerr << tmp << ":";
			tmp=(int32)((((int32)srcSeconds)%3600)/60);
			if (tmp<10) cerr << "0";
			cerr << tmp << ":";
			tmp=(int32)(((int32)srcSeconds)%60);
			if (tmp<10) cerr << "0";
			cerr << tmp << endl;

			tmp=(int32)destSeconds/3600;
			cerr << "length of destination video: ";
			if (tmp<10) cerr << "0";
			cerr << tmp << ":";
			tmp=(int32)((((int32)destSeconds)%3600)/60);
			if (tmp<10) cerr << "0";
			cerr << tmp << ":";
			tmp=(int32)(((int32)destSeconds)%60);
			if (tmp<10) cerr << "0";
			cerr << tmp << endl;

			destSeconds=-1.0;
			destPercent=10.0;
			cerr << endl;
			cerr << "setting length of destination video to default values." << endl;
			cerr << endl;
		} // if
	} // if

	if ((destSeconds!=-1.0) && (destSeconds<=0.0)) {
		cerr << endl;
		cerr << "Warning: length of destination video is not valid!" << endl;
		cerr << endl;
		cerr << "setting length of destination video to default values." << endl;
		cerr << endl;
		destSeconds=-1.0;
		destPercent=10.0;
	} // if
	if (destPercent!=-1.0) {
		if ((destPercent<=0.0) || (destPercent>100.0)) {
			cerr << endl;
			cerr << "Warning: length of destination video is not valid!" << endl;
			cerr << endl;
			cerr << "setting length of destination video to default values." << endl;
			cerr << endl;
			destSeconds=-1.0;
			destPercent=10.0;
		} // if
		destSeconds = srcSeconds*destPercent/100.0;
	} else {
		destPercent = destSeconds/srcSeconds;
	} // else


	// set metadata files
	AVDocumentFilename = param.AVDocumentFilename;
	videoLevelFilename = param.videoLevelFilename;
	VDXFilename        = param.VDXFilename;

	return;
} // VideoAbstract


std::ostream& operator<< (ostream& os, const VideoAbstract &data)
{
	os << "srcFilename           : " << data.srcFilename << endl;
	os << "destFilename          : " << data.destFilename << endl;
	os << "destMv1File           : " << data.destMv1File << endl;
	os << "srcAudioWavFile       : " << data.srcAudioWavFile << endl;
	//os << "srcMp2File            : " << data.srcMp2File << endl;
	//os << "srcWavFile            : " << data.srcWavFile << endl;
	os << "destWavFile           : " << data.destWavFile << endl;
	os << "destMp2File           : " << data.destMp2File << endl;
	os << "srcSize[DIM_X]        : " << (int32)data.srcSize[DIM_X] << endl;
	os << "srcSize[DIM_Y]        : " << (int32)data.srcSize[DIM_Y] << endl;
	os << "srcSize[DIM_Z]        : " << (int32)data.srcSize[DIM_Z] << endl;
	os << "srcSize[DIM_BAND]     : " << (int32)data.srcSize[DIM_BAND] << endl;
	os << "videoFrames           : " << data.videoFrames << endl;
	os << "lastFrameToDecode     : " << data.lastFrameToDecode << endl;
	os << "frameRate             : " << data.frameRate << endl;
	os << "srcSeconds            : " << data.srcSeconds << endl;
	os << "destSeconds           : " << data.destSeconds << endl;
	os << "destPercent           : " << data.destPercent << endl;

	os << endl;

	os << endl << "scene information:" << endl;
	os << "numberOfFrames        : " << data.videoScenes.numberOfFrames << endl;
	os << "numberOfScenes        : " << data.videoScenes.numberOfScenes << endl;
	os << "numberOfSeconds       : " << data.videoScenes.numberOfSeconds << endl;
	os << "frameRate             : " << data.videoScenes.frameRate << endl;

	os << endl << "selected frames:" << endl;
	os << "selected frames       : " << data.videoScenes.numberOfSelectedFrames << endl;
	os << "selected scenes       : " << data.videoScenes.numberOfSelectedScenes << endl;
	os << "selected seconds      : " << data.videoScenes.numberOfSelectedSeconds << endl;
	
	os << endl;


	
	list<SceneDescription>::const_iterator it;
	for (it=data.videoScenes.scenes.begin(); it!=data.videoScenes.scenes.end(); it++) {
		os << "scene: " << (*it).firstFrame << ".." << (*it).lastFrame;
		os << "\t  " << (int)(*it).priorityLength;
		os << "\t  " << (int)(*it).priorityAction;
		os << "\t  " << (int)(*it).priorityCluster;
		os << "\t  " << (int)(*it).priorityPosition;

		if ((*it).getSelected() == SELECTED) os << "\t  selected";
		os << endl;
	} // for
	
	return os;
} // operator <<


void 
VideoAbstract::setPriority(ActionData &action, VideoStatisticData &statistic, FaceData &faces) 
{
	list<SceneDescription>::iterator it;
	for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {

		// set priority length
		(*it).priorityLength=128;
		
		// set priority action
		ieee64 sum=0.0;
		for (int32 i=(*it).firstFrame; i<(*it).lastFrame; i++) {
			sum+=action.imgAction(i) + action.motionAction(i);
		} // for
		if ((*it).lastFrame-(*it).firstFrame > 0) {
			sum/=((*it).lastFrame-(*it).firstFrame);
		} // if
		sum*=255.0;
		if (sum>255.0) sum=255.0;
		if (sum<0.0) sum=0.0;
		(*it).priorityAction=(uint8)sum; 
		
		// set priority contrast
		sum=0.0;
		for (int32 i=(*it).firstFrame; i<=(*it).lastFrame; i++) {
			sum+=statistic.imgData[i].contrastHistogram + statistic.imgData[i].contrastScale;
		} // for
		if ((*it).lastFrame-(*it).firstFrame+1 > 0) {
			sum/=((*it).lastFrame-(*it).firstFrame+1);
		} // if
		if (sum>255.0) sum=255.0;
		if (sum<0.0) sum=0.0;
		(*it).priorityContrast=(uint8)sum; 
		
		
		// set priority faces
		sum=0.0;
		list<FaceItem>::const_iterator faceIt;
		for (faceIt=faces.faces.begin(); faceIt!=faces.faces.end(); faceIt++) {
			//FaceItem tmp = (*faceIt);
			if ((*faceIt).p1[DIM_X] != -1) {
				if (((*faceIt).frame >= (*it).firstFrame) && ((*faceIt).frame <= (*it).lastFrame)) {
					int32 width  = labs((*faceIt).p2[DIM_X]-(*faceIt).p1[DIM_X]);
					int32 height = labs((*faceIt).p2[DIM_Y]-(*faceIt).p1[DIM_Y]);
					int32 sizeValue=(width+height)/2;

					// increase sum
					sum += 10 + sizeValue;
				} // if
			} // if
		} // for
		if (sum<0.0) sum=0.0;
		if (sum>255.0) sum=255.0;
		(*it).priorityFaces=(uint8)sum; 

	} // for

	return;
} // setPriority


void 
VideoAbstract::setPriorityLength() 
{
	// get average length
	ieee64 totalSec=0.0;
	int32 nShots=0;
	ieee64 maxSec=0.0;
	ieee64 minSec=10e10;
	list<SceneDescription>::iterator it;
	for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
		int32 length = (*it).lastFrame - (*it).firstFrame+1;
		ieee64 sec =length/frameRate;
		if (maxSec <sec) maxSec=sec;
		if (minSec >sec) minSec=sec;
		totalSec+=sec;
		nShots++;
	} // for
	if (nShots==0) nShots=1;
	ieee64 avgLength = totalSec/nShots;

	ieee64 f1 = 255.0/(avgLength-minSec);
	ieee64 f2 = 255.0/(maxSec-avgLength);
	for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
		int32 length = (*it).lastFrame - (*it).firstFrame+1;
		ieee64 sec =length/frameRate;

		ieee64 v=0.0;
		if (sec <=avgLength) v = 255.0 - (avgLength - sec)*f1;
		if (sec > avgLength) v = 255.0 - (sec - avgLength)*f2;

		if (v<0.0)   v=0.0;
		if (v>255.0) v=255.0;
		(*it).priorityLength = (uint8)v;
	} // for

	return;
} // setPriorityLength



void 
VideoAbstract::updatePositionInformation(Signal<uint8> &selectedFrames)
{
	// initialize values
	list<SceneDescription>::iterator it = videoScenes.scenes.begin();
	for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
		(*it).priorityPosition = numberOfFrames+1000;
	} // for
	
	// do we have selected frames?
	bool framesSelected=false;
	for (int32 i=0; i<numberOfFrames; i++) {
		if (selectedFrames(i) == 1) {
			framesSelected=true;
			break;
		} // if
	} // for
	if (!framesSelected) {
		for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
			(*it).priorityPosition = 0;
		} // for
	} else {
		for (int32 i=0; i<numberOfFrames; i++) {
			if (selectedFrames(i) == 1) {
				
				// set priority values (distance to selected frame)
				for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
					int32 center = ((*it).firstFrame+(*it).lastFrame)/2;
					int32 distance = labs(center-i);
					if ((*it).priorityPosition>distance) (*it).priorityPosition=distance;
				} // for
				while ((i+2<numberOfFrames) && (selectedFrames(i+2) == 1)) i++;
			} // if
		} // for
		
		// normalize priorityPosition to 0..255
		int32 minValue = numberOfFrames+1000;
		int32 maxValue = -minValue;
		for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
			if (minValue > (*it).priorityPosition) minValue = (*it).priorityPosition; 
			if (maxValue < (*it).priorityPosition) maxValue = (*it).priorityPosition; 
		} // for
		
		// scale values to 0..255
		maxValue-=minValue;
		if (maxValue<1) maxValue=1;
		ieee64 f=255.0/maxValue;
		for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
			int32 v = (int32)(((*it).priorityPosition - minValue)*f);
			if (v<0) v=0;
			if (v>255) v=255;
			(*it).priorityPosition = (uint8)v;
		} // for
	} // else
	return;
} // updatePositionInformation


void 
VideoAbstract::scalePriorityValues()
{
	list<SceneDescription>::iterator it = videoScenes.scenes.begin();

	// set priority length
	int32 minValue =  1000;
	int32 maxValue = -1000;
	for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
		uint8 v = (*it).priorityLength;
		if (minValue > v) minValue=v;
		if (maxValue < v) maxValue=v;
	} // for
	if (minValue == maxValue) {
		// set all values to 128
		for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
			(*it).priorityLength = 128;
		} // for
	} else {
		if ((minValue!=0) || (maxValue!=255)) {
			// scale values to 0..255
			maxValue-=minValue;
			if (maxValue<1) maxValue=1;
			ieee64 f=255.0/maxValue;
			for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
				int32 v = (int32)(((*it).priorityLength - minValue)*f);
				if (v<0) v=0;
				if (v>255) v=255;
				(*it).priorityLength = (uint8)v;
			} // for
		} // if
	} // else

	// set priority action
	minValue =  1000;
	maxValue = -1000;
	for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
		uint8 v = (*it).priorityAction;
		if (minValue > v) minValue=v;
		if (maxValue < v) maxValue=v;
	} // for
	if (minValue == maxValue) {
		// set all values to 128
		for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
			(*it).priorityAction = 128;
		} // for
	} else {
		if ((minValue!=0) || (maxValue!=255)) {
			// scale values to 0..255
			maxValue-=minValue;
			if (maxValue<1) maxValue=1;
			ieee64 f=255.0/maxValue;
			for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
				int32 v = (int32)(((*it).priorityAction - minValue)*f);
				if (v<0) v=0;
				if (v>255) v=255;
				(*it).priorityAction = (uint8)v;
			} // for
		} // if
	} // else

	// set priority contrast
	minValue =  1000;
	maxValue = -1000;
	for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
		uint8 v = (*it).priorityContrast;
		if (minValue > v) minValue=v;
		if (maxValue < v) maxValue=v;
	} // for
	if (minValue == maxValue) {
		// set all values to 128
		for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
			(*it).priorityContrast = 128;
		} // for
	} else {
		if ((minValue!=0) || (maxValue!=255)) {
			// scale values to 0..255
			maxValue-=minValue;
			if (maxValue<1) maxValue=1;
			ieee64 f=255.0/maxValue;
			for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
				int32 v = (int32)(((*it).priorityContrast - minValue)*f);
				if (v<0) v=0;
				if (v>255) v=255;
				(*it).priorityContrast = (uint8)v;
			} // for
		} // if
	} // else

	// set priority faces
	minValue =  1000;
	maxValue = -1000;
	for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
		uint8 v = (*it).priorityFaces;
		if (minValue > v) minValue=v;
		if (maxValue < v) maxValue=v;
	} // for
	if (minValue == maxValue) {
		// set all values to 128
		for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
			(*it).priorityFaces = 128;
		} // for
	} else {
		if ((minValue!=0) || (maxValue!=255)) {
			// scale values to 0..255
			maxValue-=minValue;
			if (maxValue<1) maxValue=1;
			ieee64 f=255.0/maxValue;
			for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
				int32 v = (int32)(((*it).priorityFaces - minValue)*f);
				if (v<0) v=0;
				if (v>255) v=255;
				(*it).priorityFaces = (uint8)v;
			} // for
		} // if
	} // else

	// set priority cluster
	minValue =  1000;
	maxValue = -1000;
	for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
		uint8 v = (*it).priorityCluster;
		if (minValue > v) minValue=v;
		if (maxValue < v) maxValue=v;
	} // for
	if (minValue == maxValue) {
		// set all values to 128
		for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
			(*it).priorityCluster = 128;
		} // for
	} else {
		if ((minValue!=0) || (maxValue!=255)) {
			// scale values to 0..255
			maxValue-=minValue;
			if (maxValue<1) maxValue=1;
			ieee64 f=255.0/maxValue;
			for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
				int32 v = (int32)(((*it).priorityCluster - minValue)*f);
				if (v<0) v=0;
				if (v>255) v=255;
				(*it).priorityCluster = (uint8)v;
			} // for
		} // if
	} // else

	// set priority position
	minValue =  1000;
	maxValue = -1000;
	for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
		uint8 v = (*it).priorityPosition;
		if (minValue > v) minValue=v;
		if (maxValue < v) maxValue=v;
	} // for
	if (minValue == maxValue) {
		// set all values to 128
		for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
			(*it).priorityPosition = 128;
		} // for
	} else {
		if ((minValue!=0) || (maxValue!=255)) {
			// scale values to 0..255
			maxValue-=minValue;
			if (maxValue<1) maxValue=1;
			ieee64 f=255.0/maxValue;
			for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
				int32 v = (int32)(((*it).priorityPosition - minValue)*f);
				if (v<0) v=0;
				if (v>255) v=255;
				(*it).priorityPosition = (uint8)v;
			} // for
		} // if
	} // else

	return;
} // scalePriorityValues




void 
VideoAbstract::updateClusterInformation(Signal<uint8> &selectedFrames)
{
	// initialize signal
	SignalSize clusterSize (numberOfCluster,2);
	Signal<int32> clusterData (clusterSize);
	for (int32 i=0; i<numberOfCluster; i++) {
		clusterData (i,0) = 0; // number of frames in cluster
		clusterData (i,1) = 0; // number of selected frames in cluster
	} // for

	list<SceneDescription>::iterator it = videoScenes.scenes.begin();
	for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
		int32 left = (*it).firstFrame;
		int32 right= (*it).lastFrame;
		int32 sceneID = (*it).sceneID;
		if (sceneID >=0) {
			clusterData (sceneID,0)+=(right-left+1);
			for (int32 i=left; i<=right; i++) {
				if (selectedFrames(i) == 1) {
					clusterData (sceneID,1)++;
				} // if
			} // for
		} // if
	} // for
	// reduce the number of frames for each cluster by 2x the number of selected frames
	// get minumum and maximum
	int32 minValue = 1000000;
	int32 maxValue = -minValue;
	for (int32 i=0; i<numberOfCluster; i++) {
		if (clusterData (i,0) != 0) {
			clusterData (i,0) -= 2*clusterData (i,1);
			if (minValue > clusterData (i,0)) minValue = clusterData (i,0);
			if (maxValue < clusterData (i,0)) maxValue = clusterData (i,0);
		} // if
	} // for
	// scale values to 0..255
	maxValue-=minValue;
	if (maxValue<1) maxValue=1;
	ieee64 f=255.0/maxValue;
	for (int32 i=0; i<numberOfCluster; i++) {
		if (clusterData (i,0) != 0) {
			clusterData (i,0)-=minValue;
			clusterData (i,0) = (int32)(clusterData (i,0)*f);
			if (clusterData (i,0)<0) clusterData (i,0)=0;
			if (clusterData (i,0)>255) clusterData (i,0)=255;
		} // if
	} // for
	// store the priority for each shot
	for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
		int32 sceneID = (*it).sceneID;
		(*it).priorityCluster = clusterData (sceneID,0);
	} // for

	return;
} // updateClusterInformation


void 
VideoAbstract::selectFrames() 
{
	// create a copy of the video
	if (destPercent>= 100.0) {
		SceneDescription scene (0, numberOfFrames-1, frameRate, 0);
		videoScenes.addScene(scene);
		list<SceneDescription>::iterator it = videoScenes.scenes.begin();
		videoScenes.selectScene(*it);

		lastFrameToDecode = -1;		
		return;
	} // if


	cout << "start module: action detection" << endl;
	ActionData action;
	{
		ActionDetection reader (srcFilename);
		reader.process (action);
	}


	cout << "start module: video statistics" << endl;
	VideoStatisticData statistic;
	{
		createGrayscale = false;
		VideoStatistic vs (srcFilename);
		vs.process(statistic);
		createGrayscale = ((statistic.color == TYPE_GRAYSCALE) || (statistic.color == TYPE_NEAR_GRAYSCALE));
	}


    FaceData faces;
	if (analyseFaces) {
		cout << "start module: face detection" << endl;
		// initialize face detection
		FaceDetectionParameter param;
		//param.verbose = true;
		//param.defaultFaceSize=FACE_DETECTION_DEFAULT_FACE_SIZE_MEDIUM;
		param.defaultFaceSize=FACE_DETECTION_SIZE_LARGE;
		param.stepWidth = 8;
		param.mpgFilename = srcFilename;
		param.inputData= FACE_INPUT_VIDEO;
		FaceDetection fd(param);
		fd.processVideo(faces);
	} // if


	// initialize shot change
	cout << "start module: cut detection" << endl;
	CutDetectionData cuts;
	int32 minimumNumberOfAudioCuts = 0;
	{
		CutDetection reader (srcFilename);
		reader.process (cuts);
		minimumNumberOfAudioCuts = 3 + 
			cuts.numberOfDissolves + 
			cuts.numberOfFadeIn + 
			cuts.numberOfFadeOut + 
			cuts.numberOfHardCuts +
			cuts.numberOfUnknownSceneChanges;

		bool select=true;
		SceneDetectionData clusterScene; 
		SceneDetection detectClusterScene (srcFilename, numberOfCluster, select);
		detectClusterScene.process (clusterScene);

		Signal<int32> shotFrameNumber;
		reader.getShotFrameNumbers (cuts, shotFrameNumber);
		for (int32 i=0; i<cuts.numberOfFrames; i++) {
			if (shotFrameNumber(i,0) != 0) {
				int32 firstFrame = shotFrameNumber(i,1);
				int32 lastFrame = shotFrameNumber(i,2);
				
				int32 sceneID=0;
				list<SceneDetectionItem>::iterator it;
				for (it=clusterScene.scenes.begin(); it!=clusterScene.scenes.end(); it++) {
					if ((*it).referenceFrame == i) sceneID = (*it).sceneID;
				} // for

				SceneDescription scene (firstFrame, lastFrame, frameRate, sceneID);
				videoScenes.addScene(scene);
			} // if
		} // for
	}

	// initialize audio
	cout << endl << "start module: audio analysis" << endl;
	AudioCutData audioCuts;
	{

		// extract audio from MPEG file
		AudioDecoderData audioData;
		AudioDecoder readerAudio (srcFilename);
		readerAudio.process(audioData);
		srcAudioWavFile = readerAudio.getAudioPathFilename();
		noAudio = (!audioData.hasAudio);


		//AudioConvertData audioConvertData;
		//AudioConvert convert(srcFilename);
		//convert.process (audioConvertData);
		//convert.getPath(srcAudioWavFile);
		//srcAudioWavFile+=audioConvertData.audioFilename;


		// analyse audio track
		//noAudio = audioConvertData.noAudio;
		if (noAudio) {
			SignalSize audioSize;
			audioSize[DIM_X] = numberOfFrames;
			audioSize[DIM_Y] = 1;
			audioSize[DIM_Z] = 1;
			audioSize[DIM_BAND] = 1;
			audioCuts.cuts.resize(audioSize);
			audioCuts.blockLengthMilliSeconds = (int32)(1000.0/frameRate);  // one block matches with one frame
			audioCuts.filename="";
			audioCuts.numberOfBlocks = numberOfFrames;
			audioCuts.numberOfCuts = numberOfFrames;
			for (int32 i=0; i<numberOfFrames; i++) {
				audioCuts.cuts(i)=128;
			} // for
		} else {
			int32 blockLengthMilliSeconds=(int32)(1000.0/frameRate);  // one block matches with one frame
			
			AudioCut reader(srcAudioWavFile, minimumNumberOfAudioCuts, blockLengthMilliSeconds);
			reader.process(audioCuts);
		}// if
	}

	// set priority values of scenes: action, statistic, and faces 
	setPriority(action, statistic, faces);

	// set priority values of scenes: length
	setPriorityLength();


// print debug information
#if 0
	const int maxP=40;
	ieee64 totalSec=0;
	int32 nShots=0;
	SignalSize tmpSize (maxP);
	Signal<int32> tmp (tmpSize);
	for (int32 i=0; i<maxP; i++) tmp(i)=0;
	list<SceneDescription>::iterator it;
	for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
		int32 length = (*it).lastFrame - (*it).firstFrame+1;
		ieee64 sec =length/frameRate;
		totalSec+=sec;
		nShots++;
		int32 pos = (int32)sec;
		if (pos<0) pos=0;
		if (pos>maxP-1) pos=maxP-1;
		tmp(pos)++;
	} // for
	cerr << "avg length shots: " << totalSec/nShots << endl;
	for (int32 i=0; i<maxP; i++) cout << tmp(i) << endl;
	exit(1);
#endif

	// remove shots with negative ID (predefined cluster centers, e.g. black frames, white frames)
	list<SceneDescription>::iterator it;
	for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
		if ((*it).sceneID < 0) {
			videoScenes.ignoreScene(*it);
		} // if
	} // for
	//cout << videoScenes << endl;



	// remove very short shots
	ieee64 lengthShortScene=0.5;
	for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
		if ((*it).lengthSeconds < lengthShortScene) {
			if (destSeconds+0.01 < videoScenes.numberOfSeconds-videoScenes.numberOfIgnoredSeconds) {
				videoScenes.ignoreScene(*it);
			} // if
		} // if
	} // for
	lengthShortScene=1.2;
	for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
		if ((*it).lengthSeconds < lengthShortScene) {
			if (destSeconds+0.01 < videoScenes.numberOfSeconds-videoScenes.numberOfIgnoredSeconds) {
				videoScenes.ignoreScene(*it);
			} // if
		} // if
	} // for
	lengthShortScene=2.0;
	for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
		if ((*it).lengthSeconds < lengthShortScene) {
			if (destSeconds+0.01 < videoScenes.numberOfSeconds-videoScenes.numberOfIgnoredSeconds) {
				videoScenes.ignoreScene(*it);
			} // if
		} // if
	} // for


	
	// select shots
#if 1
	cout << endl;
	cout << "first.. last   len action  contr   faces  clust    pos      sum   sel" << endl;
#endif
	ieee64 selectedSeconds=0.0;
	while (true) {

		// find audio cuts
		videoScenes.updateAudioSelectedSeconds (audioCuts);
		selectedSeconds = videoScenes.audioSelectedSeconds;
		if (selectedSeconds >= destSeconds) break;
		if (videoScenes.numberOfSelectedScenes >= videoScenes.numberOfScenes) break;

		//update cluster and scene-information (priorityCluster)
		updateClusterInformation(videoScenes.audioSelectedFramesList);

		//update shot-information (priorityPosition)
		updatePositionInformation(videoScenes.audioSelectedFramesList);
		scalePriorityValues();

		// display shot information
		//cout << videoScenes << endl;


		// select next shot
		int32 maxSummarizedPriority = -10000;
		list<SceneDescription>::iterator maxIt = videoScenes.scenes.begin();
		for (it=videoScenes.scenes.begin(); it!=videoScenes.scenes.end(); it++) {
			if ((*it).getSelected() != SELECTED) {
				(*it).updateSummarizedPriority();

				if (maxSummarizedPriority < (*it).summarizedPriority) {
					maxSummarizedPriority = (*it).summarizedPriority;
					maxIt = it;
				} // if
			} // if
		} // for
		videoScenes.selectScene(*maxIt);

#if 1
		cout << (*maxIt);
#endif

		// display shot information
		//cout << videoScenes << endl;

	} // while

	
	// update selected frames
	videoScenes.updateAudioSelectedSeconds (audioCuts);
	lastFrameToDecode = videoScenes.audioSelectedFramesList.getSize()[DIM_X]-1;
	for (int32 i=lastFrameToDecode; i>=0; i--) {
		if (videoScenes.audioSelectedFramesList(i) == 1) {
			lastFrameToDecode=i+2;
			break;
		} // if
	} // for
	if (lastFrameToDecode+3>=srcSize[DIM_Z]) {
		lastFrameToDecode = -1;
	} // if

	// print selected frames
#if 1
	videoScenes.printSelected();
	videoScenes.printAudioSelected();
#endif
	return;
} // selectFrames



	//cout << (*this) << endl;

	// create scene descriptions

	/*
	// example: two scenes with a hard cut
	SceneDescription scene1 (
		SceneChange(1912), 
		SceneChange(1958)
		);
	SceneDescription scene2 (
		SceneChange(2028),
		SceneChange(2055)
		);
	*/

	/*
	// example: two scenes with four different fades
	SceneDescription scene1 (
		SceneChange(1912, SCENE_CHANGE_FADE, 10, 0), 
		SceneChange(1958, SCENE_CHANGE_FADE, 20, 128)
		);
	SceneDescription scene2 (
		SceneChange(2028, SCENE_CHANGE_FADE,  5, 255),
		SceneChange(2055, SCENE_CHANGE_FADE, 15,  64)
		);
	*/


	
	// example: two scenes with one dissolve between
	/*
	SceneDescription scene1 (
		SceneChange(554),
		SceneChange(623, SCENE_CHANGE_DISSOLVE,  4, 0)
		);
	SceneDescription scene2 (
		SceneChange(889, SCENE_CHANGE_DISSOLVE,  4, 0),
		SceneChange(944, SCENE_CHANGE_DISSOLVE,  7, 0)
		);
	SceneDescription scene3 (
		SceneChange(960, SCENE_CHANGE_DISSOLVE,  7, 0),
		SceneChange(1005, SCENE_CHANGE_DISSOLVE, 10, 0)
		);
	SceneDescription scene4 (
		SceneChange(1018, SCENE_CHANGE_DISSOLVE, 10, 0),
		SceneChange(1090)
		);
	

	// add scenes
	videoScenes->addScene(scene1);
	videoScenes->addScene(scene2);
	videoScenes->addScene(scene3);
	videoScenes->addScene(scene4);
	*/



void 
VideoAbstract::mergeImages(Image<uint8> &src1, Image<uint8> &src2, uint8 percentImage1, Image<uint8> &dest)
{
	int32 v;
	int32 f1 = percentImage1; 
	int32 f2 = 100 - percentImage1; 

	ImageSize size = src1.getSize();
	dest.resize(size);
	for (int band=0; band<size[DIM_BAND]; band++) {
		for (int y=0; y<size[DIM_Y]; y++) {
			for (int x=0; x<size[DIM_X]; x++) {
				v = (f1*src1(x,y,band) + f2*src2(x,y,band)) / 100;
				if (v>255) v=255;
				if (v<0) v=0;
				dest(x,y,band) = v;
			} // for
		} // for
	} // for

	return;
} // mergeImages

void
VideoAbstract::createFadeIn (VideoDecoder &reader, int32 firstFrame, int16 numberOfFrames, Video<uint8> &vid, uint8 color) 
{
	if (numberOfFrames<1) numberOfFrames=1;
	VideoSize size (srcSize[DIM_X], srcSize[DIM_Y], numberOfFrames, srcSize[DIM_BAND]);
	vid.resize(size);
	Image<uint8> tmp;
	for (int z=0; z<numberOfFrames; z++) {

		// read frame
		tmp = reader.getImage(z+firstFrame);

		// calculate fade in
		int16 framePart = z * 120 / numberOfFrames;
		int16 colorPart = 120 - framePart;
		for (int32 b=0; b<size[DIM_BAND]; b++) {
			for (int32 y=0; y<size[DIM_Y]; y++) {
				for (int32 x=0; x<size[DIM_X]; x++) {
					tmp(x,y,b) = (uint8)((colorPart*color + framePart*tmp(x,y,b)) / 120);					
				} // for
			} // for
		} // for

		// store frame
		vid.setImage (z, tmp);
	} // for

	return;
} // createFadeIn


void
VideoAbstract::createFadeOut (VideoDecoder &reader, int32 lastFrame, int16 numberOfFrames, Video<uint8> &vid, uint8 color) 
{
	if (numberOfFrames<1) numberOfFrames=1;
	VideoSize size (srcSize[DIM_X], srcSize[DIM_Y], numberOfFrames, srcSize[DIM_BAND]);
	vid.resize(size);
	Image<uint8> tmp;
	for (int z=0; z<numberOfFrames; z++) {

		// read frame
		tmp = reader.getImage(z+lastFrame-numberOfFrames);

		// calculate fade out
		int16 colorPart = (z+1) * 120 / numberOfFrames;
		int16 framePart = 120 - colorPart;
		for (int32 b=0; b<size[DIM_BAND]; b++) {
			for (int32 y=0; y<size[DIM_Y]; y++) {
				for (int32 x=0; x<size[DIM_X]; x++) {
					tmp(x,y,b) = (uint8)((colorPart*color + framePart*tmp(x,y,b)) / 120);					
				} // for
			} // for
		} // for

		// store frame
		vid.setImage (z, tmp);
	} // for

	return;
} // createFadeOut

void
VideoAbstract::createDissolve (VideoDecoder &reader, int32 lastFrameFirstScene, int32 firstFrameSecondScene, int16 numberOfFrames, Video<uint8> &vid) 
{
	if (numberOfFrames<1) numberOfFrames=1;
	VideoSize size (srcSize[DIM_X], srcSize[DIM_Y], numberOfFrames, srcSize[DIM_BAND]);
	vid.resize(size);
	Image<uint8> img;

	// read first video
	for (int z=0; z<numberOfFrames; z++) {
		vid.setImage(z, reader.getImage(z+lastFrameFirstScene-numberOfFrames+1));
	} // for

	for (int z=0; z<numberOfFrames; z++) {
		// read frame
		img = reader.getImage(z+firstFrameSecondScene);

		// calculate fade in
		// example: fade over 5 frames, percent of first scene: 10%, 30%, 50%, 70%, 90%
		//          difference = 20%, start value = difference / 2 = 10%
		int16 part2 = z * 120 / numberOfFrames + 60 / numberOfFrames;
		int16 part1 = 120 - part2;
		for (int32 b=0; b<size[DIM_BAND]; b++) {
			for (int32 y=0; y<size[DIM_Y]; y++) {
				for (int32 x=0; x<size[DIM_X]; x++) {
					vid(x,y,z,b) = (uint8)((part1*vid(x,y,z,b) + part2*img(x,y,b)) / 120);					
				} // for
			} // for
		} // for
	} // for

	return;
} // createDissolve


void 
VideoAbstract::createDissolve (Image<uint8> &img1, Image<uint8> &img2, int16 numberOfFrames, Video<uint8> &vid)
{
	if (numberOfFrames<1) numberOfFrames=1;
	ImageSize imgSize = img1.getSize();
	VideoSize size (imgSize[DIM_X], imgSize[DIM_Y], numberOfFrames, imgSize[DIM_BAND]);
	vid.resize(size);

	// store images
	for (int z=0; z<numberOfFrames; z++) {

		int16 part2 = z * 120 / numberOfFrames + 60 / numberOfFrames;
		int16 part1 = 120 - part2;
		for (int32 b=0; b<size[DIM_BAND]; b++) {
			for (int32 y=0; y<size[DIM_Y]; y++) {
				for (int32 x=0; x<size[DIM_X]; x++) {
					vid(x,y,z,b) = (uint8)((part1*img1(x,y,b) + part2*img2(x,y,b)) / 120);					
				} // for
			} // for
		} // for
	} // for

	return;
} // createDissolve


void 
VideoAbstract::createMpegVideo() 
{
	if (videoScenes.audioSelectedFrames < 1) {
		cerr << endl << "VideoAbstract: no frames selected" << endl << endl;
		exit(1);
	} // if
	
	// process video 
	videoFrames = 0;
	try {

		// open reader
		int cacheSize=3;
		VideoCodec codec=CODEC_RGB;
		VideoDecoder reader (srcFilename, cacheSize, lastFrameToDecode, codec);
		ieee64 framerate = reader.getFrameRate();

		// increase bitrate by 10 percent
		ieee64 bitrate = reader.getBitRate();
		bitrate*=1.1;

		bool cropTo16=true;

		// open writer
		OMpegVideo writer (destMv1File, imgSize, framerate, bitrate, createGrayscale, cropTo16);


		// process selected frames
		Image<uint8> img;
		cerr << endl;
		int32 lastFrame = 0;
		for (int32 frame=0; frame<videoScenes.audioSelectedFramesList.getSize()[DIM_X]; frame++) {

			// frame selected?
			if (videoScenes.audioSelectedFramesList(frame) == 1) { 
				videoFrames++;

				// read and store frame
				img = reader.getImage(frame);
				writer.write (img);
				lastFrame = frame;
				//cerr << "\rstore frame: " << frame << "  " << ends;
			} // if
		} // for

		// close process
		writer.close();
		//reader.close();

    } catch (...) {
        cerr << endl << "VideoAbstract: could not process video!" << endl;
        exit (1);
    } // try

	string timecode;
	videoScenes.getTimeCode (videoFrames, timecode);
	cerr << endl << "length of abstract: " << timecode << "  (" << videoFrames << " frames)" << endl;
	cerr << endl;
	
	return;
} // createMpegVideo


/*
			// calculate and store "fade in"
			if ((*it).mode[0] == SCENE_CHANGE_FADE) {
				Video<uint8> vid;
				createFadeIn (reader, (*it).frame[0], (*it).sceneChangeLength[0], vid, (*it).fadeColor[0]);
				cerr << "frames: " << ((*it).frame[0])<< ".." << ((*it).frame[0]+vid.getSize()[DIM_Z])
					<< " ["<< (vid.getSize()[DIM_Z]) << "] \t\t [fade in  ]" << endl;
				for (int32 z=0; z<vid.getSize()[DIM_Z]; z++) {
					videoFrames++;
					writer.write (vid.getImage(z));
				} // for
			} // if

			// calculate and store frames with "no effect"
			int32 noEffectStart = (*it).frame[0];
			int32 noEffectEnd = (*it).frame[1];
			if ((*it).mode[0] != SCENE_CHANGE_HARD_CUT) {
				noEffectStart += (*it).sceneChangeLength[0];
			} // if
			if ((*it).mode[1] != SCENE_CHANGE_HARD_CUT) {
				noEffectEnd -= (*it).sceneChangeLength[1];
			} // if
			cerr << "frames: " << noEffectStart << ".." << noEffectEnd << " [" << (noEffectEnd-noEffectStart+1) <<"] \t\t [no effect]" << endl;
			for (int32 z=noEffectStart; z<=noEffectEnd; z++) {
				videoFrames++;
				writer.write (reader.getImage(z));
			} // for


			// calculate and store frames at the end of the scene
			// (1) fade out
			if ((*it).mode[1] == SCENE_CHANGE_FADE) {
				Video<uint8> vid;
				createFadeOut (reader, (*it).frame[1], (*it).sceneChangeLength[1], vid, (*it).fadeColor[1]);
				cerr << "frames: " << ((*it).frame[1]-(*it).sceneChangeLength[1]+1) << ".." << ((*it).frame[1]) 
					<< " [" << (vid.getSize()[DIM_Z]) << "] \t\t [fade out ]" << endl;
				for (int32 z=0; z<vid.getSize()[DIM_Z]; z++) {
					videoFrames++;
					writer.write (vid.getImage(z));
				} // for
			} // if

			// (2) dissolve
			if ((*it).mode[1] == SCENE_CHANGE_DISSOLVE) {

				// get next scene description
				list<SceneDescription>::iterator nextSceneIt = it;
				nextSceneIt++;
				if (nextSceneIt == videoScenes.scenes.end()) {
					cerr << endl << "VideoAbstract: last scene ends with a dissolve"<< endl << endl;
					exit(1);
				} // if

				// calculate dissolve
				Video<uint8> vid;
				createDissolve (reader, (*it).frame[1], (*nextSceneIt).frame[0], (*it).sceneChangeLength[1], vid);
				cerr << "frames: " << ((*it).frame[1]-(*it).sceneChangeLength[1]+1) << ".." << (*it).frame[1] << ","
					<< ((*nextSceneIt).frame[0]) << ".." << ((*nextSceneIt).frame[0]+(*it).sceneChangeLength[1]-1) 
					<< " [" << (vid.getSize()[DIM_Z]) << "] \t [dissolve ]" << endl;
				for (int32 z=0; z<vid.getSize()[DIM_Z]; z++) {
					videoFrames++;
					writer.write (vid.getImage(z));
				} // for
			} // if
*/



/*
		for (int32 i=0; i<srcSize[DIM_Z]; i++) {

			// no effect
			if (selectedFrames(i) == 1) { // TODO
				cerr << "select frame: " << i << "    " << endl;
				videoFrames++;
				writer.write (reader.getImage(i));
			} // if

			// dissolve
			if (selectedFrames(i) == 2) {
				int32 start  = i;
				int32 stop   = i;
				int32 start2 = i;
				int32 stop2  = i;
				while (selectedFrames(i) != 3) {
					i++;
				} // while
				stop = i;
				while (selectedFrames(i) != 4) {
					i++;
				} // while
				start2 = i;
				int32 length = stop-start;
				if (length < 1) length = 1;
				stop2 = start2 + length;
				cout << "dissolve frames: [" << start << "," << stop-1 << "][" << start2 << "," << stop2-1 << "]" << endl;

				// allocate memory for images
				if (maxDissolveSize < length) {

					// delete old images
					if (scene1 != NULL) {
						delete[] scene1;
						scene1 = NULL;
					} // if
					if (scene2 != NULL) {
						delete[] scene2;
						scene2 = NULL;
					} // if

					// allocate memory for images
					maxDissolveSize = length;
					scene1 = new Image<uint8>[maxDissolveSize];
					scene2 = new Image<uint8>[maxDissolveSize];
				} // if

				// read images
				for (int j=start; j<stop; j++) {
					scene1[j-start] = reader.getImage(j);
				} // for
				for (int j=start2; j<stop2; j++) {
					scene2[j-start2] = reader.getImage(j);
				} // for

				// merge images
				for (int j=0; j<length; j++) {
					cerr << "merge frames: [" << j+start << "," << j+start2 << "]" << endl;
					ieee64 percentImage1 = 100;
					if (length > 0) {
						percentImage1 = 100.0 * (length - j) / (length+1);
					} // if
					if (percentImage1 < 0)   percentImage1 = 0;
					if (percentImage1 > 100) percentImage1 = 100;

					mergeImages(scene1[j], scene2[j], (uint8) percentImage1, img);
					videoFrames++;
					writer.write (img);
				} // for

				i=stop2-1;
			} // if

		} // for
*/

	// delete old images
/*	if (scene1 != NULL) {
		delete[] scene1;
		scene1 = NULL;
	} // if
	if (scene2 != NULL) {
		delete[] scene2;
		scene2 = NULL;
	} // if
*/	



void 
VideoAbstract::createMpegAudio() 
{
	
	if (noAudio) {
	    cerr << "audio track not found!" << endl;
	    return;
	} // if

		
	// store the selected samples as WAV file
	int32 totalSamples = 0;
	WaveHeader audioHeader;
	try {
		// open reader
		IWaveAudio reader (srcAudioWavFile);
		audioHeader = reader.getHeader ();
		int32 srcAudioSamples = audioHeader.getNumberOfSamples();

		
		// set audio size
		ieee64 samplesPerSecond = audioHeader.getSamplesPerSec();
		int32 totalAudioSamples = (int32) (videoScenes.audioSelectedSeconds * samplesPerSecond);
		if (totalAudioSamples > srcAudioSamples) totalAudioSamples = srcAudioSamples;
		
		// set destination header
		WaveHeader destAudioHeader;
		int32 dataLength = 2 * audioHeader.getChannels() * totalAudioSamples;
		destAudioHeader.setHeader (audioHeader.getChannels(), audioHeader.getSamplesPerSec(), audioHeader.getBitsPerSample(), dataLength);
		
		// debug information
		//ieee64 realAudioSize = audioSize[DIM_X]/samplesPerSecond;
		//ieee64 realVideoSize = srcSeconds;
		
		// copy audio samples
		FILE *fh_in =NULL;
		FILE *fh_out=NULL;
		OpenFile::process(&fh_out, destWavFile, F_WRITE);
		if (fh_out==NULL) {
		    cerr << "VideoAbstract: could not open temporary WAV destination file." << endl;
		} // if

		int32 counter_out = 0;		
		counter_out = fwrite ((char*)destAudioHeader.getRiffBuffer(), 1, WAV_HEADER_SIZE, fh_out);
		if (counter_out != WAV_HEADER_SIZE) {
			cerr << "VideoAbstract: could not create wav header!" << endl;
			exit(1);
		} // if
		
		uint8 *buf = NULL;
		buf = new uint8 [1024];
		if (buf == NULL) {
			cerr << endl << "VideoAbstract: cannot allocate memory for audio buffer." << endl << endl;
			exit(1);
		} // if
		
		
		int32 numberOfFrames = videoScenes.audioSelectedFramesList.getSize()[DIM_X];
		int32 i=0;
		cout << "store audio track:" << endl;
		bool endOfWavFileReached=false;
		while (i<numberOfFrames) {
			
			while ((i<numberOfFrames) && (videoScenes.audioSelectedFramesList(i)!=1)) i++;
			int32 firstFrame = i;
			while ((i<numberOfFrames) && (videoScenes.audioSelectedFramesList(i)==1)) i++;
			
			if (firstFrame < i-1) {
				
				// get start/stop position
				ieee64 startFrame=firstFrame;
				ieee64 lastFrame =i;
				ieee64 startFrameTime = startFrame / frameRate;
				ieee64 lastFrameTime  = lastFrame  / frameRate;
				int32 startSample = (int32)(startFrameTime * samplesPerSecond);
				int32 lastSample  = (int32)(lastFrameTime  * samplesPerSecond);
				
				if (totalSamples+lastSample-startSample > totalAudioSamples) {
					int32 diffSamples=totalSamples+lastSample-startSample-totalAudioSamples;
					lastSample -= diffSamples;
				} // if
				totalSamples+=lastSample-startSample;
				int32 totalBytes = 2 * (lastSample-startSample) * audioHeader.getChannels();
				
				// select data
				fh_in = reader.getFileHandle (startSample);
				while (totalBytes>0) {
					int32 readBytes=1024;
					if (readBytes>totalBytes) readBytes=totalBytes;
					totalBytes-=readBytes;
					
					// read data
					int32 counter_in = 0;		
					counter_in = fread(buf,1, readBytes, fh_in);	
					if (counter_in != readBytes) {
						for (int i=counter_in; i<1024; i++) buf[i]=0;
						counter_in=readBytes;
						endOfWavFileReached = true;
					} // if
					
					// write data
					counter_out = 0;		
					counter_out = fwrite ((char*)buf, 1, counter_in, fh_out);
					if (counter_out != counter_in) {
						cerr << "VideoAbstract: could not copy wav audio file!" << endl;
						exit(1);
					} // if
				} // while
				
			} // if
		} // while
#if 0
		if (endOfWavFileReached) {
			cerr << "VideoAbstract: end of WAV file reached (replace last samples with zeros)" << endl;
		} // if
#endif
		
		fflush (fh_out);
		fclose (fh_out);
		if (buf!=NULL) delete[] buf;
		buf=NULL;
	} catch (...) {
		cerr << "VideoAbstract: could not process audio (WAV), disable audio." << endl;
		noAudio=true;
		return;
	} // try


	// process audio (PCM to MP2)
	try {
		OMpegAudio::Wav2Mp2 (destWavFile, destMp2File, audioHeader.getSamplesPerSec(), audioHeader.getChannels());
	} catch (...) {
		cerr << "VideoAbstract: could not process audio (MP2), disable audio." << endl;
		noAudio=true;
		return;
	} // try

	return;
} // createMpegAudio

void 
VideoAbstract::multiplexAudioVideo() 
{
	if (noAudio) { 
		// rename file
		FileHandling::renameFile (destMv1File, destFilename);

	} else {

		// multiplex audio/video
		try {
			OMpegMplex::process(destMv1File, destMp2File, destFilename);
			
		} catch (...) {
			cerr << endl << "VideoAbstract: could not process video." << endl;
			exit (1);
		} // try
	} // else


	return;
} // multiplexAudioVideo


void 
VideoAbstract::process() 
{
	cerr << "select frames..." << endl;
	selectFrames();
	
	if (createAbstract) {
		cerr << endl << "process video..." << endl;
		createMpegVideo();
		cerr << "process video...done" << endl;

		cerr << endl << "process audio..." << endl;
		createMpegAudio();
		cerr << "process audio...done" << endl;
		
		cerr << endl << "multiplexing audio/audio..." << ends;
		multiplexAudioVideo();
		cerr << "\rmultiplexing audio/audio...done" << endl << endl;
	} // if

	return;
} // process

