/*
   This file is part of the MoCA video library.
   Name: ConvolutionFilter.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: {kuehne,kopf}@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/* -*- C++ -*-

   ConvolutionFilter.h  -- implementation of a convolution filter

   Author:    Stephan Kopf, Gerald Kuehne
              Lehrstuhl fuer Praktische Informatik IV
              Universitaet Mannheim, Germany
              email: {kuehne,kopf}@pi4.informatik.uni-mannheim.de

*/



#ifndef CONVOLUTION_FILTER_H
#define CONVOLUTION_FILTER_H

#include <SampleBlock.h>
#include <MocaSignal.h>
#include <FilterBase.h>

/*
  Signal<ieee64> maskData : contains the values of the convolution mask
  Directions direction    : direction of the convolution, e. g. DIR_X, DIR_XYZ, ... (see enum.h)
  bool scale              : scales image, if the sum of maskData is not zero => the sum of maskData is scaled to 1
  ieee64 scalingFactor    : if (scale==true) && (scalingFactor != 1) => this scalingFactor is used
  SampleBlock<T> src      : source image
  SampleBlock<T> dest     : destination image
  T offset                : add offset value to image, e.g. wit mask=(-1,0,1) to avoid negative values, set offset=128
                            and scalingFactor=4 
  
  
  maskData is a signal with the values of the convolution mask, e. g. maskData = (0.25, 0.5, 0,25)

  Setting scale = true, the values in maskData are normalized, so than
  SUM (maskData(0)+...+maskData(n)) == 1

  Example:  ~Gauss filter (convolutionMask = [1,2,1] )
  ConvolutionFilter<uint8> cf (convolutionMask, DIR_XY, true);
  cf.process(src, dest);

  Example: ~"signed" Derivation filter (convolutionMask = [-1,0,1], scalingFactor=2)
  ConvolutionFilter<int16> cf (convolutionMask, DIR_X, true, scalingFactor);
  cf.process(src, dest);

  Example: ~"unsigned" Derivation filter (convolutionMask = [-1,0,1], scalingFactor=4, offset=128="uint8"/2)
  ConvolutionFilter<uint8> cf (convolutionMask, DIR_X, true, scalingFactor);
  cf.process(src, dest, offset);
  
*/

enum {CONVOLUTION_NOSCALE=-1, CONVOLUTION_AUTOSCALE=-2};

template<class T>
class ConvolutionFilter : public FilterBase<T>
{
public:
    ConvolutionFilter (Signal<ieee64> &maskData, Directions direction= DIR_XY, ieee64 scalingFactor=CONVOLUTION_NOSCALE)
    {
        maskSize = maskData.getSize()[DIM_X];
        checkValue (maskSize % 2, "ConvolutionFilter: Can only process odd mask sizes");

        this->direction = direction;
        
        mask = new ieee64[maskSize];
        checkMemory (mask);

        sizeType i;
        ieee64 scaleF = 0.0;
        for (i = 0; i < maskSize; i++) {
            mask[i] = (ieee64) maskData(i);
            scaleF += mask[i];
        } // for
        
        if (scalingFactor != CONVOLUTION_NOSCALE) {
            if (scalingFactor==CONVOLUTION_AUTOSCALE) {
                if (scaleF != 0) {
                    for (i = 0; i < maskSize; i++) {
                        mask[i] /= scaleF;       
                    } // for
                } // if
            } else {
                if (scalingFactor != 0) {
                    for (i = 0; i < maskSize; i++) {
                        mask[i] /= scalingFactor;       
                    } // for
                } // if
            } // else
        } // if

/*        cout << endl << "mask: "; 
        for (i = 0; i < maskSize; i++) {
            cout << mask[i] << "  ";
        } // for
        cout << endl;
*/
        
        maskRadius = maskSize/2;        
        mask += maskRadius;        
    };
    
    ~ConvolutionFilter ()
    {
        delete [] (mask - maskRadius);
    };

    //! Apply convolution filter to src, write out into dest
    void process (const SampleBlock<T>& src, SampleBlock<T>& dest);

    //! Apply convolution filter to src, write into same image
    void process (SampleBlock<T>& srcDest);

    //! Apply convolution filter to src, write out into dest
    void process (const SampleBlock<T>& src, SampleBlock<T>& dest, T offset);

    //! Apply convolution filter to src, write into same image
    void process (SampleBlock<T>& srcDest, T offset);

    //! Returns the filter mask size
    sizeType getMaskSize() const { return maskSize; }

    //! Gives access to the mask entries
    const ieee64* getMaskPtr() { return mask - maskRadius; }

protected:
    //! Filter from src (type T) to dest (type T)
    void filterLineSD (const T* srcP,  const offsetType* srcOffsetP,
                     ieee64* destP, const offsetType* destOffsetP,
                     sizeType size, T offset);

    //! Filter from src (type T) to dest (type T)
    void filterLineD (ieee64* srcP,  const offsetType* srcOffsetP,
                     ieee64* destP, const offsetType* destOffsetP,
                     sizeType size, T offset);

    //! the mask (e.g. 1 2 1)
    ieee64* mask;

    //! size of mask
    sizeType maskSize;
    
    //! half of maskSize
    sizeType maskRadius;

    //! convolution direction: DIR_X, DIR_Y, DIR_Z
    Directions direction;
    
    //! temporary storage (needed for inplace filtering)
    ieee64* srcLineS;
    ieee64* srcLineD;
}; // ConvolutionFilter

#include <ConvolutionFilter.cpt>

#endif

