/*
   This file is part of the MoCA video library.
   Name: DownSampleFilter.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: {kuehne,kopf}@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/* -*- C++ -*-

   DownSampleFilter.h  -- samples a SampleBlock down

   Author:    Gerald Kuehne
              Lehrstuhl fuer Praktische Informatik IV
              Universitaet Mannheim, Germany
              email: {kuehne,kopf}@pi4.informatik.uni-mannheim.de

   $Id: DownSampleFilter.h,v 1.2 2004/03/02 17:04:11 kopf Exp $
*/



#ifndef DOWNSAMPLEFILTER_H
#define DOWNSAMPLEFILTER_H

// C++ includes
#include <vector>

// MoCA includes
#include <Image.h>
#include <Video.h>
#include <MocaEnum.h>
#include <MocaExceptions.h>
#include <SampleBlock.h>





/*!
  \brief Provides functions for downsampling an image or a video

  The class provides element functions for downsampling an image or
  a video by a factor of 2. It provides also a function to create a
  pyramid, i.e. a collection of layers differing in resolution.

  \remarks Currently we are using a Gaussian mask (3 pixel
           size) for preprocessing.
*/

template<class T>
class DownSampleFilter 
{
public:
    //! The constructor
    //DownSampleFilter(int numberOfPyramids, uint32 direction = (DIR_X | DIR_Y | DIR_Z));
    DownSampleFilter(int numberOfPyramids, uint32 direction);
    //! The destructor
    ~DownSampleFilter();

    //! Samples an image or video down by a factor of 2
    void process(const SampleBlock<T>* src, SampleBlock<T>* dest);
    

    
    //! Creates a pyramid by downsampling an image multiple times
    void process(const Image<T>&, ImageVector<T>&) throw (MocaException);
    //! Creates a pyramid by downsampling a video multiple times
    void process(const Video<T>&, VideoVector<T>&) throw (MocaException);

    
protected:
    //! Calculates the size of the downsampled image or video
    void calculateSize (const SampleBlockSize&, SampleBlockSize&);

    //! Filters one sample block line (datatype S -> datatype D)
    void filterLineSD (const T* srcP,  const offsetType* srcOffsetP,
                       T* destP, const offsetType* destOffsetP,
                       sizeType);


    //! Filters one sample block line (datatype D -> datatype D)
    void filterLineD (T* srcP,  const offsetType* srcOffsetP,
                      T* destP, const offsetType* destOffsetP,
                      sizeType);


    //! the number of pyramid images
	int numberOfPyramids;

    //! Contains the downsampling directions (e.g. DIR_X | DIR_Y)
    uint32 directions;

    //! The Gaussian mask for preprocessing (e.g. 1 2 1)
    double* mask;

    //! The radius of the Gaussian mask
    sizeType maskRadius;

    //! Temporary storage needed for inplace filtering of a sample line
    uint8* srcLine;
}; // DownSampleFilter

#include <DownSampleFilter.cpt>

#endif

