/*
   This file is part of the MoCA video library.
   Name: RotationFilter.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


/* -*- C++ -*-

   RotationFilter.h  -- implementation of the gauss filter

   Author:    Stephan Kopf
              Lehrstuhl fuer Praktische Informatik IV
              Universitaet Mannheim, Germany
              email: kopf@pi4.informatik.uni-mannheim.de

*/



#ifndef ROTATION_FILTER_H
#define ROTATION_FILTER_H

// C includes
#include <math.h>

// MoCA includes
#include <ImageFilterBase.h>
#include <FlipFilter.h>
#include <TransposeFilter.h>
#include <ShearFilter.h>
#include <CropFilter.h>
#include <Image.h>


/*
  Rotates an image by an arbitrary angle.

  The rotation algorithm is adapted from
  "A Fast Algorithm for General Raster Rotation"
  by Alan Paeth, Graphics Interface '86, pp. 77-81.
*/

template<class T>
class RotationFilter : public ImageFilterBase<T>
{
private:    
    static const ieee64 TOLERANCE;
public:
    static const ieee64 PI;
    
public:  

	// use the following parameter with the constructor:
	// ieee64 angle = 45.0;
	// RotationFilter<uint8> rf (angle * FACE_DETECTION_PI/180.0);
	// rf.process (img);
    RotationFilter (ieee64 angle) {
        this->angle = angle;        
    } // constructor

    //! Apply rotation filter to src, write out into dest
    void process (const Image<T>& src, Image<T>& dest);

    //! Apply rotation filter to src, write into same image
    void process (Image<T>& srcDest)
	{
		Image<T> tmp = srcDest;
		process (tmp, srcDest);
		return;
	}

private:

    // Rotate an image by an angle in the interval [-Pi/2, +Pi/2].
    void rotate (const Image<T>& src, Image<T>& dest, ieee64 angle);
    
    //! rotation angle
    ieee64 angle;

}; // RotationFilter

#include <RotationFilter.cpt>

#endif

