/*
    ConvertColorSpace.cpp     convert images from one color space to another
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



#include <ConvertColorSpace.h>


// clip values to 0..255
inline int clip(int x) { if (x<0) return 0; if (x>255) return 255; return x; }


// convert image to image, available modes: Y/YUV/RGB/YUV8
void 
ConvertColorSpace::process(VideoCodec srcMode, VideoCodec destMode, const Image<uint8> &src, Image<uint8> &dest)
{
	bool error = false;

	// color modes: RGB, YUV444, YUV422, YUV420, Y, YUV8 

	Image<uint8> tmp;
	switch (srcMode) {

	case CODEC_RGB:
		switch (destMode) {
		case CODEC_RGB:
			dest=src;
			break;
		case CODEC_YUV444:
			RGB_YUV444 (src, dest);
			break;
		case CODEC_YUV422:
			RGB_YUV444 (src, tmp);
			YUV444_YUV422 (tmp, dest);
			break;
		case CODEC_YUV420:
			RGB_YUV444 (src, tmp);
			YUV444_YUV420 (tmp, dest);
			break;
		case CODEC_Y:
			RGB_Y (src, dest);
			break;
		case CODEC_YUV8:
			RGB_YUV444 (src, tmp);
			YUV444_YUV8 (tmp, dest);
			break;
		default:
			error = true;
		}
		break;

	case CODEC_YUV444:
		switch (destMode) {
		case CODEC_YUV444:
			dest=src;
			break;
		case CODEC_RGB:
			YUV444_RGB (src, dest);
			break;
		case CODEC_YUV422:
			YUV444_YUV422 (src, dest);
			break;
		case CODEC_YUV420:
			YUV444_YUV420 (src, dest);
			break;
		case CODEC_Y:
			YUV444_Y (src, dest);
			break;
		case CODEC_YUV8:
			YUV444_YUV8 (src, dest);
			break;
		default:
			error = true;
		}
		break;

	case CODEC_YUV422:
		switch (destMode) {
		case CODEC_YUV422:
			dest=src;
			break;
		case CODEC_RGB:
			YUV422_YUV444 (src, tmp);
			YUV444_RGB (tmp, dest);
			break;
		case CODEC_YUV444:
			YUV422_YUV444 (src, dest);
			break;
		case CODEC_YUV420:
			YUV422_YUV444 (src, tmp);
			YUV444_YUV420 (tmp, dest);
			break;
		case CODEC_Y:
			YUV422_Y (src, dest);
			break;
		case CODEC_YUV8:
			YUV422_YUV444 (src, tmp);
			YUV444_YUV8 (tmp, dest);
			break;
		default:
			error = true;
		}
		break;

	case CODEC_YUV420:
		switch (destMode) {
		case CODEC_YUV420:
			dest=src;
			break;
		case CODEC_RGB:
			YUV420_YUV444 (src, tmp);
			YUV444_RGB (tmp, dest);
			break;
		case CODEC_YUV444:
			YUV420_YUV444 (src, dest);
			break;
		case CODEC_YUV422:
			YUV420_YUV444 (src, tmp);
			YUV444_YUV422 (tmp, dest);
			break;
		case CODEC_Y:
			YUV420_Y (src, dest);
			break;
		case CODEC_YUV8:
			YUV420_YUV444 (src, tmp);
			YUV444_YUV8 (tmp, dest);
			break;
		default:
			error = true;
		}
		break;

	case CODEC_Y:
		switch (destMode) {
		case CODEC_Y:
			dest=src;
			break;
		case CODEC_RGB:
			Y_RGB (src, dest);
			break;
		case CODEC_YUV444:
			Y_YUV444 (src, dest);
			break;
		case CODEC_YUV422:
			Y_YUV444 (src, tmp);
			YUV444_YUV422 (tmp, dest);
			break;
		case CODEC_YUV420:
			Y_YUV444 (src, tmp);
			YUV444_YUV420 (tmp, dest);
			break;
		case CODEC_YUV8:
			Y_YUV444 (src, tmp);
			YUV444_YUV8 (tmp, dest);
			break;
		default:
			error = true;
		}
		break;

	case CODEC_YUV8:
		switch (destMode) {
		case CODEC_YUV8:
			dest=src;
			break;
		case CODEC_RGB:
			YUV8_YUV444 (src, tmp);
			YUV444_RGB (tmp, dest);
			break;
		case CODEC_YUV444:
			YUV8_YUV444 (src, dest);
			break;
		case CODEC_YUV422:
			YUV8_YUV444 (src, tmp);
			YUV444_YUV422 (tmp, dest);
			break;
		case CODEC_YUV420:
			YUV8_YUV444 (src, tmp);
			YUV444_YUV420 (tmp, dest);
			break;
		case CODEC_Y:
			YUV8_YUV444 (src, tmp);
			YUV444_Y (tmp, dest);
			break;
		default:
			error = true;
		}
		break;

	default:
		error = true;
	}

	if (error) {
		cerr << "ConvertColorSpace::process:conversion not possible!" << endl;
		exit(1);
	}

	return;
}



// -------------------------YUV conversion --------------------------------------// 


void 
ConvertColorSpace::YUV444_RGB (const Image<uint8> &src, Image<uint8> &dest)
{
	// set size of destination image
	ImageSize size=src.getSize();
	size[DIM_BAND]=3;
	dest.resize(size);
	int32 sx = size[DIM_X];
	int32 sy = size[DIM_Y];

	// set values of destination image
	for (int32 y=0; y<sy; y++) {

		const uint8* yp = src.getSamplePtr(0,y,0);
		const uint8* up = src.getSamplePtr(0,y,1);
		const uint8* vp = src.getSamplePtr(0,y,2);

		uint8* rp = dest.getSamplePtr(0,y,0);
		uint8* gp = dest.getSamplePtr(0,y,1);
		uint8* bp = dest.getSamplePtr(0,y,2);

		for (int32 x=0; x<sx; x++) {
			int yy = 298*((*yp) - 16);
			int uu =      (*up) -128;
			int vv =      (*vp) -128;
			
			// clipping is required
			*rp = clip( (          409*vv + yy)>>8 );
			*gp = clip( (-100*uu - 208*vv + yy)>>8 );
			*bp = clip( ( 517*uu          + yy)>>8 );

			yp++; up++; vp++;
			rp++; gp++; bp++;

		} // for
	} // for
	return;
}


void 
ConvertColorSpace::RGB_YUV444 (const Image<uint8> &src, Image<uint8> &dest)
{
	// set size of destination image
	ImageSize size=src.getSize();
	size[DIM_BAND]=3;
	dest.resize(size);
	int32 sx = size[DIM_X];
	int32 sy = size[DIM_Y];

	// set values of destination image
	for (int32 y=0; y<sy; y++) {

		const uint8* rp = src.getSamplePtr(0,y,0);
		const uint8* gp = src.getSamplePtr(0,y,1);
		const uint8* bp = src.getSamplePtr(0,y,2);

		uint8* yp = dest.getSamplePtr(0,y,0);
		uint8* up = dest.getSamplePtr(0,y,1);
		uint8* vp = dest.getSamplePtr(0,y,2);

		for (int32 x=0; x<sx; x++) {

			// clipping is not required
			*yp++ = (( 66*(*rp) +  129*(*gp) +  25*(*bp) +  4096) >>8);
			*up++ = ((-38*(*rp) +  -74*(*gp) + 112*(*bp) + 32768) >>8);
			*vp++ = ((112*(*rp) +  -94*(*gp) + -18*(*bp) + 32768) >>8);

			rp++; gp++; bp++;
		} // for
	} // for
	return;
}



void 
ConvertColorSpace::YUV444_YUV422 (const Image<uint8> &src, Image<uint8> &dest)
{
	// set size of destination image
	ImageSize size=src.getSize();
	size[DIM_BAND]=3;
	dest.resize(size);

	int32 sx = size[DIM_X];
	int32 dx = (sx/2)*2;
	int32 sy = size[DIM_Y];


	// set values of destination image
	for (int32 y=0; y<sy; y++) {

		const uint8* y0 = src.getSamplePtr(0,y,0);

		const uint8* u00 = src.getSamplePtr(0,y,1);
		const uint8* v00 = src.getSamplePtr(0,y,2);
		const uint8* u01 = u00+1;
		const uint8* v01 = v00+1;

		uint8* y1 = dest.getSamplePtr(0,y,0);
		uint8* u1 = dest.getSamplePtr(0,y,1);
		uint8* v1 = dest.getSamplePtr(0,y,2);

		for (int32 x=0; x<dx; x+=2) {

			// copy 2 pixels
			*y1++ = *y0++;
			*y1++ = *y0++;

			// copy 1 pixel for u and v planes
			*u1++ = ((*u00) + (*u01)) >> 1;
			*v1++ = ((*v00) + (*v01)) >> 1;

			// set pointers to next pixel
			u00+=2;
			u01+=2;
			v00+=2;
			v01+=2;
		} // for
	} // for

	// handle odd number of columns
	if (sx != dx) {
		int32 px0 = dx/2;
		int32 px1 = dx;
		for (int32 y=0; y<sy; y++) {

			// full resolution
			dest(px1,y,0) = src (px1,y,0);

			// half resolution
			dest(px0,y,1) = src (px1,y,1);
			dest(px0,y,2) = src (px1,y,2);
		} // for
	} // if
	return;
}

void 
ConvertColorSpace::YUV422_YUV444 (const Image<uint8> &src, Image<uint8> &dest)
{
	// set size of destination image
	ImageSize size=src.getSize();
	size[DIM_BAND]=3;
	dest.resize(size);

	int32 sx = size[DIM_X];
	int32 dx = (sx/2)*2;
	int32 sy = size[DIM_Y];

	// set values of destination image
	for (int32 y=0; y<sy; y++) {

		const uint8* y0 = src.getSamplePtr(0,y,0);
		const uint8* u0 = src.getSamplePtr(0,y,1);
		const uint8* v0 = src.getSamplePtr(0,y,2);

		uint8* y1 = dest.getSamplePtr(0,y,0);
		uint8* u1 = dest.getSamplePtr(0,y,1);
		uint8* v1 = dest.getSamplePtr(0,y,2);

		for (int32 x=0; x<dx; x+=2) {

			// copy 2 pixels
			*y1++ = *y0++;
			*y1++ = *y0++;

			// copy pixel 2 times for u and v planes
			*u1++ = *u0;
			*u1++ = *u0++;
			*v1++ = *v0;
			*v1++ = *v0++;
		} // for
	} // for

	// handle odd number of columns
	if (sx != dx) {
		int32 px0 = dx;
		int32 px1 = dx/2;
		for (int32 y=0; y<sy; y++) {

			// full resolution
			dest(px0,y,0) = src (px0,y,0);

			// half resolution
			dest(px0,y,1) = src (px1,y,1);
			dest(px0,y,2) = src (px1,y,2);
		} // for
	} // if

	return;
}


void 
ConvertColorSpace::YUV444_YUV420 (const Image<uint8> &src, Image<uint8> &dest)
{
	// set size of destination image
	ImageSize size=src.getSize();
	size[DIM_BAND]=3;
	dest.resize(size);

	int32 sx = size[DIM_X];
	int32 sy = size[DIM_Y];
	int32 dx = (sx/2)*2;
	int32 dy = (sy/2)*2;

	// set values of destination image
	for (int32 y=0; y<dy; y+=2) {

		const uint8* y0  = src.getSamplePtr(0,y,0);
		const uint8* u00 = src.getSamplePtr(0,y,1);
		const uint8* u01 = u00+1;
		const uint8* v00 = src.getSamplePtr(0,y,2);
		const uint8* v01 = v00+1;

		const uint8* y1  = src.getSamplePtr(0,y+1,0);
		const uint8* u10 = src.getSamplePtr(0,y+1,1);
		const uint8* u11 = u10+1;
		const uint8* v10 = src.getSamplePtr(0,y+1,2);
		const uint8* v11 = v10+1;

		uint8* y2 = dest.getSamplePtr(0,y,0);
		uint8* u2 = dest.getSamplePtr(0,y/2,1);
		uint8* v2 = dest.getSamplePtr(0,y/2,2);

		uint8* y3 = dest.getSamplePtr(0,y+1,0);

		for (int32 x=0; x<dx; x+=2) {

			// copy 2 pixels to first line
			*y2++ = *y0++;
			*y2++ = *y0++;

			// copy 2 pixels to second line
			*y3++ = *y1++;
			*y3++ = *y1++;

			// copy 1 pixel for u plane
			*u2++ = ((*u00) + (*u01) + (*u10) + (*u11)) >> 2;
			u00+=2; u01+=2; u10+=2; u11+=2; 

			// copy 1 pixel for v plane
			*v2++ = ((*v00) + (*v01) + (*v10) + (*v11)) >> 2;
			v00+=2; v01+=2; v10+=2; v11+=2; 


		} // for
	} // for

	// handle odd number of columns
	if (sx != dx) {
		int32 px0 = dx/2;
		int32 px1 = dx;

		for (int32 y=0; y<dy; y+=2) {

			// full resolution
			dest(px1,y  ,0) = src (px1,y  ,0);
			dest(px1,y+1,0) = src (px1,y+1,0);

			// half resolution
			int32 py0 = y/2;
			dest(px0,py0,1) = (src (px1,y,1) + src (px1,y+1,1)) >> 1;
			dest(px0,py0,2) = (src (px1,y,2) + src (px1,y+1,2)) >> 1;

		} // for
	} // if


	// handle odd number of rows
	if (sy != dy) {
		int32 py0 = dy/2;
		int32 py1 = dy;

		for (int32 x=0; x<dx; x+=2) {

			// full resolution
			dest(x,  py1,0) = src (x,  py1,0);
			dest(x+1,py1,0) = src (x+1,py1,0);

			// half resolution
			int32 px0 = x/2;
			dest(px0,py0,1) = (src (x,py1,1) + src (x+1,py1,1)) >> 1;
			dest(px0,py0,2) = (src (x,py1,2) + src (x+1,py1,2)) >> 1;
		} // for
	} // if

	// handle odd number of rows and columns (lower right corner)
	if ((sx != dx) && (sy != dy)) {
		dest(dx,dy,0)     = src(dx,dy,0);
		dest(dx/2,dy/2,1) = src(dx,dy,1);
		dest(dx/2,dy/2,2) = src(dx,dy,2);
	} // if
	
	return;
}

void 
ConvertColorSpace::YUV420_YUV444 (const Image<uint8> &src, Image<uint8> &dest)
{
	// set size of destination image
	ImageSize size=src.getSize();
	size[DIM_BAND]=3;
	dest.resize(size);

	int32 sx = size[DIM_X];
	int32 sy = size[DIM_Y];
	int32 dx = (sx/2)*2;
	int32 dy = (sy/2)*2;

	// set values of destination image
	for (int32 y=0; y<dy; y+=2) {

		const uint8* y0 = src.getSamplePtr(0,y,0);
		const uint8* u0 = src.getSamplePtr(0,y/2,1);
		const uint8* v0 = src.getSamplePtr(0,y/2,2);

		const uint8* y1 = src.getSamplePtr(0,y+1,0);

		uint8* y2 = dest.getSamplePtr(0,y,0);
		uint8* u2 = dest.getSamplePtr(0,y,1);
		uint8* v2 = dest.getSamplePtr(0,y,2);

		uint8* y3 = dest.getSamplePtr(0,y+1,0);
		uint8* u3 = dest.getSamplePtr(0,y+1,1);
		uint8* v3 = dest.getSamplePtr(0,y+1,2);

		for (int32 x=0; x<dx; x+=2) {

			// copy 2 pixels to first line
			*y2++ = *y0++;
			*y2++ = *y0++;

			// copy 2 pixels to first line
			*y3++ = *y1++;
			*y3++ = *y1++;

			// copy pixel 4 times for u and v planes
			*u2++ = *u0;
			*u2++ = *u0;
			*u3++ = *u0;
			*u3++ = *u0;
			u0++;

			*v2++ = *v0;
			*v2++ = *v0;
			*v3++ = *v0;
			*v3++ = *v0;
			v0++;
		} // for
	} // for

	// handle odd number of columns
	if (sx != dx) {
		int32 px0 = dx;
		int32 px1 = dx/2;
		for (int32 y=0; y<sy; y+=2) {

			// full resolution
			dest(px0,y,0)   = src (px0,y,0);
			dest(px0,y+1,0) = src (px0,y+1,0);

			// half resolution
			dest(px0,y,1)   = src (px1,y/2,1);
			dest(px0,y+1,1) = src (px1,y/2,1);
			dest(px0,y,2)   = src (px1,y/2,2);
			dest(px0,y+1,2) = src (px1,y/2,2);
		} // for
	} // if

	

	// handle odd number of rows
	if (sy != dy) {
		int32 py0 = dy;
		int32 py1 = dy/2;
		for (int32 x=0; x<sx; x+=2) {

			// full resolution
			dest(x,  py0,0) = src (x,  py0,0);
			dest(x+1,py0,0) = src (x+1,py0,0);

			// half resolution
			dest(x,  py0,1) = src (x/2, py1,1);
			dest(x+1,py0,1) = src (x/2, py1,1);
			dest(x,  py0,2) = src (x/2, py1,2);
			dest(x+1,py0,2) = src (x/2, py1,2);
		} // for
	} // if

	

	// handle odd number of rows and columns (lower right corner)
	if ((sx != dx) && (sy != dy)) {
		dest(dx,dy,0)     = src(dx,dy,0);
		dest(dx,dy,1) = src(dx/2,dy/2,1);
		dest(dx,dy,2) = src(dx/2,dy/2,2);
	} // if
	
	
	return;
}

// -------------------------Y conversion --------------------------------------// 

void 
ConvertColorSpace::Y_YUV444 (const Image<uint8> &src, Image<uint8> &dest)
{
	// set size of destination image
	ImageSize size=src.getSize();
	size[DIM_BAND]=3;
	dest.resize(size);

	// set values of destination image
	for (int32 y=0; y<size[DIM_Y]; y++) {

		const uint8* y0 = src.getSamplePtr(0,y,0);
		uint8* y1 = dest.getSamplePtr(0,y,0);
		uint8* u1 = dest.getSamplePtr(0,y,1);
		uint8* v1 = dest.getSamplePtr(0,y,2);

		for (int32 x=0; x<size[DIM_X]; x++) {
			*y1++ = *y0;
			y0++;
			*u1++ = 128;
			*v1++ = 128;
		} // for
	} // for
	return;
}

void 
ConvertColorSpace::Y_YUV422 (const Image<uint8> &src, Image<uint8> &dest)
{
	Y_YUV444 (src, dest);
	return;
}

void 
ConvertColorSpace::Y_YUV420 (const Image<uint8> &src, Image<uint8> &dest)
{
	Y_YUV444 (src, dest);
	return;
}

void 
ConvertColorSpace::Y_RGB (const Image<uint8> &src, Image<uint8> &dest)
{
	// set size of destination image
	ImageSize size=src.getSize();
	size[DIM_BAND]=3;
	dest.resize(size);

	// set values of destination image
	for (int32 y=0; y<size[DIM_Y]; y++) {

		const uint8* y0 = src.getSamplePtr(0,y,0);
		uint8* r1 = dest.getSamplePtr(0,y,0);
		uint8* g1 = dest.getSamplePtr(0,y,1);
		uint8* b1 = dest.getSamplePtr(0,y,2);

		for (int32 x=0; x<size[DIM_X]; x++) {
			*r1++ = *y0;
			*g1++ = *y0;
			*b1++ = *y0;
			y0++;
		} // for
	} // for
	return;
}

void 
ConvertColorSpace::YUV444_Y (const Image<uint8> &src, Image<uint8> &dest)
{
	// set size of destination image
	ImageSize size=src.getSize();
	size[DIM_BAND]=1;
	dest.resize(size);

	// set values of destination image
	for (int32 y=0; y<size[DIM_Y]; y++) {

		const uint8* y0 = src.getSamplePtr(0,y,0);
		uint8* y1 = dest.getSamplePtr(0,y,0);

		for (int32 x=0; x<size[DIM_X]; x++) {
			*y1++ = *y0;
			y0++;
		} // for
	} // for
	return;
}

void 
ConvertColorSpace::YUV422_Y (const Image<uint8> &src, Image<uint8> &dest)
{
	YUV444_Y (src, dest);
	return;
}

void 
ConvertColorSpace::YUV420_Y (const Image<uint8> &src, Image<uint8> &dest)
{
	YUV444_Y (src, dest);
	return;
}


void 
ConvertColorSpace::RGB_Y (const Image<uint8> &src, Image<uint8> &dest)
{
	// set size of destination image
	ImageSize size=src.getSize();
	size[DIM_BAND]=1;
	dest.resize(size);

	// set values of destination image
	for (int32 y=0; y<size[DIM_Y]; y++) {

		const uint8* r0 = src.getSamplePtr(0,y,0);
		const uint8* g0 = src.getSamplePtr(0,y,1);
		const uint8* b0 = src.getSamplePtr(0,y,2);
		uint8* y1 = dest.getSamplePtr(0,y,0);

		for (int32 x=0; x<size[DIM_X]; x++) {

			*y1++ = (( 66*(*r0) +  129*(*g0) +  25*(*b0) +  4096) >>8);
			r0++; g0++; b0++;
		} // for

	} // for
	return;
}


// -------------------------YUV8 conversion --------------------------------------// 
// 8 bits are stored for each puixel in the YUV8 colorspace:
//      4 most significant Y (luminance) bits, 
//      2 U and 2 V most significant (chrominance) bits
// the YUV8 colorspace is only used for computation (e.g., region growing)
// the YUV8 colorspace should not be used to display an image (very low visual quality)
void 
ConvertColorSpace::YUV8_YUV444 (const Image<uint8> &src, Image<uint8> &dest)
{
	// set size of destination image
	ImageSize size=src.getSize();
	size[DIM_BAND]=3;
	dest.resize(size);

	// set values of destination image
	for (int32 y=0; y<size[DIM_Y]; y++) {

		const uint8* y0 = src.getSamplePtr(0,y,0);

		uint8* y1 = dest.getSamplePtr(0,y,0);
		uint8* u1 = dest.getSamplePtr(0,y,1);
		uint8* v1 = dest.getSamplePtr(0,y,2);

		for (int32 x=0; x<size[DIM_X]; x++) {
			*y1++ = ((*y0 & 0x0f) << 4) + 8;
			*u1++ = ((*y0 & 0x30) << 2) + 32;
			*v1++ = ((*y0 & 0xc0)     ) + 32;
			y0++;
		} // for
	} // for
	return;
}


void 
ConvertColorSpace::YUV444_YUV8 (const Image<uint8> &src, Image<uint8> &dest)
{
	// set size of destination image
	ImageSize size=src.getSize();
	size[DIM_BAND]=1;
	dest.resize(size);

	// set values of destination image
	for (int32 y=0; y<size[DIM_Y]; y++) {

		const uint8* y0 = src.getSamplePtr(0,y,0);
		const uint8* u0 = src.getSamplePtr(0,y,1);
		const uint8* v0 = src.getSamplePtr(0,y,2);
		uint8* y1 = dest.getSamplePtr(0,y,0);

		for (int32 x=0; x<size[DIM_X]; x++) {

			*y1++ =  (*y0 >> 4)  +  ((*u0 & 0xc0) >> 2)  +  (*v0 & 0xc0);
			y0++; u0++; v0++;
		} // for
	} // for
	return;
}


// -------------------------copy raw data to linear buffer -------------------------// 

void 
ImageBuffer::getImageSize (VideoCodec codec, ImageSize &size, int32 &sx, int32 &sy, int32 &ux, int32 &uy)
{
	// set dimension for Y plane
	sx = size[DIM_X];
	sy = size[DIM_Y];

	// set dimension for UV plane
	ux = 0;
	uy = 0;
	switch (codec) {
	case CODEC_RGB:
	case CODEC_YUV444:
		ux = sx;
		uy = sy;
		break;
	case CODEC_YUV422:
		ux = ((sx+1)/2);
		uy = sy;
		break;
	case CODEC_YUV420:
		ux = ((sx+1)/2);
		uy = ((sy+1)/2);
		break;
	case CODEC_Y:
	case CODEC_YUV8:
		break;
	default:
		cerr << "ImageBuffer::videocodec is unknown!" << endl;
		exit(1);
	}

	return;
}


void 
ImageBuffer::setImage (VideoCodec codec, const Image<uint8> &src)
{
	ptr = NULL;	
	size = src.getSize();
	this->codec = codec;

	// get image size
	int32 sx, sy, ux, uy;
	getImageSize (codec, size, sx,sy,ux,uy);


	// set buffer size
	bufSize = sx*sy + 2*(ux*uy);

		
	SignalSize destSize (bufSize,1);
	img.resize(destSize);
	ptr = img.getSamplePtr (0,0);

	// copy Y data
	uint8* ip = img.getSamplePtr(0,0);
	for (int32 y=0; y<sy; y++) {

		const uint8* sp = src.getSamplePtr(0,y,0);
		for (int32 x=0; x<sx; x++) {
			*ip++ = *sp;
			sp++;
		} // for
	} // for

	// copy UV data
	for (int32 b=1; b<size[DIM_BAND]; b++) {
		for (int32 y=0; y<uy; y++) {
			
			const uint8* sp = src.getSamplePtr(0,y,b);
			for (int32 x=0; x<ux; x++) {
				*ip++ = *sp;
				sp++;
			} // for
		} // for
	} // for

	return;
}


void 
ImageBuffer::getImage (VideoCodec destCodec, Image<uint8> &dest)
{
	dest.resize(size);

	// get image size
	int32 sx, sy, ux, uy;
	getImageSize (codec, size, sx,sy,ux,uy);


	// copy Y data
	const uint8* ip = img.getSamplePtr(0,0);
	for (int32 y=0; y<sy; y++) {

		uint8* sp = dest.getSamplePtr(0,y,0);
		for (int32 x=0; x<sx; x++) {
			*sp++ = *ip;
			ip++;
		} // for
	} // for

	// copy UV data
	for (int32 b=1; b<size[DIM_BAND]; b++) {
		for (int32 y=0; y<uy; y++) {
			
			uint8* sp = dest.getSamplePtr(0,y,b);
			for (int32 x=0; x<ux; x++) {
				*sp++ = *ip;
				ip++;
			} // for
		} // for
	} // for

	// convert to destination color space
	ConvertColorSpace::process(codec,destCodec, dest);
	return;
}
