/*
   This file is part of the MoCA video library.
   Name: ConvertColorSpace.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/* -*- C++ -*-

   ConvertColorSpace.h  -- convert images from one color space to another

   Author:    Stephan Kopf
              Lehrstuhl fuer Praktische Informatik IV
              Universitaet Mannheim, Germany
              email: kopf@pi4.informatik.uni-mannheim.de

*/



#ifndef CONVERT_COLOR_SPACE_H
#define CONVERT_COLOR_SPACE_H

#include <MocaTypes.h>
#include <MocaEnum.h>
#include <Image.h>
#include <OImageStream.h>

struct ImageBuffer
{
public:
	ImageBuffer () {
		ptr = NULL;
		bufSize = 0;
		codec = CODEC_RGB;
	}
	void setImage (VideoCodec codec, const Image<uint8> &img);
	void getImage (VideoCodec codec, Image<uint8> &img);

private:
	void getImageSize (VideoCodec codec, ImageSize &size, int32 &sx, int32 &sy, int32 &ux, int32 &uy);

	Signal<uint8> img; // raw data
	ImageSize size;    // original image size
	VideoCodec codec;  // original codec

public:
	int32 bufSize;	   // size of buffer
	uint8 *ptr;		   // pointer to data
};


class ConvertColorSpace
{
public:

	// convert image to image, 
	// color modes: RGB, YUV444, YUV422, YUV420, Y 
	static void process(VideoCodec srcMode, VideoCodec destMode, Image<uint8> &srcDest)
	{
		Image<uint8> tmp = srcDest;
		process (srcMode, destMode, tmp, srcDest);
		return;
	}

	// convert image to image, 
	// color modes: RGB, YUV444, YUV422, YUV420, Y 
	static void process(VideoCodec srcMode, VideoCodec destMode, const Image<uint8> &src, Image<uint8> &dest);


private:

	// YUV conversion
	// YUV444: full resolution in x and y direction for UV planes
	// YUV422: full resolution in y direction for UV planes, half resolution in x direction
	//         (only a part of the pixels are used in the UV bands:
	//             if (BAND==1) || (BAND==2): image size: [0..(DIM_X+1)/2] x [0..DIM_Y]
	// YUV420: half resolution in x and y direction for UV planes
	//         (only a part of the pixels are used in the UV bands:
	//             if (BAND==1) || (BAND==2): image size: [0..(DIM_X+1)/2] x [0..(DIM_Y+1)/2]
	static void RGB_YUV444    (const Image<uint8> &src, Image<uint8> &dest);
	static void YUV444_RGB    (const Image<uint8> &src, Image<uint8> &dest);
	static void YUV444_YUV422 (const Image<uint8> &src, Image<uint8> &dest);
	static void YUV422_YUV444 (const Image<uint8> &src, Image<uint8> &dest);
	static void YUV444_YUV420 (const Image<uint8> &src, Image<uint8> &dest);
	static void YUV420_YUV444 (const Image<uint8> &src, Image<uint8> &dest);

	// Y conversion
	static void Y_YUV444 (const Image<uint8> &src, Image<uint8> &dest);
	static void Y_YUV422 (const Image<uint8> &src, Image<uint8> &dest);
	static void Y_YUV420 (const Image<uint8> &src, Image<uint8> &dest);
	static void Y_RGB    (const Image<uint8> &src, Image<uint8> &dest);
	static void YUV444_Y (const Image<uint8> &src, Image<uint8> &dest);
	static void YUV422_Y (const Image<uint8> &src, Image<uint8> &dest);
	static void YUV420_Y (const Image<uint8> &src, Image<uint8> &dest);
	static void RGB_Y    (const Image<uint8> &src, Image<uint8> &dest);

	// YUV8 conversion
	static void YUV8_YUV444 (const Image<uint8> &src, Image<uint8> &dest);
	static void YUV444_YUV8 (const Image<uint8> &src, Image<uint8> &dest);
};

#endif

