/*
    IYuvVideo.cpp     read YUV (raw) videos
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <IYuvVideo.h>



IYuvVideo::IYuvVideo (const string &pathFile, int32 width, int32 height, VideoCodec cMode, bool interleave, bool silent)
{
	pathFilename        = pathFile;
	videoSize[DIM_X]    = width;
	videoSize[DIM_Y]    = height;
	videoSize[DIM_Z]    = 1;
	videoSize[DIM_BAND] = 1;
	codec = cMode;
	fh_in = NULL;
	getUVplanes = false;
	this->interleave = interleave;
	this->silent     = silent;
	
	bufferSize = 0;
	switch (codec) {
	case CODEC_Y:
		bufferSize = width*height;
		videoSize[DIM_BAND] = 1;
		getUVplanes = false;
		break;
	case CODEC_YUV420:
		bufferSize = width*height + 2*((width+1)/2)*((height+1)/2);
		videoSize[DIM_BAND] = 3;
		getUVplanes = true;
		break;
	case CODEC_YUV422:
		bufferSize = 2*width*height;
		videoSize[DIM_BAND] = 3;
		getUVplanes = true;
		break;
	case CODEC_YUV444:
		bufferSize = 3*width*height;
		videoSize[DIM_BAND] = 3;
		getUVplanes = true;
		break;
	case CODEC_RGB:
		cerr << "IYuvVideo::RGB video mode not available!" << endl;
		exit(1);
		break;
	case CODEC_YUV8:
		cerr << "IYuvVideo::YUV8 video mode not available!" << endl;
		exit(1);
		break;
	default:
		cerr << "IYuvVideo::unknown video mode!" << endl;
		exit(1);
		break;
	} // switch
	

	// set image size
	imgSize[DIM_X]    = videoSize[DIM_X];
	imgSize[DIM_Y]    = videoSize[DIM_Y];
	imgSize[DIM_Z]    = 1;
	imgSize[DIM_BAND] = videoSize[DIM_BAND];
	
	// get size of video
	int32 sizeInBytes=0;
	FileHandling::getFileSize (pathFilename, sizeInBytes);
	videoSize[DIM_Z] = sizeInBytes / bufferSize;
	if (sizeInBytes % bufferSize != 0) {
		if (!silent) cerr << "IYuvVideo: Warning: the size of the YUV-video is not correct!" << endl << endl;
	} // if
}



IYuvVideo::IYuvVideo (const string &pathFile)
{
	pathFilename = pathFile;
	interleave = false;
	silent = false;

	// get size of video
	int32 sizeInBytes=0;
	FileHandling::getFileSize (pathFilename, sizeInBytes);

	// get video size
	int32 width=0;
	int32 height=0;
	VideoCodec c = CODEC_YUV420;
	ieee64 dist=10e10;
	getResolution(sizeInBytes, width, height, c, dist, interleave);


	// set default values
	const ieee64 DEFAULT_DIST = 64.0;
	if (dist > DEFAULT_DIST) {
		cerr << "IYuvVideo: Error: the size of the YUV-video could not be identified automatically!" << endl;
		cerr << "Specify the exact resolution, please!" << endl;
		exit(1);
	} // if
	

	videoSize[DIM_X]    = width;
	videoSize[DIM_Y]    = height;
	videoSize[DIM_Z]    = 1;
	videoSize[DIM_BAND] = 1;
	codec = c;
	fh_in = NULL;
	getUVplanes = false;
	
	bufferSize = 0;
	switch (codec) {
	case CODEC_Y:
		bufferSize = width*height;
		videoSize[DIM_BAND] = 1;
		getUVplanes = false;
		break;
	case CODEC_YUV420:
		bufferSize = width*height + 2*((width+1)/2)*((height+1)/2);
		videoSize[DIM_BAND] = 3;
		getUVplanes = true;
		break;
	case CODEC_YUV422:
		bufferSize = 2*width*height;
		videoSize[DIM_BAND] = 3;
		getUVplanes = true;
		break;
	case CODEC_YUV444:
		bufferSize = 3*width*height;
		videoSize[DIM_BAND] = 3;
		getUVplanes = true;
		break;
	default:
		cerr << "IYuvVideo::unknown video mode!" << endl;
		exit(1);
		break;
	} // switch

	// set image size
	imgSize[DIM_X]    = videoSize[DIM_X];
	imgSize[DIM_Y]    = videoSize[DIM_Y];
	imgSize[DIM_Z]    = 1;
	imgSize[DIM_BAND] = videoSize[DIM_BAND];
	
	videoSize[DIM_Z] = sizeInBytes / bufferSize;

	cerr << "video resolution: " 
		 << videoSize[DIM_X] << " x " << videoSize[DIM_Y] << " x " << videoSize[DIM_Z];
	switch (codec) {
	case CODEC_Y:
		cerr << " (grayscale)";
		break;
	case CODEC_YUV420:
		cerr << " (YUV420-color)";
		break;
	case CODEC_YUV422:
		cerr << " (YUV422-color)";
		break;
	case CODEC_YUV444:
		cerr << " (YUV444-color)";
		break;
	default:
		cerr << " (unknown)";
		exit(1);
		break;
	} // switch
	cerr << " ("<< dist << ")" << endl;
	if (sizeInBytes % bufferSize != 0) {
		if (!silent) cerr << "IYuvVideo: Warning: the size of the YUV-video is not correct!" << endl << endl;
	} // if
}



void
IYuvVideo::getResolution(int32 sizeInBytes, int32 &width, int32 &height, VideoCodec &codec, ieee64 &diff, bool &interleave)
{
	const int32 RES_COUNT = 29;
	const int32 defaultResolutions[RES_COUNT][2] = {
		{2048,1024},{1600,960},
		{800,600},
		{768,576},{768,480},{720,576},{720,486},{720,480},{704,576},{704,480},{704,288},
		{672,488},{640,512},{640,480},
		{512,400},
		{480,576},{480,384},{400,320},
		{384,288},{360,288},{352,576},{352,288},{352,240},{320,256},{320,240},
		{256,256},{200,160},
		{192,144},{176,144},
	};

	int32  optW          = defaultResolutions[0][0];
	int32  optH          = defaultResolutions[0][1];
	ieee64 optDiff       = 10e10;
	bool optInterleave = false;
	VideoCodec optC = CODEC_Y;


	for (int32 r=0; r<RES_COUNT; r++) {
		for (int32 i=0; i<=3; i++) {
			for (int32 l=0; l<=1; l++) {
				
				VideoCodec testC = CODEC_Y;
				if (i==1) testC = CODEC_YUV420;
				if (i==2) testC = CODEC_YUV422;
				if (i==3) testC = CODEC_YUV444;
				
				bool testI = false;
				if (l==1) testI = true;
				
				ieee64 testDiff=10e10;
				
				// check video sizes
				if (checkSize (defaultResolutions[r][0], defaultResolutions[r][1], testC, testI, sizeInBytes, testDiff)) {
					if (optDiff > testDiff) {
						optW = defaultResolutions[r][0];
						optH = defaultResolutions[r][1];
						optDiff = testDiff;
						optC = testC;
						optInterleave = testI;
					} // if
				} // if
			} // for
		} // for
	} // for


	width  = optW;
	height = optH;
	codec  = optC;
	diff   = optDiff;
	interleave = optInterleave;

	// correct resolution not found?
	if (diff > 64) {

		for (int32 r=0; r<RES_COUNT; r++) {
			cerr << "\rtest resolution: " << defaultResolutions[r][0] << " x " << defaultResolutions[r][1] << "    " << ends;
			for (int32 i=0; i<=3; i++) {
				for (int32 l=0; l<=1; l++) {
					
					VideoCodec testC = CODEC_Y;
					if (i==1) testC = CODEC_YUV420;
					if (i==2) testC = CODEC_YUV422;
					if (i==3) testC = CODEC_YUV444;
					
					bool testI = false;
					if (l==1) testI = true;
					
					// assume one resolution and decode images
					ieee64 testDiff=10e10;
					getImgDifference (
						pathFilename, 
						defaultResolutions[r][0], defaultResolutions[r][1], testC, testI, 
						testDiff) ;

					// get difference
					if (optDiff > testDiff) {
						optW = defaultResolutions[r][0];
						optH = defaultResolutions[r][1];
						optDiff = testDiff;
						optC = testC;
						optInterleave = testI;
					} // if
				} // for
			} // for
		} // for
		cerr << endl;

		width  = optW;
		height = optH;
		codec  = optC;
		diff   = optDiff;
		interleave = optInterleave;
	} // if
	return;
} 


void 
IYuvVideo::getImgDifference (string &fn, int32 width, int32 height, VideoCodec c, bool interleave, ieee64 &diff) 
{
	diff = 10e10;

	IYuvVideo reader (fn, width, height, c, interleave, true);
	Image<uint8> img0, img1;
	reader.getImage(0, img0);
	reader.getImage(1, img1);
	
	ImageSize s = img0.getSize();
	int32 w = s[DIM_X]/2;
	int32 h = s[DIM_X]/2;
	if (w>32) w=32;
	if (h>32) h=32;
	int32 offX = s[DIM_X]-w;
	int32 offY = s[DIM_Y]-h;

	int32 d=0;
	for (int32 b=0; b<s[DIM_BAND]; b++) {
		for (int32 y=0; y<h; y++) {
			for (int32 x=0; x<w; x++) {
				d+=labs (((int32) img0(x,y,b)) - img1(x,y,b));
				d+=labs (((int32) img0(offX+x,y,b)) - img1(offX+x,y,b));
				d+=labs (((int32) img0(offX+x,offY+y,b)) - img1(offX+x,offY+y,b));
				d+=labs (((int32) img0(x,offY+y,b)) - img1(x,offY+y,b));
			} // for
		} // for
	} // for
	ieee64 f = 4 * w * h * s[DIM_BAND];
	if (f<1) f=1;
	diff = ((ieee64)d) / f;

	return;
}

bool 
IYuvVideo::checkSize (int32 width, int32 height, VideoCodec c, bool interleave, int32 vidSize, ieee64 &imgDifference) 
{
	bool resOkay = false;
	int32 testS = 99999999;
	
	if (c == CODEC_Y) {
		testS = width*height;
	} // if
	if (c == CODEC_YUV420) {
		testS = width*height + 2*((width+1)/2)*((height+1)/2);
	} // if
	if (c == CODEC_YUV422) {
		testS = 2*width*height;
	} // if
	if (c == CODEC_YUV444) {
		testS = 3*width*height;
	} // if
	if (vidSize % testS == 0) resOkay = true;
	
	imgDifference = 10e10;
	if (resOkay) getImgDifference (pathFilename, width, height, c, interleave, imgDifference);

	return resOkay;
}


void
IYuvVideo::getImage(int32 frame, Image<uint8> &img)
{
	if (frame<0) frame=0;
	if (frame>=videoSize[DIM_Z]) frame=videoSize[DIM_Z]-1;

	// open file
	open();

	// seek file position
	long seekPos = bufferSize*frame;
	int code = fseek(fh_in, seekPos, SEEK_SET);
	if (code != 0) {
		cerr << "IYuvVideo: Cannot seek file position!" << endl;
		exit(1);
	} // if

	// initialize image
	img.resize(imgSize);

	// get Y plane
	uint8 buf;
	for (int32 y=0; y<imgSize[DIM_Y]; y++) {
		for (int32 x=0; x<imgSize[DIM_X]; x++) {
			fread(&buf,1, 1, fh_in);
			img(x,y,0) = buf;
		} // for
	} // for

	
	if (getUVplanes) {

		int32 width  = imgSize[DIM_X];
		int32 height = imgSize[DIM_Y];
		if (codec == CODEC_YUV420) {
			width /=2;
			height /=2;
		} // if
		if (codec == CODEC_YUV422) {
			width /=2;
		} // if

		if (interleave) {

			// get U and V plane
			for (int32 y=0; y<height; y++) {
				for (int32 x=0; x<width; x++) {
					fread(&buf,1, 1, fh_in);
					img(x,y,1) = buf;
					fread(&buf,1, 1, fh_in);
					img(x,y,2) = buf;
				} // for
			} // for

		} else {
			// get U plane
			for (int32 y=0; y<height; y++) {
				for (int32 x=0; x<width; x++) {
					fread(&buf,1, 1, fh_in);
					img(x,y,1) = buf;
				} // for
			} // for
			
			// get V plane
			for (int32 y=0; y<height; y++) {
				for (int32 x=0; x<width; x++) {
					fread(&buf,1, 1, fh_in);
					img(x,y,2) = buf;
				} // for
			} // for
		} // else

	} // if

	return;
} // getImage


void
IYuvVideo::getVideo(Video<uint8> &vid)
{
	// initialize video
	try {
		vid.resize(videoSize);
	} catch(...) {
		cerr << "IYuvVideo::callon allocate memory for video" << endl;
		exit(1);
	} // catch
	
	// open file
	open();


	// seek file position
	bool error = false;
	for (int32 z=0; z<videoSize[DIM_Z]; z++) {

		try {
			long seekPos = bufferSize*z;
			int code = fseek(fh_in, seekPos, SEEK_SET);
			if (code != 0) error = true;
		} catch (...) {
			error = true;
		} // catch
		if (error) {
			cerr << "IYuvVideo: Cannot seek file position!" << endl;
			exit(1);
		} // if


		try {
			uint8 buf;
			// get Y plane
			for (int32 y=0; y<imgSize[DIM_Y]; y++) {
				for (int32 x=0; x<imgSize[DIM_X]; x++) {
					fread(&buf,1, 1, fh_in);
					vid(x,y,z,0) = buf;
				} // for
			} // for

			
			if (getUVplanes) {
				
				// get U plane
				for (int32 y=0; y<imgSize[DIM_Y]/2; y++) {
					for (int32 x=0; x<imgSize[DIM_X]/2; x++) {
						fread(&buf,1, 1, fh_in);
						int32 px = 2*x;
						int32 py = 2*y;
						vid(px,  py,  z, 1) = buf;
						vid(px+1,py,  z, 1) = buf;
						vid(px,  py+1,z, 1) = buf;
						vid(px+1,py+1,z, 1) = buf;
					} // for
				} // for
				
				// get V plane
				for (int32 y=0; y<imgSize[DIM_Y]/2; y++) {
					for (int32 x=0; x<imgSize[DIM_X]/2; x++) {
						fread(&buf,1, 1, fh_in);
						int32 px = 2*x;
						int32 py = 2*y;
						vid(px,  py,  z, 2) = buf;
						vid(px+1,py,  z, 2) = buf;
						vid(px,  py+1,z, 2) = buf;
						vid(px+1,py+1,z, 2) = buf;
					} // for
				} // for
			} // if
		} catch (...) {
			error = true;
		} // catch
		if (error) {
			cerr << "IYuvVideo: Cannot read data!" << endl;
			exit(1);
		} // if
	} // for

	return;
} // getVideo

