/*
    WaveHeader.cpp     handle header in WAVE files
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <WaveHeader.h>

//const int WaveHeader::WAV_HEADER_SIZE  =  44;
const int WaveHeader::FORMAT_DATA_SIZE =  16;

const int WaveHeader::FORMAT_TAG_WAVE_PCM  =   1;
const int WaveHeader::FORMAT_TAG__MULAW    = 257;
const int WaveHeader::FORMAT_TAG_IBM_ALAW  = 258;
const int WaveHeader::FORMAT_TAG_IBM_ADPCM = 259;

const int WaveHeader::CHANNELS_MONO   = 1;
const int WaveHeader::CHANNELS_STEREO = 2;


std::ostream& operator<< (ostream& os, const WaveHeader &header)
{
	os << "length after riff header     : " << header.riffLength << endl;
	os << "length of format data        : " << header.fmtLength << endl;
	os << "format tag                   : " << header.formatTag << endl;
	os << "channels                     : " << header.channels << endl;
	os << "samples per second           : " << header.samplesPerSec << endl;
	os << "average bytes per second     : " << header.avgBytesPerSec << endl;
	os << "block alignment              : " << header.blockAlign << endl;
	os << "bits per sample              : " << header.bitsPerSample << endl;
	os << "length of pcm samples [byte] : " << header.dataLength << endl;
	
	return os;
} // operator<<

OTagStream& operator<< (OTagStream& os, const WaveHeader &data)
{

	// video parameters
    TagFormat::writeNodeValue(os, "riffLength",     data.riffLength);
    TagFormat::writeNodeValue(os, "fmtLength",      data.fmtLength);
    TagFormat::writeNodeValue(os, "formatTag",      data.formatTag);
    TagFormat::writeNodeValue(os, "channels",       data.channels);
    TagFormat::writeNodeValue(os, "samplesPerSec",  data.samplesPerSec);
    TagFormat::writeNodeValue(os, "avgBytesPerSec", data.avgBytesPerSec);
    TagFormat::writeNodeValue(os, "blockAlign",     data.blockAlign);
    TagFormat::writeNodeValue(os, "bitsPerSample",  data.bitsPerSample);
    TagFormat::writeNodeValue(os, "dataLength",     data.dataLength);
	os << endl;
	
    return os;
} // operator<< 

ITagStream& operator>> (ITagStream& is, WaveHeader &data)
{
    char c;
    bool failFlag = false;
	
	if (is >> c) {
		is.putback(c);
		
		// video parameters
		TagFormat::readNodeValue(is, "riffLength",     data.riffLength);
		TagFormat::readNodeValue(is, "fmtLength",      data.fmtLength);
		TagFormat::readNodeValue(is, "formatTag",      data.formatTag);
		TagFormat::readNodeValue(is, "channels",       data.channels);
		TagFormat::readNodeValue(is, "samplesPerSec",  data.samplesPerSec);
		TagFormat::readNodeValue(is, "avgBytesPerSec", data.avgBytesPerSec);
		TagFormat::readNodeValue(is, "blockAlign",     data.blockAlign);
		TagFormat::readNodeValue(is, "bitsPerSample",  data.bitsPerSample);
		TagFormat::readNodeValue(is, "dataLength",     data.dataLength);
	} // if
	if (is.rdstate() & ios::failbit) {
		failFlag = true;
	} // if
	
    if (failFlag) {
		throw MocaException ("WaveHeader : Illegal format of index file!");
	} else {
        is.clear();
    } 
    return is;
} // operator>>

// constructor
WaveHeader::WaveHeader () 
{
	riffId[0]  ='R'; riffId[1]='I'; riffId[2]='F'; riffId[3]='F';
	riffLength = WAV_HEADER_SIZE - 8;
	
	// wave id
	wavId[0]='W'; wavId[1]='A'; wavId[2]='V'; wavId[3]='E';
	
	// format header
	fmtId[0]  ='f'; fmtId[1]='m'; fmtId[2]='t'; fmtId[3]=' ';   
	fmtLength = FORMAT_DATA_SIZE;
	
	// format data
	formatTag      = FORMAT_TAG_WAVE_PCM;	
	channels       = CHANNELS_STEREO;	
	samplesPerSec  =  44100;
	avgBytesPerSec = 176400;  // SamplesPerSec x 16bit x 2 (stereo)
	blockAlign     =      4;  // one block = 4 bytes	
	bitsPerSample  =     16;
	
	// data header
	dataHeader[0]='d'; dataHeader[1]='a'; dataHeader[2]='t'; dataHeader[3]='a'; 
	dataLength   = 0;	
	
	updateBuffer ();
} // constructor


// operator=
WaveHeader& 
WaveHeader::operator=(const WaveHeader &other) 
{
	if (this != &other) {
		riffId[0]=other.riffId[0]; riffId[1]=other.riffId[1]; 
		riffId[2]=other.riffId[2]; riffId[3]=other.riffId[3];
		
		riffLength=other.riffLength; 
		
		wavId[0]=other.wavId[0]; wavId[1]=other.wavId[1]; 
		wavId[2]=other.wavId[2]; wavId[3]=other.wavId[3];
		
		fmtId[0]=other.fmtId[0]; fmtId[1]=other.fmtId[1]; 
		fmtId[2]=other.fmtId[2]; fmtId[3]=other.fmtId[3];
		
		fmtLength=other.fmtLength; 
		
		formatTag      = other.formatTag; 
		channels       = other.channels; 
		samplesPerSec  = other.samplesPerSec; 
		avgBytesPerSec = other.avgBytesPerSec; 
		blockAlign     = other.blockAlign; 
		bitsPerSample  = other.bitsPerSample; 
		
		dataHeader[0]=other.dataHeader[0]; dataHeader[1]=other.dataHeader[1]; 
		dataHeader[2]=other.dataHeader[2]; dataHeader[3]=other.dataHeader[3];
		
		dataLength = other.dataLength; 
		
		for (int i=0; i<WAV_HEADER_SIZE; i++) {
			riffBuffer[i]=other.riffBuffer[i]; 
		} // for
	} // if

	return *this;
} // operator=


// setHeader
void 
WaveHeader::setHeader (uint16 channels, uint32 samplesPerSec, uint16 bitsPerSample, uint32 dataLength) 
{
	this->channels       = channels;
	this->samplesPerSec  = samplesPerSec;
	this->bitsPerSample  = bitsPerSample;
	this->dataLength     = dataLength;
	
	riffLength     = dataLength + WAV_HEADER_SIZE - 8;
	blockAlign     = ((bitsPerSample + 7) / 8) * channels;
	avgBytesPerSec = samplesPerSec * blockAlign;
	
	updateBuffer ();
	return;
} // setHeader



// read header information
// return true, if header is valid, else false
bool 
WaveHeader::setHeader (uint8 *buf, bool silent) 
{
	if ((buf[0] != 'R') || (buf[1] != 'I') || (buf[2] != 'F') || (buf[3] != 'F')) {
		if (!silent)
			cerr << endl << "WavHeader: tag RIFF not found." << endl << endl;
		return false;
	} // if
	if ((buf[8] != 'W') || (buf[9] != 'A') || (buf[10] != 'V') || (buf[11] != 'E')) {
		if (!silent)
			cerr << endl << "WavHeader: tag WAVE not found." << endl << endl;
		return false;
	} // if
	if ((buf[12] != 'f') || (buf[13] != 'm') || (buf[14] != 't') || (buf[15] != ' ')) {
		if (!silent)
			cerr << endl << "WavHeader: tag fmt  not found." << endl << endl;
		return false;
	} // if
	if ((buf[36] != 'd') || (buf[37] != 'a') || (buf[38] != 't') || (buf[39] != 'a')) {
		if (!silent)
			cerr << endl << "WavHeader: tag data not found." << endl << endl;
		return false;
	} // if
	
	uint16 channels = 0;
	uint32 samplesPerSec = 0;
	uint16 bitsPerSample = 0;
	uint32 dataLength    = 0;
	
	channels      = buf[22] + ((buf[23]) << 8);
	samplesPerSec = buf[24] + ((buf[25]) << 8) + ((buf[26]) << 16) + ((buf[27]) << 24);
	bitsPerSample = buf[34] + ((buf[35]) << 8);
	dataLength    = buf[40] + ((buf[41]) << 8) + ((buf[42]) << 16) + ((buf[43]) << 24);

	setHeader (channels, samplesPerSec, bitsPerSample, dataLength);
	return true;
} // readHeader



// updateBuffer
void 
WaveHeader::updateBuffer () 
{
	uint32 tmp;
	
	riffBuffer[0]=riffId[0]; riffBuffer[1]=riffId[1]; riffBuffer[2]=riffId[2]; riffBuffer[3]=riffId[3];
	tmp = riffLength;
	riffBuffer[4]=tmp; tmp >>= 8; riffBuffer[5]=tmp; tmp >>= 8;
	riffBuffer[6]=tmp; tmp >>= 8; riffBuffer[7]=tmp; tmp >>= 8;
	
	riffBuffer[8] =wavId[0]; riffBuffer[9]=wavId[1]; riffBuffer[10]=wavId[2]; riffBuffer[11]=wavId[3];
	
	riffBuffer[12]=fmtId[0]; riffBuffer[13]=fmtId[1]; riffBuffer[14]=fmtId[2]; riffBuffer[15]=fmtId[3];
	tmp = fmtLength;
	riffBuffer[16]=tmp; tmp >>= 8; riffBuffer[17]=tmp; tmp >>= 8;
	riffBuffer[18]=tmp; tmp >>= 8; riffBuffer[19]=tmp; tmp >>= 8;
	
	tmp = formatTag;
	riffBuffer[20]=tmp; tmp >>= 8; riffBuffer[21]=tmp; tmp >>= 8;
	tmp = channels;
	riffBuffer[22]=tmp; tmp >>= 8; riffBuffer[23]=tmp; tmp >>= 8;
	tmp = samplesPerSec;
	riffBuffer[24]=tmp; tmp >>= 8; riffBuffer[25]=tmp; tmp >>= 8;
	riffBuffer[26]=tmp; tmp >>= 8; riffBuffer[27]=tmp; tmp >>= 8;
	tmp = avgBytesPerSec;
	riffBuffer[28]=tmp; tmp >>= 8; riffBuffer[29]=tmp; tmp >>= 8;
	riffBuffer[30]=tmp; tmp >>= 8; riffBuffer[31]=tmp; tmp >>= 8;
	tmp = blockAlign;
	riffBuffer[32]=tmp; tmp >>= 8; riffBuffer[33]=tmp; tmp >>= 8;
	tmp = bitsPerSample;
	riffBuffer[34]=tmp; tmp >>= 8; riffBuffer[35]=tmp; tmp >>= 8;
	
	riffBuffer[36] =dataHeader[0]; riffBuffer[37]=dataHeader[1]; riffBuffer[38]=dataHeader[2]; riffBuffer[39]=dataHeader[3];
	tmp = dataLength;
	riffBuffer[40]=tmp; tmp >>= 8; riffBuffer[41]=tmp; tmp >>= 8;
	riffBuffer[42]=tmp; tmp >>= 8; riffBuffer[43]=tmp; tmp >>= 8;
	return;
} // updateBuffer

