/*
   This file is part of the MoCA video library.
   Name: WaveHeader.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/* -*- C++ -*-

 WaveHeader.h -- Interface for handling WAV audio files
 
  Author: Stephan Kopf
  Lehrstuhl fuer Praktische Informatik IV
  Universitaet Mannheim, Germany
  email: kopf@pi4.informatik.uni-mannheim.de
  
   
*/

#ifndef WAVE_HEADER_H
#define WAVE_HEADER_H

// C++ includes
#include <iostream>
using namespace std;

// MoCA includes
#include <MocaTypes.h>
#include <OTagStream.h>


static const int WAV_HEADER_SIZE = 44;

class WaveHeader 
{
public:
	// constant values

private:
	// constant values
	static const int FORMAT_DATA_SIZE;
	
	static const int FORMAT_TAG_WAVE_PCM;
	static const int FORMAT_TAG__MULAW;
	static const int FORMAT_TAG_IBM_ALAW;
	static const int FORMAT_TAG_IBM_ADPCM;
	
	static const int CHANNELS_MONO;
	static const int CHANNELS_STEREO;

public:	
	// constructor
	WaveHeader ();
	
	// destructor
	~WaveHeader (){};

	// operator=
    WaveHeader& operator=(const WaveHeader &other);
	
	// initialize with four values
	void setHeader (uint16 channels, uint32 samplesPerSec, uint16 bitsPerSample, uint32 dataLength);

	// return true, if header is valid, else false,
	// silent: disable error messages
	bool setHeader (uint8 *buf, bool silent=false);

	// operator <<
	friend std::ostream& operator<< (ostream&, const WaveHeader&);

	uint8* getRiffBuffer () {
		return &riffBuffer[0];
	} // getRiffBuffer 

	uint32 getDataLength () {
		return dataLength;
	} // getDataLength 

	uint16 getChannels () {
		return channels;
	} // getChannels 

	uint16 getBlockAlign () {
		return blockAlign;
	} // getBlockAlign 

	uint16 getBitsPerSample () {
		return bitsPerSample;
	} // getBitsPerSample 

	uint32 getSamplesPerSec () {
		return samplesPerSec;
	} // getSamplesPerSec 

	uint32 getNumberOfSamples () {
		uint32 s = dataLength / blockAlign;
		return s;
	} // getNumberOfSamples


private:	
	// updateBuffer
	void updateBuffer ();

public:
	// riff header
	uint8  riffId[4];           // "RIFF"
	uint32 riffLength;          // filelength - 8
	
	// wave id
	uint8  wavId[4];            // "WAVE"
	
	// format header
	uint8  fmtId[4];            // "fmt " 
	uint32 fmtLength;           // length of format data, for PCM wave files should be 16 
	
	// format data
	uint16 formatTag;		    // 1 = WAVE_FORMAT_PCM, 257=FORMAT_MULAW, 258=IBM_FORMAT_ALAW, 259=IBM_FORMAT_ADPCM 
	uint16 channels;		    // mono=1, stereo=2 
	uint32 samplesPerSec;		// samples per second at which channel should be played, e. g. 44100
	uint32 avgBytesPerSec;		// average data per second 
	uint16 blockAlign;			// block alognment in bytes, e. g. 20 bits,mono => use 3 bytes per block
								// e. g. 20 bits,stereo => use 6 bytes per block
	uint16 bitsPerSample;		// format specific for PCM, 20 bits per sample 
	
	// data header
    uint8  dataHeader[4];       // "data"
    uint32 dataLength;			// length of data in bytes
    // raw pcm data follows: signed int

	uint8 riffBuffer [WAV_HEADER_SIZE];

};
std::ostream& operator<< (ostream&, const WaveHeader&);
OTagStream& operator<< (OTagStream& os, const WaveHeader &data);
ITagStream& operator>> (ITagStream& is, WaveHeader &data);

#endif
