/*
   This file is part of the MoCA video library.
   Name: SampleBlock.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: {kuehne,kopf}@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/* -*- C++ -*-

   file.h  -- short description

   Author:    Gerald Kuehne
              Stephan Kopf
              Lehrstuhl fuer Praktische Informatik IV
              Universitaet Mannheim, Germany
              email: {kuehne,kopf}@pi4.informatik.uni-mannheim.de

   $Id: SampleBlock.h,v 1.2 2004/03/05 15:21:24 kopf Exp $
*/

#ifndef SAMPLE_BLOCK_H
#define SAMPLE_BLOCK_H

// C++ includes
#include <list>
using namespace std;

// MoCA includes
#include <SampleBlockSize.h>
#include <SampleBlockPosition.h>
#include <MocaError.h>
#include <SampleBlockBase.h>
#include <MocaExceptions.h>

template <class T>
class SampleBlock : protected SampleBlockBase
{
    friend class SubSampleBlock<T>;

public:
    // Construction/Destruction
    SampleBlock (const SampleBlockSize& size     = SampleBlockSize (0, 0, 0, 0),
                 const SampleBlockSize& tileSize = SampleBlockSize (0, 0, 0, 0),
                 BorderMode borderMode           = BORDER_REPLICATE);
    SampleBlock (const SampleBlock&);

    virtual ~SampleBlock ();

    SampleBlock& operator= (const SampleBlock&);


    // Interface
    void resize  (const SampleBlockSize& size);
    void realloc (const SampleBlockSize& size, const SampleBlockSize& tileSize,
                  BorderMode borderMode);

    SampleBlockSize getSize () const;
    SampleBlockSize getTileSize () const;
    SampleBlockSize getNumberOfTiles () const;

    BorderMode getBorderMode () const;
    void setBorderMode (BorderMode borderMode);

    T* getSamplePtr (indexType x, indexType y, indexType z, indexType band);
    const T* getSamplePtr (indexType x, indexType y, indexType z, indexType band) const;
    T* getTilePtr   (indexType x, indexType y, indexType z, indexType band) const;

    const offsetType& getSampleOffset     (sizeType dimension, indexType i) const;
    const offsetType& getSampleOffsetX    (indexType i) const;
    const offsetType& getSampleOffsetY    (indexType i) const;
    const offsetType& getSampleOffsetZ    (indexType i) const;
    const offsetType& getSampleOffsetBand (indexType i) const;

    const offsetType& getTileOffset     (sizeType dimension) const;
    const offsetType& getTileOffsetX    () const;
    const offsetType& getTileOffsetY    () const;
    const offsetType& getTileOffsetZ    () const;
    const offsetType& getTileOffsetBand () const;

    //! execute function for each position in the sample block
    void forEach(void (*f) (T*));

	// clears the sampleblock with background value
	void clear (T value=(T)0);

	// copy a sampleblock to position, destPos defindes the position in the destination sampleblock
	void copy (const SampleBlock&, const SampleBlockPosition& destPos = SampleBlockPosition (0,0,0,0));

	// copy a sampleblock to position, srcPos defindes the position in the sourcesampleblock, 
	// and destPos defindes the position in the destination sampleblock
	void copy (const SampleBlock&, const SampleBlockPosition& srcPos, const SampleBlockPosition& destPos);


	// copy a sampleblock to position, destPos defindes the position in the destination sampleblock
	// copy all pixels, except pixels with the specified color
	void copyExcept (const SampleBlock&, const SampleBlockPosition& destPos, T color);

	// copy a sampleblock to position, srcPos defindes the position in the sourcesampleblock, 
	// and destPos defindes the position in the destination sampleblock
	// copy all pixels, except pixels with the specified color
	void copyExcept (const SampleBlock&, const SampleBlockPosition& srcPos, const SampleBlockPosition& destPos, T color);


	// find the position of the minimum/maximum value in the sampleblock
	void minValue (SampleBlockPosition &pos);
	void maxValue (SampleBlockPosition &pos);
	T minValue ();
	T maxValue ();


}; // SampleBlock



template<class T>
inline
SampleBlock<T>::
SampleBlock (const SampleBlockSize& size, const SampleBlockSize& tileSize,
             BorderMode borderMode)
        : SampleBlockBase (sizeof(T), size, tileSize, borderMode)
{}

template<class T>
inline
SampleBlock<T>::
SampleBlock (const SampleBlock& other)
        : SampleBlockBase  (other)
{}

template<class T>
SampleBlock<T>::
~SampleBlock ()
{}

template<class T>
inline
SampleBlock<T>&
SampleBlock<T>::operator= (const SampleBlock<T>& other)
{
    SampleBlockBase::operator= (other);
    return *this;
}

template<class T>
inline
void
SampleBlock<T>::
resize (const SampleBlockSize& size)
{
    SampleBlockBase::resize (size);
}


template<class T>
inline
void
SampleBlock<T>::
realloc (const SampleBlockSize& size, const SampleBlockSize& tileSize,
         BorderMode borderMode)
{
    SampleBlockBase::realloc (dataSize, size, tileSize, borderMode);
}

template<class T>
inline
SampleBlockSize
SampleBlock<T>::
getSize () const
{
    return SampleBlockBase::getSize();
}

template<class T>
inline
SampleBlockSize
SampleBlock<T>::
getTileSize () const
{
    return SampleBlockBase::getTileSize();
}

template<class T>
inline
SampleBlockSize
SampleBlock<T>::
getNumberOfTiles () const
{
    return SampleBlockBase::getNumberOfTiles();
}

template<class T>
inline
BorderMode
SampleBlock<T>::
getBorderMode () const
{
    return SampleBlockBase::getBorderMode ();
}

template<class T>
inline
void
SampleBlock<T>::
setBorderMode (BorderMode borderMode)
{
    SampleBlockBase::setBorderMode (borderMode);
	return;
}

template<class T>
inline
T*
SampleBlock<T>::
getSamplePtr (indexType x, indexType y, indexType z, indexType band)
{
    return (T*) SampleBlockBase::getSamplePtr (x, y, z, band);
}

template<class T>
inline
const T*
SampleBlock<T>::
getSamplePtr (indexType x, indexType y, indexType z, indexType band) const
{
    return (T*) SampleBlockBase::getSamplePtr (x, y, z, band);
}

template<class T>
inline
T*
SampleBlock<T>::
getTilePtr (indexType x, indexType y, indexType z, indexType band) const
{
    return (T*) SampleBlockBase::getTilePtr (x, y, z, band);
}

template<class T>
inline
const offsetType&
SampleBlock<T>::
getSampleOffset (sizeType dimension, indexType i) const
{
    return SampleBlockBase::getSampleOffset (dimension, i);
}

template<class T>
inline
const offsetType&
SampleBlock<T>::
getSampleOffsetX (indexType i) const
{
    return SampleBlockBase::getSampleOffsetX (i);
}

template<class T>
inline
const offsetType&
SampleBlock<T>::
getSampleOffsetY (indexType i) const
{
    return SampleBlockBase::getSampleOffsetY (i);
}

template<class T>
inline
const offsetType&
SampleBlock<T>::
getSampleOffsetZ (indexType i) const
{
    return SampleBlockBase::getSampleOffsetZ (i);
}

template<class T>
inline
const offsetType&
SampleBlock<T>::
getSampleOffsetBand (indexType i) const
{
    return SampleBlockBase::getSampleOffsetBand (i);
}


template<class T>
inline
const offsetType&
SampleBlock<T>::
getTileOffset (sizeType dimension) const
{
    return SampleBlockBase::getTileOffset (dimension);
}

template<class T>
inline
const offsetType&
SampleBlock<T>::
getTileOffsetX () const
{
    return SampleBlockBase::getTileOffsetX ();
}

template<class T>
inline
const offsetType&
SampleBlock<T>::
getTileOffsetY () const
{
    return SampleBlockBase::getTileOffsetY ();
}

template<class T>
inline
const offsetType&
SampleBlock<T>::
getTileOffsetZ () const
{
    return SampleBlockBase::getTileOffsetZ ();
}

template<class T>
inline
const offsetType&
SampleBlock<T>::
getTileOffsetBand () const
{
    return SampleBlockBase::getTileOffsetBand ();
}

/*!

  execute function f for each element in SampleBlock

  \param f function pointer to function which takes a T* as argument

 */

template<class T>
inline
void
SampleBlock<T>::forEach(void(*f)(T*))
{
    for (indexType band = 0; band < size[DIM_BAND]; band++) {
        for (indexType z = 0; z < size[DIM_Z]; z++) {
            for (indexType y = 0; y < size[DIM_Y]; y++) {
                const offsetType* offsetPtr  = &getSampleOffsetX (0);
                T* posPtr = (T*) SampleBlockBase::getSamplePtr (0, y, z, band);
                for (indexType x = 0; x < size[DIM_X]; x++) {
                    f(posPtr);
                    posPtr += *offsetPtr;
                    offsetPtr += 1;
                } // for
            } // for
        } // for
    } // for

} // SampleBlock<T>::for_each


template<class T>
inline
void 
SampleBlock<T>::clear (T value)
{
    for (indexType band = 0; band < size[DIM_BAND]; band++) {
        for (indexType z = 0; z < size[DIM_Z]; z++) {
            for (indexType y = 0; y < size[DIM_Y]; y++) {
                const offsetType* offsetPtr  = &getSampleOffsetX (0);
                T* posPtr = (T*) SampleBlockBase::getSamplePtr (0, y, z, band);
                for (indexType x = 0; x < size[DIM_X]; x++) {
                    *posPtr = value;
                    posPtr += *offsetPtr;
                    offsetPtr += 1;
                } // for
            } // for
        } // for
    } // for

	return;
}


// destPos is the position in the destination sampleblock
template<class T>
inline
void 
SampleBlock<T>::copy (const SampleBlock &src, const SampleBlockPosition &destPos)
{
	SampleBlockSize sizeSrc = src.getSize();
	indexType minX=destPos[DIM_X];
	indexType maxX=minX+sizeSrc[DIM_X];
	if (maxX>size[DIM_X]) maxX=size[DIM_X];

	indexType minY=destPos[DIM_Y];
	indexType maxY=minY+sizeSrc[DIM_Y];
	if (maxY>size[DIM_Y]) maxY=size[DIM_Y];

	indexType minBand=destPos[DIM_BAND];
	indexType maxBand=minBand+sizeSrc[DIM_BAND];
	if (maxBand>size[DIM_BAND]) maxBand=size[DIM_BAND];

	indexType minZ=destPos[DIM_Z];
	indexType maxZ=minZ+sizeSrc[DIM_Z];
	if (maxZ>size[DIM_Z]) maxZ=size[DIM_Z];

    for (indexType band = minBand; band < maxBand; band++) {
        for (indexType z = minZ; z < maxZ; z++) {
	        for (indexType y = minY; y < maxY; y++) {

				// pointer and offset to dest
                const offsetType* destOff  = &getSampleOffsetX (minX);
				T* destPtr = (T*) SampleBlockBase::getSamplePtr (minX, y, z, band);

				// pointer and offset to src
                const offsetType* srcOff  = &src.getSampleOffsetX (0);
				T* srcPtr = (T*) src.getSamplePtr (0, y-minY, z-minZ, band-minBand);

				// copy values
				for (indexType x = minX; x < maxX; x++) {
					*destPtr = *srcPtr;
					srcPtr  += *srcOff;
					destPtr += *destOff;
                    srcOff++;
                    destOff++;
				} // for
			} // for
        } // for
    } // for
	return;
}




// destPos is the position in the destination sampleblock
template<class T>
inline
void 
SampleBlock<T>::copy (const SampleBlock &src, const SampleBlockPosition &srcPos, const SampleBlockPosition &destPos)
{
	SampleBlockSize sizeSrc = src.getSize();
	sizeSrc[DIM_X]   -=srcPos[DIM_X];
	sizeSrc[DIM_Y]   -=srcPos[DIM_Y];
	sizeSrc[DIM_Z]   -=srcPos[DIM_Z];
	sizeSrc[DIM_BAND]-=srcPos[DIM_BAND];

	indexType minX=destPos[DIM_X];
	indexType maxX=minX+sizeSrc[DIM_X];
	if (maxX>size[DIM_X]) maxX=size[DIM_X];

	indexType minY=destPos[DIM_Y];
	indexType maxY=minY+sizeSrc[DIM_Y];
	if (maxY>size[DIM_Y]) maxY=size[DIM_Y];

	indexType minBand=destPos[DIM_BAND];
	indexType maxBand=minBand+sizeSrc[DIM_BAND];
	if (maxBand>size[DIM_BAND]) maxBand=size[DIM_BAND];

	indexType minZ=destPos[DIM_Z];
	indexType maxZ=minZ+sizeSrc[DIM_Z];
	if (maxZ>size[DIM_Z]) maxZ=size[DIM_Z];

    for (indexType band = minBand; band < maxBand; band++) {
        for (indexType z = minZ; z < maxZ; z++) {
	        for (indexType y = minY; y < maxY; y++) {

				// pointer and offset to src
                const offsetType* destOff  = &getSampleOffsetX (minX);
				T* destPtr = (T*) SampleBlockBase::getSamplePtr (minX, y, z, band);

				// pointer and offset to dest
                const offsetType* srcOff  = &src.getSampleOffsetX (srcPos[DIM_X]);
				T* srcPtr = (T*) src.getSamplePtr (srcPos[DIM_X], y-minY+srcPos[DIM_Y], z-minZ+srcPos[DIM_Z], band-minBand+srcPos[DIM_BAND]);

				// copy values
				for (indexType x = minX; x < maxX; x++) {
					*destPtr = *srcPtr;
					srcPtr  += *srcOff;
					destPtr += *destOff;
                    srcOff++;
                    destOff++;
				} // for
			} // for
        } // for
    } // for
	return;
}


// destPos is the position in the destination sampleblock
// all pixels are copied, except pixels with the specified color
template<class T>
inline
void 
SampleBlock<T>::copyExcept (const SampleBlock &src, const SampleBlockPosition &destPos, T color)
{
	SampleBlockSize sizeSrc = src.getSize();
	indexType minX=destPos[DIM_X];
	indexType maxX=minX+sizeSrc[DIM_X];
	if (maxX>size[DIM_X]) maxX=size[DIM_X];

	indexType minY=destPos[DIM_Y];
	indexType maxY=minY+sizeSrc[DIM_Y];
	if (maxY>size[DIM_Y]) maxY=size[DIM_Y];

	indexType minBand=destPos[DIM_BAND];
	indexType maxBand=minBand+sizeSrc[DIM_BAND];
	if (maxBand>size[DIM_BAND]) maxBand=size[DIM_BAND];

	indexType minZ=destPos[DIM_Z];
	indexType maxZ=minZ+sizeSrc[DIM_Z];
	if (maxZ>size[DIM_Z]) maxZ=size[DIM_Z];

    for (indexType band = minBand; band < maxBand; band++) {
        for (indexType z = minZ; z < maxZ; z++) {
	        for (indexType y = minY; y < maxY; y++) {

				// pointer and offset to dest
                const offsetType* destOff  = &getSampleOffsetX (minX);
				T* destPtr = (T*) SampleBlockBase::getSamplePtr (minX, y, z, band);

				// pointer and offset to src
                const offsetType* srcOff  = &src.getSampleOffsetX (0);
				T* srcPtr = (T*) src.getSamplePtr (0, y-minY, z-minZ, band-minBand);

				// copy values
				for (indexType x = minX; x < maxX; x++) {
					if (*srcPtr != color) {
						*destPtr = *srcPtr;
					} // if
					srcPtr  += *srcOff;
					destPtr += *destOff;
                    srcOff++;
                    destOff++;
				} // for
			} // for
        } // for
    } // for
	return;
}




// destPos is the position in the destination sampleblock
template<class T>
inline
void 
SampleBlock<T>::copyExcept (const SampleBlock &src, const SampleBlockPosition &srcPos, 
					        const SampleBlockPosition &destPos, T color)
{
	SampleBlockSize sizeSrc = src.getSize();
	sizeSrc[DIM_X]   -=srcPos[DIM_X];
	sizeSrc[DIM_Y]   -=srcPos[DIM_Y];
	sizeSrc[DIM_Z]   -=srcPos[DIM_Z];
	sizeSrc[DIM_BAND]-=srcPos[DIM_BAND];

	indexType minX=destPos[DIM_X];
	indexType maxX=minX+sizeSrc[DIM_X];
	if (maxX>size[DIM_X]) maxX=size[DIM_X];

	indexType minY=destPos[DIM_Y];
	indexType maxY=minY+sizeSrc[DIM_Y];
	if (maxY>size[DIM_Y]) maxY=size[DIM_Y];

	indexType minBand=destPos[DIM_BAND];
	indexType maxBand=minBand+sizeSrc[DIM_BAND];
	if (maxBand>size[DIM_BAND]) maxBand=size[DIM_BAND];

	indexType minZ=destPos[DIM_Z];
	indexType maxZ=minZ+sizeSrc[DIM_Z];
	if (maxZ>size[DIM_Z]) maxZ=size[DIM_Z];

    for (indexType band = minBand; band < maxBand; band++) {
        for (indexType z = minZ; z < maxZ; z++) {
	        for (indexType y = minY; y < maxY; y++) {

				// pointer and offset to src
                const offsetType* destOff  = &getSampleOffsetX (minX);
				T* destPtr = (T*) SampleBlockBase::getSamplePtr (minX, y, z, band);

				// pointer and offset to dest
                const offsetType* srcOff  = &src.getSampleOffsetX (srcPos[DIM_X]);
				T* srcPtr = (T*) src.getSamplePtr (srcPos[DIM_X], y-minY+srcPos[DIM_Y], z-minZ+srcPos[DIM_Z], band-minBand+srcPos[DIM_BAND]);

				// copy values
				for (indexType x = minX; x < maxX; x++) {
					if (*srcPtr != color) {
						*destPtr = *srcPtr;
					} // if
					srcPtr  += *srcOff;
					destPtr += *destOff;
                    srcOff++;
                    destOff++;
				} // for
			} // for
        } // for
    } // for
	return;
}





// find the position of the minimum value in the sampleblock
template<class T>
inline
void 
SampleBlock<T>::minValue (SampleBlockPosition &pos)
{
	pos[DIM_X]=0;
	pos[DIM_Y]=0;
	pos[DIM_Z]=0;
	pos[DIM_BAND]=0;
	T v = *((T*)SampleBlockBase::getSamplePtr (0,0,0,0));

    for (indexType band = 0; band < size[DIM_BAND]; band++) {
        for (indexType z = 0; z < size[DIM_Z]; z++) {
	        for (indexType y = 0; y < size[DIM_Y]; y++) {

				// pointer and offset to src
                const offsetType* off  = &getSampleOffsetX (0);
				T* ptr = (T*) SampleBlockBase::getSamplePtr (0, y, z, band);

				// copy values
				for (indexType x = 0; x < size[DIM_X]; x++) {
					if (v>*ptr) {
						v=*ptr;
						pos[DIM_X]=x;
						pos[DIM_Y]=y;
						pos[DIM_Z]=z;
						pos[DIM_BAND]=band;
					} // if
					ptr  += *off;
                    off++;
				} // for
			} // for
        } // for
    } // for

	return;
}


// find the position of the maximum value in the sampleblock
template<class T>
inline
void 
SampleBlock<T>::maxValue (SampleBlockPosition &pos)
{
	pos[DIM_X]=0;
	pos[DIM_Y]=0;
	pos[DIM_Z]=0;
	pos[DIM_BAND]=0;
	T v = *((T*)SampleBlockBase::getSamplePtr (0,0,0,0));

    for (indexType band = 0; band < size[DIM_BAND]; band++) {
        for (indexType z = 0; z < size[DIM_Z]; z++) {
	        for (indexType y = 0; y < size[DIM_Y]; y++) {

				// pointer and offset to src
                const offsetType* off  = &getSampleOffsetX (0);
				T* ptr = (T*) SampleBlockBase::getSamplePtr (0, y, z, band);

				// copy values
				for (indexType x = 0; x < size[DIM_X]; x++) {
					if (v<*ptr) {
						v=*ptr;
						pos[DIM_X]=x;
						pos[DIM_Y]=y;
						pos[DIM_Z]=z;
						pos[DIM_BAND]=band;
					} // if
					ptr  += *off;
                    off++;
				} // for
			} // for
        } // for
    } // for

	return;
}


// find the minimum value in the sampleblock
template<class T>
inline T SampleBlock<T>::minValue ()
{
	SampleBlockPosition pos;
	minValue (pos);
	T v = *((T*)SampleBlockBase::getSamplePtr (pos[DIM_X],pos[DIM_Y],pos[DIM_Z],pos[DIM_BAND]));

	return v;
}

// find the maximum value in the sampleblock
template<class T>
inline T SampleBlock<T>::maxValue ()
{
	SampleBlockPosition pos;
	maxValue (pos);
	T v = *((T*)SampleBlockBase::getSamplePtr (pos[DIM_X],pos[DIM_Y],pos[DIM_Z],pos[DIM_BAND]));

	return v;
}


// send data to stdout, a file, ...
template<class T>
std::ostream& operator<< (ostream& os, const SampleBlock<T>& data)
{
	SampleBlockSize size = data.getSize();
    for (indexType band = 0; band < size[DIM_BAND]; band++) {
        for (indexType z = 0; z < size[DIM_Z]; z++) {
            for (indexType y = 0; y < size[DIM_Y]; y++) {
                const offsetType* offsetPtr  = &data.getSampleOffsetX (0);
                T* posPtr = (T*) data.getSamplePtr (0, y, z, band);
                for (indexType x = 0; x < size[DIM_X]; x++) {
					os << (*posPtr) << endl;
                    posPtr += *offsetPtr;
                    offsetPtr += 1;
                } // for
            } // for
        } // for
    } // for

	return os;
} // operator<<


#endif
