/*
    LUsolve.cpp     Solution of a linear equation system using LU decomposition
    Copyright (C) 2003 Markus Beier

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <LUsolve.h>

#define TINY 1.0e-20

void
LUsolve::
ludcmp(Matrix<ieee64>& a, int& n, Vector<int>& indx, ieee64& d) throw (MocaException)
{
	int i,imax=0,j,k;
	double big,dum,sum,temp;

    Vector<ieee64> vv(n, 1);

	d=1.0;
	for (i=1;i<=n;i++) {
		big=0.0;
		for (j=1;j<=n;j++)
			if ((temp=fabs(a[i][j])) > big) big=temp;
		if (big == 0.0) {
            throw MocaException("Singular matrix in routine LUDCMP");
        }
		vv[i]=1.0/big;
	}
	for (j=1;j<=n;j++) {
		for (i=1;i<j;i++) {
			sum=a[i][j];
			for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
		}
		big=0.0;
		for (i=j;i<=n;i++) {
			sum=a[i][j];
			for (k=1;k<j;k++)
				sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
			if ( (dum=vv[i]*fabs(sum)) >= big) {
				big=dum;
				imax=i;
			}
		}
		if (j != imax) {
			for (k=1;k<=n;k++) {
				dum=a[imax][k];
				a[imax][k]=a[j][k];
				a[j][k]=dum;
			}
			d = -d;
			vv[imax]=vv[j];
		}
		indx[j]=imax;
		if (a[j][j] == 0.0) a[j][j]=TINY;
		if (j != n) {
			dum=1.0/(a[j][j]);
			for (i=j+1;i<=n;i++) a[i][j] *= dum;
		}
	}

}


void
LUsolve::
lubksb(Matrix<ieee64>& a, int& n, Vector<int>& indx, Vector<ieee64>& b)
{
	int32 i,ii=0,ip,j;
	double sum;

	for (i=1;i<=n;i++) {
		ip=indx[i];
		sum=b[ip];
		b[ip]=b[i];
		if (ii)
			for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
		else if (sum) ii=i;
		b[i]=sum;
	}
	for (i=n;i>=1;i--) {
		sum=b[i];
		for (j=i+1;j<=n;j++) sum -= a[i][j]*b[j];
		b[i]=sum/a[i][i];
	}
}


void
LUsolve::
solve (Matrix<ieee64>& A, Vector<ieee64>& a, Vector<ieee64>& b) throw (MocaException)
{
    A.setIndexStart(1);

    b.setIndexStart(1);
    a.setIndexStart(1);

    sizeType rows = (A.getSize())[DIM_ROW];
    sizeType cols = (A.getSize())[DIM_COL];

    if (rows != cols) {
        throw MocaException("LUsolve.solve: Input Matrix must be square");

    }

    // LU decomposition
    Vector<int> indx(cols, 1);
    ieee64 d;

	int tmpRows = rows;
    ludcmp (A, tmpRows, indx, d);
	rows = tmpRows;

#if 0
	// debug
    cout << "Decomposed Matrix A:" << endl;
    for (int i = 1; i <= rows; i++) {
        for (int j = 1; j <= rows; j++) {
            cout << A[i][j] << " ";
        }
        cout << endl;
    }
#endif



    // check, if det(A) != 0
    for (int i=1 ; i <=rows; i++) {
        if ( fabs(A[i][i]) < 1.0e-14 ) {
            throw MocaException("LUsolve.solve: Input Matrix is singular");
        }
    }

    // back substitution
	tmpRows = rows;
	lubksb(A, tmpRows, indx, b);
	rows = tmpRows;


    for (int j = 1; j <= rows; j++) {
        a[j] = b[j];
    }


    A.setIndexStart(0);
    b.setIndexStart(0);
    a.setIndexStart(0);

    return;
}
