/*
    SvdFit.cpp     least squares fit by singular value decomposition
    Copyright (C) 2002 Gerald Kuehne

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <SvdFit.h>

static ieee64 at,bt,ct;
#define PYTHAG(a,b) ((at=fabs(a)) > (bt=fabs(b)) ? \
(ct=bt/at,at*sqrt(1.0+ct*ct)) : (bt ? (ct=at/bt,bt*sqrt(1.0+ct*ct)): 0.0))

static ieee64 maxarg1,maxarg2;
#define SVD_MAX(a,b) (maxarg1=(a),maxarg2=(b),(maxarg1) > (maxarg2) ?\
	(maxarg1) : (maxarg2))
#define SVD_SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))

void
SvdFit::
svdcmp(Matrix<ieee64>& a, Vector<ieee64>& w, Matrix<ieee64>& v) throw (MocaException)
{
    int m = (a.getSize())[DIM_ROW];
    int n = (a.getSize())[DIM_COL];

	int flag,i,its,j,jj,k,l=0,nm=0;
	ieee64 c,f,h,s,x,y,z;
	ieee64 anorm=0.0,g=0.0,scale=0.0;
	Vector<ieee64> rv1(n, 1);

	if (m < n) {
        throw MocaException("SVDCMP: You must augment A with extra zero rows");
    }

	for (i=1;i<=n;i++) {
		l=i+1;
		rv1[i]=scale*g;
		g=s=scale=0.0;
		if (i <= m) {
			for (k=i;k<=m;k++) scale += fabs(a[k][i]);
			if (scale) {
				for (k=i;k<=m;k++) {
					a[k][i] /= scale;
					s += a[k][i]*a[k][i];
				}
				f=a[i][i];
				g = -SVD_SIGN(sqrt(s),f);
				h=f*g-s;
				a[i][i]=f-g;
				if (i != n) {
					for (j=l;j<=n;j++) {
						for (s=0.0,k=i;k<=m;k++) s += a[k][i]*a[k][j];
						f=s/h;
						for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
					}
				}
				for (k=i;k<=m;k++) a[k][i] *= scale;
			}
		}
		w[i]=scale*g;
		g=s=scale=0.0;
		if (i <= m && i != n) {
			for (k=l;k<=n;k++) scale += fabs(a[i][k]);
			if (scale) {
				for (k=l;k<=n;k++) {
					a[i][k] /= scale;
					s += a[i][k]*a[i][k];
				}
				f=a[i][l];
				g = -SVD_SIGN(sqrt(s),f);
				h=f*g-s;
				a[i][l]=f-g;
				for (k=l;k<=n;k++) rv1[k]=a[i][k]/h;
				if (i != m) {
					for (j=l;j<=m;j++) {
						for (s=0.0,k=l;k<=n;k++) s += a[j][k]*a[i][k];
						for (k=l;k<=n;k++) a[j][k] += s*rv1[k];
					}
				}
				for (k=l;k<=n;k++) a[i][k] *= scale;
			}
		}
		anorm=SVD_MAX(anorm,(fabs(w[i])+fabs(rv1[i])));
	}
	for (i=n;i>=1;i--) {
		if (i < n) {
			if (g) {
				for (j=l;j<=n;j++)
					v[j][i]=(a[i][j]/a[i][l])/g;
				for (j=l;j<=n;j++) {
					for (s=0.0,k=l;k<=n;k++) s += a[i][k]*v[k][j];
					for (k=l;k<=n;k++) v[k][j] += s*v[k][i];
				}
			}
			for (j=l;j<=n;j++) v[i][j]=v[j][i]=0.0;
		}
		v[i][i]=1.0;
		g=rv1[i];
		l=i;
	}
	for (i=n;i>=1;i--) {
		l=i+1;
		g=w[i];
		if (i < n)
			for (j=l;j<=n;j++) a[i][j]=0.0;
		if (g) {
			g=1.0/g;
			if (i != n) {
				for (j=l;j<=n;j++) {
					for (s=0.0,k=l;k<=m;k++) s += a[k][i]*a[k][j];
					f=(s/a[i][i])*g;
					for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
				}
			}
			for (j=i;j<=m;j++) a[j][i] *= g;
		} else {
			for (j=i;j<=m;j++) a[j][i]=0.0;
		}
		++a[i][i];
	}
	for (k=n;k>=1;k--) {
		for (its=1;its<=30;its++) {
			flag=1;
			for (l=k;l>=1;l--) {
				nm=l-1;
				if ((ieee64)(fabs(rv1[l])+anorm) == anorm) {
					flag=0;
					break;
				}
				if ((ieee64)(fabs(w[nm])+anorm) == anorm) break;
			}
			if (flag) {
				c=0.0;
				s=1.0;
				for (i=l;i<=k;i++) {
					f=s*rv1[i];
					rv1[i]=c*rv1[i];
					if ((ieee64)(fabs(f)+anorm) == anorm) break;
					g=w[i];
					h=PYTHAG(f,g);
					w[i]=h;
					h=1.0/h;
					c=g*h;
					s=(-f*h);
					for (j=1;j<=m;j++) {
						y=a[j][nm];
						z=a[j][i];
						a[j][nm]=y*c+z*s;
						a[j][i]=z*c-y*s;
					}
				}
			}
			z=w[k];
			if (l == k) {
				if (z < 0.0) {
					w[k] = -z;
					for (j=1;j<=n;j++) v[j][k]=(-v[j][k]);
				}
				break;
			}
			if (its == 30) {
                throw MocaException("No convergence in 30 SVDCMP iterations");
            }

			x=w[l];
			nm=k-1;
			y=w[nm];
			g=rv1[nm];
			h=rv1[k];
			f=((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
			g=PYTHAG(f,1.0);
			f=((x-z)*(x+z)+h*((y/(f+SVD_SIGN(g,f)))-h))/x;
			c=s=1.0;
			for (j=l;j<=nm;j++) {
				i=j+1;
				g=rv1[i];
				y=w[i];
				h=s*g;
				g=c*g;
				z=PYTHAG(f,h);
				rv1[j]=z;
				c=f/z;
				s=h/z;
				f=x*c+g*s;
				g=g*c-x*s;
				h=y*s;
				y=y*c;
				for (jj=1;jj<=n;jj++) {
					x=v[jj][j];
					z=v[jj][i];
					v[jj][j]=x*c+z*s;
					v[jj][i]=z*c-x*s;
				}
				z=PYTHAG(f,h);
				w[j]=z;
				if (z) {
					z=1.0/z;
					c=f*z;
					s=h*z;
				}
				f=(c*g)+(s*y);
				x=(c*y)-(s*g);
				for (jj=1;jj<=m;jj++) {
					y=a[jj][j];
					z=a[jj][i];
					a[jj][j]=y*c+z*s;
					a[jj][i]=z*c-y*s;
				}
			}
			rv1[l]=0.0;
			rv1[k]=f;
			w[k]=x;
		}
	}
}

void
SvdFit::
svbksb(Matrix<ieee64>& u, Vector<ieee64>& w, Matrix<ieee64>& v,
       Vector<ieee64>& b, Vector<ieee64>& x)
{
    int m = (u.getSize())[DIM_ROW];
    int n = (u.getSize())[DIM_COL];

    Vector<ieee64> tmp(n, 1);
	int j;
	for (j = 1; j <= n; j++) {
		ieee64 s = 0.0;
		if (w[j]) {
			for (int i = 1; i <= m; i++) s += u[i][j]*b[i];
			s /= w[j];
		}
		tmp[j]=s;
	}
	for (j = 1; j <= n; j++) {
		ieee64 s = 0.0;
		for (int jj = 1; jj <= n; jj++) s += v[j][jj]*tmp[jj];
		x[j]=s;
	}
}


#define TOLERANCE 1.0e-5

ieee64
SvdFit::
solve (Matrix<ieee64>& A, Vector<ieee64>& a, Vector<ieee64>& b)
{
    A.setIndexStart(1);
    Matrix<ieee64> u = A;

    b.setIndexStart(1);
    a.setIndexStart(1);


    sizeType rows = (A.getSize())[DIM_ROW];
    sizeType cols = (A.getSize())[DIM_COL];

    Matrix<ieee64> v(MatrixSize(cols,cols), 1);
    Vector<ieee64> w(cols, 1);

    // singular value decomposition
    svdcmp(u,w,v);

	ieee64 wmax = 0.0;
	int j;
	for (j = 1; j <= cols; j++)
		if (w[j] > wmax)
            wmax = w[j];

	ieee64 thresh = TOLERANCE * wmax;
	for (j = 1; j <= cols; j++)
		if (w[j] < thresh)
            w[j]=0.0;

    // back substitution
	svbksb(u,w,v,b,a);

    // error calculation
	ieee64 chisq = 0.0;
	for (int i = 1; i <= rows; i++) {
        ieee64 sum = 0.0;
		for (int j = 1; j <= cols; j++) {
            sum += a[j]*A[i][j];
        }
        ieee64 tmp = b[i] - sum;
        chisq += tmp*tmp;
	}

    A.setIndexStart(0);
    b.setIndexStart(0);
    a.setIndexStart(0);

    return chisq;
}




