/* 
 * vo_stdout.c, interface to send raw data to stdout 
 *
 *
 * Hacked into mplayer by
 * 
 * Stephan Kopf <kopf@informatik.uni-mannheim.de>
 * 
 * 
 * based on code from:
 * 
 * Copyright (C) 1996, MPEG Software Simulation Group. All Rights Reserved. 
 *
 * Aaron Holtzman <aholtzma@ess.engr.uvic.ca>
 *
 * 15 & 16 bpp support added by Franck Sicard <Franck.Sicard@solsoft.fr>
 *
 * Xv image suuport by Gerd Knorr <kraxel@goldbach.in-berlin.de>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#ifdef WIN32
// include routines to switch to binary mode  
#include <fcntl.h>
#endif

#include "config.h"
#include "video_out.h"
#include "video_out_internal.h"

#define STDOUT_DECODE  'd'  /* decode video (yuv) to stdout (default)     */
#define STDOUT_STATUS  's'  /* get status information (number of frames)  */
static int config_stdout = STDOUT_DECODE;
 

static vo_info_t info = 
{
	"send decoded video data to stdout",
	"stdout",
	"Stephan Kopf <kopf@informatik.uni-mannheim.de>",
	""
};

LIBVO_EXTERN (stdout)

static int image_width;
static int image_height;
static int num_bytes;
static int framenum = 0;

static uint8_t *image=NULL;

static uint32_t
config(uint32_t width, uint32_t height, uint32_t d_width, uint32_t d_height, uint32_t fullscreen, char *title, uint32_t format)
{
    image_height = height;
    image_width = width;
    num_bytes = width*height*3/2;

    if (config_stdout != STDOUT_STATUS) {
        if(image) free(image);
	image=NULL;
	image=malloc(num_bytes);
    }
    return 0;
}

static void draw_osd(void)
{
}

static void flip_page (void)
{
    // count frames, do not copy image data
    framenum++;
    if (config_stdout == STDOUT_STATUS) {
      return;
    }
    
    // first frame?
    if (framenum==1) {
	fprintf(stdout, "\n\nMPLAYER-START-DECODING");
    }

    // write data
    int n_bytes = fwrite(image, 1, num_bytes, stdout);
    if (n_bytes != num_bytes) {
      if (n_bytes == 0) {
	//perror("mplayer (vo_stdout): Error writing image to stdout!");
	exit(1);
      }
      //perror("mplayer (vo_stdout): Error writing image to stdout!");
    }
    return;
}

static uint32_t draw_slice(uint8_t *srcimg[], int stride[], int w,int h,int x,int y)
{
    // count frames, do not copy image data
    if (config_stdout == STDOUT_STATUS) {
      return 0;
    }

    int i;
    // copy Y:
    uint8_t *dst=image+image_width*y+x;
    uint8_t *src=srcimg[0];
    for(i=0;i<h;i++){
        memcpy(dst,src,w);
        src+=stride[0];
        dst+=image_width;
    }
    {
        // copy U+V:
        uint8_t *src1=srcimg[1];
	uint8_t *src2=srcimg[2];
	uint8_t *dst=image+image_width*image_height+image_width*(y/2)+(x/2);
	for(i=0;i<h/2;i++){
	  memcpy(dst,src1,w/2);
	  memcpy(dst+image_width/2,src2,w/2);
	  src1+=stride[1];
	  src2+=stride[2];
	  dst+=image_width;
	}
    }    
    return 0;
}


static uint32_t draw_frame(uint8_t * src[])
{
    return 0;
}

static uint32_t
query_format(uint32_t format)
{
    if(format==IMGFMT_YV12) return 1;
    return 0;
}

static void
uninit(void)
{
    // write number of frames to stdout
    if (config_stdout == STDOUT_STATUS) {
	fprintf(stdout, "\nFRAMES: %d\n",framenum);
    }

    if(image) free(image);
    image=NULL;
}


static void check_events(void)
{
}

static uint32_t preinit(const char *arg)
{
    if(arg) 
    {
        int arg_unrecognized = 0;
	if (strcmp(arg, "status"))
            arg_unrecognized++;
	else
	    config_stdout = STDOUT_STATUS;

        /* If the test failed the argument is invalid */
        if (arg_unrecognized == 1)
        {
	        fprintf(stderr, "vo_stdout: Unknown subdevice: %s\n",arg);
		return ENOSYS;
	}
    }

    // do not write any messages, warnings or errors to stdout
    mp_msg_set_level(-1);


#ifdef WIN32
    //  Switch Win32 standard output to binary mode, so that piping the
    //  raw image data works. 
    int result = _setmode( _fileno(stdout), _O_BINARY );
    if (result == -1) 
    {
	fprintf(stderr, "mplayer (vo_stdout): Cannot set standard output to binary mode.\n");
	return ENOSYS;
    }
#endif

    return 0;
}

static uint32_t control(uint32_t request, void *data, ...)
{
  switch (request) {
  case VOCTRL_QUERY_FORMAT:
    return query_format(*((uint32_t*)data));
  }
  return VO_NOTIMPL;
}
